#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
__version__=''' $Id: apploggers.py 13507 2004-07-15 09:48:55Z rgbecker $ '''
__all__=('AppLogEntry','AppLogger','FileAppLogger','OptionalFileAppLogger')
import marshal, base64, zlib, time, os, sys

class AppLogEntry:
    '''A WebApp log entry''' 
    def __init__(self):
        self._records = []

    def _encrypt(self,obj):
        try:
            return marshal.dumps(obj)
        except:
            from errhandlers import ShowTracebackHandler 
            return marshal.dumps(ShowTracebackHandler().handleException(debugMemo=0))

    def _decrypt(self,cobj):
        return marshal.loads(cobj)

    def record(self,label,**kw):
        self._records.append((label,self._encrypt(kw)))

    def store_to_string(self):
        '''returns a storable form of our records'''
        D = self.__dict__.copy()
        del D['_records']
        self.record('__dict__',__dict__=D)
        return base64.encodestring(zlib.compress(marshal.dumps(self._records))).replace('\n','')

    def restore_from_string(self,ctext):
        '''replace records with version obtained by decoding string'''
        self._records = marshal.loads(zlib.decompress(base64.decodestring(ctext)))
        self.__dict__.update(self._decrypt(self._records[-1][1])['__dict__'])
        del self._records[-1]

    def __getitem__(self,i):
        r = [self._decrypt(x[1]) for x in self._records if x[0]==i]
        if len(r)==0: raise IndexError(i)
        if len(r)==1: return r[0]
        return r

    def __len__(self):
        return len(self._records)

    def __str__(self):
        records = self._records
        if not records: return '[]'
        from reportlab.lib.utils import getStringIO
        from pprint import pprint
        f = getStringIO()
        f.write('[\n')
        for k,v in records:
            pprint((k,self._decrypt(v)),f)
        f.write(']\n')
        return f.getvalue()

    def __repr__(self):
        return self.store_to_string()

class AppLogger:
    def __init__(self,**kw):
        '''allowed keywords logFN, appName'''
        self._timeStarted = time.time()
        from rlextra.utils.cgisupport import TimeStamp
        ts = TimeStamp()
        self._cds, self.id = ts.currentDateString, ts.currentId
    def store(self, **kw):
        pass
    def restore(self, key):
        pass
    def locate_logs(self):
        return []
    def record(self,label,**kw):
        pass
    def __getitem__(self,i):
        raise KeyError(i)

class FileAppLogger(AppLogger):
    def __init__(self,**kw):
        def pop(d, key, default):
            if d.has_key(key):
                v = d[key]
                del d[key]
                return v
            else:
                return default
        logFN = pop(kw, 'logFN','')
        logDir = pop(kw, 'logDir','')
        appName = pop(kw, 'appName','')
        if kw: raise ValueError('Unexpected arguments '+`kw`)
        AppLogger.__init__(self)

        if not appName:
            if not appName:
                appName = os.environ.get('SCRIPT_NAME','')
                if not appName:
                    appName = sys.argv[0]
                    if not appName:
                        try: 1/0
                        except: frame = sys.exc_traceback.tb_frame
                        frame = frame.f_back
                        upLocals = frame.f_locals
                        req = upLocals.get('req',None)
                        if req: appName = getattr(req,'url','')
                if appName: appName=os.path.splitext(os.path.basename(appName))[0]
            if not appName: appName = 'unknownapp'
        self._appName = appName
        if logFN:
            logFN = os.path.abspath(logFN)
        else:
            if not logDir: logDir = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(sys.argv[0]))),'logs',appName)
            logFN = self._make_logFN(logDir=os.path.abspath(logDir))
        assert logFN.endswith('-log.txt'), "Bad filename %s" % logFN
        self._logFN = logFN
        self._entry = AppLogEntry()

    def _make_logFN(self,cds=None,logDir=None):
        logFN = '%s-%s-log.txt' % (self._appName,cds or self._cds)
        if not logDir: logDir = os.path.dirname(self._logFN)
        return os.path.join(logDir,logFN)

    def record(self,label,**kw):
        self._entry.record(label,**kw)

    def _get_logFN(self):
        logFN = self._logFN
        logDir = os.path.dirname(logFN)
        if logDir:
            try:
                os.makedirs(logDir)
            except OSError:
                pass
        return logFN

    def store(self):
        e = self._entry
        kw = {}
        kw['timeStopped'] = t = time.time()
        kw['timing'] = e.timing = t - self._timeStarted
        kw['timeStarted'] = e.timeStarted = self._timeStarted
        self.record('__store__',**kw)
        logFN = self._get_logFN()
        try: #safe attempt at writing the log
            f = open(logFN,'a')
            f.write('%s %s\n' % (self.id,e.store_to_string()))
            f.close()
        except:
            pass

    def restore(self, key):
        logFN = self._get_logFN()
        f = open(logFN,'r')
        for l in f.readlines():
            l = l.strip().split()
            if l[0]==key:
                self._entry.restore_from_string(l[1])
                return
        raise KeyError(key)

    def logfiles(self):
        import glob
        return glob.glob(self._make_logFN(cds='*'))

    def __getitem__(self,i):
        return self._entry[i]

    def _getentry(self, key):
        cds = key.split('-')[-2][:-6]
        logFN = self._make_logFN(cds=cds)
        l = self.__class__(logFN=logFN)
        l.restore(key)
        return l._entry

    def _rerun_from_entry(self, e, app, response):
        from webapp import WebRequest
        req = WebRequest()
        req.params = e['startcgi']['params']
        resp = app.handleRequest(req)
        response.__dict__ = resp.__dict__
        return

    def rerun_from_key(self, key, app, response):
        self._rerun_from_entry(self._getentry(key),app,response)

    def rerun_from_entry_text(self, text, app, response):
        text = text.strip().split()
        key = text[0]
        e = AppLogEntry()
        e.restore_from_string(''.join(text[1:]))
        self._rerun_from_entry(e,app,response)

    def entryListing(self, back=0, limit=100, logFN=None, script='',
            list_action="test_list_log_entry",probe_action="test_log_probe"):
        """hyperlinked text of entries going back from most recent"""
        # this implementation may not be advisable if the log gets huge
        text = "<h1>Recent Log lists going back %s</h1>" % back
        try:
            a = open(logFN or self._logFN,'r')
            log = a.readlines()
            a.close()
        except:
            log = []

        L = []
        a = L.append
        for l in log:
            l = l.split()
            k = l[0]
            e = AppLogEntry()
            e.restore_from_string(''.join(l[1:]))

            action = "(none)"
            try:
                params = e['startcgi']['params']
                action = params["action"]
            except:
                pass
            try:
                error = e['error']
            except:
                error = None
            if error:
                try:
                    et = error.get('error_type','')
                    ev = error.get('error_value','')
                    from xml.sax.saxutils import escape
                    if et or ev: ev = ' error=%s(%s)' % (et,escape(ev))
                except:
                    ev = ''
                i = '<a href="%s?action=%s&id=%s"><font color="red">id=%s action=%s%s</font></a>' % (
                        script, list_action, k, k, action, ev)
            else:
                i = '<a href="%s?action=%s&id=%s">id=%s action=%s</a>' % (script,list_action,k,k,action)

            a(i)
        # cut out the requested entries
        n = len(L)
        if limit is None: limit = n
        last = max(0,n-back)
        first = max(0, n-back-limit)
        L[:] = L[first:last]
        if first>0:
            nextlink = """\n<a href="%s?action=&back=%s">get more</a>"""% (script,probe_action,(back+limit))
            a(nextlink)
        return text + "\n<pre>\n%s\n<pre>" % "\n".join(L)

    def html_logfiles_form(self,params,
            list_action="test_list_log_entry",probe_action="test_log_probe"):
        '''returns an html logfiles form'''
        contents = []
        write = contents.append
        from rlextra.utils import htmldbhelpers
        logFN = params.get("logfilename", self._logFN)
        write("current file is %s <br>" % logFN)
        logs = self.logfiles()
        write("""
            <form action="%s">
            <input type="hidden" name="action" value="%s">
            <input type="submit" name="use_file" value="use_file">
            """ % (params['SCRIPT_NAME'],probe_action))
        write(htmldbhelpers.dayselect(name="logfilename", selected=logFN, values=logs, names=logs))
        write("""</form>""")
        write(self.entryListing(int(params.get("back", "0")), logFN=logFN, script=params['SCRIPT_NAME'],
            probe_action=probe_action, list_action=list_action))
        return '\n'.join(contents)

    def html_logentry_form(self,params,retry_action="test_log_entry_retry"):
        '''returns html for a single log entry'''
        contents = []
        write = contents.append
        id = params["id"]
        e = self._getentry(id)
        write('<div style="text-align:left">')
        write("<h1>Log entry</h1>")
        write("<pre>")
        write(repr(e)) #shoudn't need escaping
        write("</pre>")
        write("""<a href="%s?action=%s&id=%s">Retry this entry</a>""" % (params['SCRIPT_NAME'],retry_action,id))
        write("<br><h1>Decoded version of the above</h1><pre>")
        from xml.sax.saxutils import escape
        write(escape(str(e)))
        write("</pre>")
        write('</div>')
        return '\n'.join(contents)

def OptionalFileAppLogger(logDir):
    if logDir: return lambda logDir=logDir: FileAppLogger(logDir=logDir)
    return AppLogger
    

class FileAppLogProbeMixin:
    '''may be mixed into a a WebApp to provide logprobing capability''' 

    def action_test_log_probe(self, request, response):
        params = request.params
        self.error_if_live(response)
        self.runView(request,response,"std_header")
        action = self.getActionName(request)
        response.write(self.logger.html_logfiles_form(params))
        self.runView(request,response,"std_trailer")
        self.setSuccessCompletionMark(action)

    def action_test_list_log_entry(self, request, response):
        params = request.params
        self.error_if_live(response)
        #id = eval(params["id"])
        self.runView(request,response,"std_header")
        response.write(self.logger.html_logentry_form(params))
        self.runView(request,response,"std_trailer")
        action = self.getActionName(request)
        self.setSuccessCompletionMark(action)

    def action_test_log_entry_retry(self, request, response):
        self.error_if_live(response)
        #id = eval(request.params["id"])
        id = request.params["id"]
        self.logger.rerun_from_key(id,self,response)
        # no completion mark!  will be added during retry!

    def action_test_upload_log_entry(self, request, response):
        self.error_if_live(response)
        response.setContentType('text/html')
        response.write("""
            <h1>Upload Log entry</h1>
            <p>Please cut-and-paste the log entry to rerun into the text box below
            and hit the RUN button</p>
            <form action="kfdserve.cgi">
            <input type="hidden" name="action" value="test_run_uploaded_log_entry">
            <textarea name="entry_text" cols=60 rows=20></textarea>
            <input type="submit" name="RUN" value="RUN">
            </form>
            """)

    def action_test_run_uploaded_log_entry(self, request, response):
        self.error_if_live(response)
        entry_text = request.params["entry_text"]
        # XXX I don't know if this is sufficient or even totally correct!!!
        entry_text = string.replace(entry_text,"&lt;","<")
        entry_text = string.replace(entry_text,"&gt;",">")
        entry_text = string.strip(entry_text)
        self.logger.rerun_from_entry_text(entry_text,self,response)
        # no completion mark!

def test():
    import apploggers, errhandlers
    from reportlab.lib.utils import get_rl_tempfile
    try:
        try:
            logFN=get_rl_tempfile('bilbo')
            l = apploggers.FileAppLogger(None,logFN=logFN)
            l.record('start',params={'a': 1, 'b':(1,2)})
            time.sleep(3.3)
            l.record('finish',msg='ended OK')
            l.store()
            del l   #throw the record away

            #create an empty one
            l = apploggers.FileAppLogger(logFN=logFN)
            l
            #get the start record
            start = l['start']
            print start
            __store__=l['__store__']
            print __store__
        except:
            print >>open('/tmp/dbg.html','w'), errhandlers.ShowTracebackHandler().handleException()
            raise
    finally:
        os.remove(logFN)

if __name__=='__main__':
    test()
