class IErrorHandler:

    """Error handlers should conform to this interface"""
    
    def handleException(self):
        """Handle the the current exception

        Return an HTML snippet which will be inserted within the body of a returned error page.
        """
        pass

import sys
import traceback
import string
from xml.sax.saxutils import escape

class Coder:

    key = 'rlErr0r'    

    def encode(self, data):
        from rlextra.utils.rc4 import encode
        import base64
        return base64.encodestring(encode(data,self.key))

    def decode(self, data):
        from rlextra.utils.rc4 import decode
        import base64
        return base64.encodestring(decode(data,self.key))
        

def printCodedTraceback(s):
    print Coder().decode(s)
    
### hack this fixes a bug in inspect for 2.3
import inspect
__code='''
def getframeinfo(frame, context=1, lineno=None):
    if istraceback(frame):
        frame = frame.tb_frame
    if not isframe(frame):
        raise TypeError('arg is not a frame or traceback object')

    filename = getsourcefile(frame) or getfile(frame)
    if lineno is None: lineno = frame.f_lineno
    if context > 0:
        start = lineno - 1 - int(divmod(context, 2)[0])
        try:
            lines, lnum = findsource(frame)
        except IOError:
            lines = index = None
        else:
            start = max(start, 1)
            start = min(start, len(lines) - context)
            lines = lines[start:start+context]
            index = lineno - 1 - start
    else:
        lines = index = None

    return (filename, lineno, frame.f_code.co_name, lines, index)

def getinnerframes(tb, context=1):
    framelist = []
    while tb:
        framelist.append((tb.tb_frame,) + getframeinfo(tb, context, tb.tb_lineno))
        tb = tb.tb_next
    return framelist
'''
exec __code in inspect.__dict__
del inspect
### end of bug fix

class ShowTracebackHandler(IErrorHandler):

    def handleException(self,debugMemo=1):
        tbt,tbv,tb = sys.exc_info()
        if sys.hexversion<0x2020000:
            html = traceback.format_exception(tbt,tbv,tb)
            html = ''.join(html)
            # Ensure goes into server error log too
            # commented out because appears to cause problems with some version of Apache
            # sys.stderr.write(errInfo)
            html = '<pre>\n%s\n</pre>' % escape(html)
        else:
            import cgitb
            html = cgitb.html((tbt,tbv,tb),context=3)
            html = html[html.find('>')+1:]
            colormap={
                '#909090':  '#606060',  #grey
                '#f0f0f8':  '#f0f0f8',  #bodybg
                '#6622aa':  '#6622aa',  #toptitle color
                '#ffffff':  '#ffffff',  #toptitle font color
                '#d8bbff':  '#d8bbff',  #filename background
                '#ffccee':  '#ffccee',  #codeline background
                }
            for old, new in [(old,new) for old,new in colormap.items() if old!=new]:
                html = html.replace(old,new)

        if debugMemo:
            from reportlab.lib.utils import DebugMemo, getStringIO
            info = getStringIO()
            dbgmemo = DebugMemo(capture_traceback=0,stdout=info)
            dbgmemo.show()
            return '%s\n<pre>\n%s</pre>' % (html,info.getvalue())
        else:
            return html

class EncryptTracebackHandler(ShowTracebackHandler):

    def handleException(self,debugMemo=1):
        html = ShowTracebackHandler.handleException(self,debugMemo=debugMemo)
        # ensure goes into server error log too
        #sys.stderr.write(errInfo)
        encodedHtml = Coder().encode(html)
        return '<!--\n%s\n-->' % encodedHtml
