"""
Utilities to get old ERS apps to run in the WebApp framework
"""

class WebAppProcessorMixIn:

    """ Mix-in class to be used with ERS Processors """

    def getErsDictFromRequest(self, app, request):
        """ Construct ERS style dictionary from WebApp request """
        import os
        from framework import context
        from webapp import include
        dictionary = request.params
        assert context(dictionary)=='html', 'ONLY SUPPORTING html AT PRESENT!'
        dictionary['OUTDIR'] = app.OUTDIR
        dictionary['HTTP_OUTDIR'] = app.HTTP_OUTDIR
        dictionary['document'] = 'NOTUSED'
        dictionary['SCRIPT_PFX'] = os.path.splitext(os.path.basename(dictionary['SCRIPT_NAME']))[0]
        if not dictionary.has_key('include'): dictionary['include'] = include
        dictionary["DESCRIPTION"] = self.description(dictionary)
        if not dictionary.has_key("outfile"):
            dictionary["outfile"] = self.getUniqueFileName(dictionary.get('SCRIPT_PFX',''),dictionary.get('currentid',None))
        if not dictionary.has_key("APPDIR"):
            # Assume current directory is application dir
            dictionary["APPDIR"] = os.getcwd()
        return dictionary


from webapp import WebApp

class ErsWebApp(WebApp):

    """ Derive from this to wrap an ERS processor. Set class variable processor to the ERS Processor to use. """ 

    processor = None # Set this in your subclasses
    
    def __init__(self,**kw):
            
        self._setKeywords(**kw)
        WebApp.__init__(
            self, 
            defaultAction = 'Test',
            #pageTemplate = 'pagetemplate.html',
            paramQuoteFunction = lambda x: x,
        )
        assert self.processor is not None, 'Must supply an ERS processor' 
        self._ersActionDir = {}
        self.addErsActions({
            'PdfText': 'Return PDF in HTTP response',
            'PdfLink': 'Make PDF and return a link to the generated file',
            'GoopHTML': 'Make HTML goop (de-flated, MD5, checksummed base64) from the data',
            'Rml': 'Translate XML to RML and return HTML escaped RML text',
            'PlainRml': 'Translate XML to RML and return unescaped RML text (not viewable in some browsers)',
        })

    def addErsActions(self, actionDir):
        self._ersActionDir.update(actionDir)

    def action_PdfText(self, request, response):
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)

        from framework import getStringIO
        import os
        fn = "%s.pdf" % os.path.splitext(os.path.basename(dictionary.get('outfile','xmlforms')))[0]
        outfile = getStringIO()
        dictionary["outfile"] = outfile #this causes RML errors unless the outfile name is
                                        #handled well in the template
        processor.makePdf(dictionary)
        pdfout = outfile.getvalue()

        response.setHeader('Content-Disposition', 'inline; filename="%s"' % fn)
        response.setContentType('application/pdf')
        response.write(pdfout)

    def action_PdfLink(self, request, response):
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)
        
        import os
        outFilename = processor.makePdf(dictionary)

        location = dictionary['HTTP_OUTDIR'] + '/' + outFilename
        dictionary["outfile"] = outFilename
        dictionary["DOWNLOAD_LOCATION"] = location
        dictionary["SUCCESS"] = 1

        import string
        L = []
        a = L.append
        a('<a href="%s.cgi">Start Over</a><br>'%dictionary['SCRIPT_PFX'])
        a('The PdfLink option creates a PDF document from the data input and returns a link to the document.<br>')
        a("<h1>Your PDF is ready...</h1>")
        a("download PDF: <a href=%s>%s</a>" % (location, outFilename))
        a("<p>%(PDF_PAGE_COUNT)d pages generated in %(PDF_GENERATION_TIME)0.4f seconds, PDF size %(PDF_FILE_SIZE)d bytes" % dictionary)

        response.write(string.join(L, "\n"))
        response.setContentType('text/html')

    def action_GoopHTML(self, request, response):
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)

        if not dictionary.has_key('datafilters'):
            dictionary['datafilters'] = 'deflate,md5,base64'
        processor.dataFilter(dictionary)

        response.write(dictionary['data'])
        response.setContentType('text/plain')

    def action_Rml(self, request, response):
        "Translate XML to RML and return HTML escaped RML text"
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)

        from xml.sax import saxutils
        rmlOut = processor.makeRml(dictionary)

        response.write('<pre>%s</pre>' % saxutils.escape(rmlOut))
        response.setContentType('text/html')

    def action_PlainRml(self, request, response):
        "Translate XML to RML and return unescaped RML text (not viewable in some browsers)"
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)

        rmlOut = processor.makeRml(dictionary)

        response.write(rmlOut)
        response.setContentType('data/x-rml')

    def action_Test(self, request, response):
        import string
        from rlextra.preppy import preppy
        processor = self.processor
        dictionary = processor.getErsDictFromRequest(self, request)
        dictionary["SAMPLES"] = processor.getSampleList(dictionary)
        sampleFilename = dictionary.get('sample', '')
        dictionary["sample"] = sampleFilename
        sampleText = ''
        if sampleFilename:
            from reportlab.lib.utils import open_and_read
            sampleText = open_and_read(dictionary["sample"].strip())
        dictionary["SAMPLETEXT"] = sampleText
        dictionary["ACTIONS"] = self._ersActionDir

        self.runView(request, response, 'ersTestPage')
        response.setContentType('text/html')
