"""Standard implementation."""
import sys, os, string, time, traceback, glob, operator
from rlextra.preppy import preppy
from rlextra.utils.cgisupport import quoteValue
from reportlab.lib.utils import getStringIO, rl_isfile, rl_glob

def context(dictionary):
	return dictionary.get('CONTEXT','html').lower()

def go(dictionary, processorClass=None):
	from rlextra.ers.webapp import include
	if not dictionary.has_key('include'): dictionary['include'] = include
	return (processorClass or BasicProcessor)().go(dictionary)

def ersPath(n=None):
	'''returns the path to the ers installation'''
	d = os.path.abspath(os.path.dirname(__file__))
	if n: d = os.path.join(d,n)
	return d

def findTemplate(template,ers=None):
	D = ['templates','text','html','.']
	ED = map(ersPath,D)
	for d in (ers and ED+D or D+ED):
		fn = os.path.normpath(os.path.join(d,template))
		for ext in ['.pyc','.py','.prep']:
			if rl_isfile(fn + ext): return d, template
	return None

def findFile(fn,D=['.'],ers=None):
	ED = map(ersPath,D)
	for d in (ers and ED+D or D+ED):
		tfn = os.path.normpath(os.path.join(d,fn))
		if rl_isfile(tfn): return tfn
	return None

def getTemplate(dictionary,template,ers=None):
	try:
		d, fn = findTemplate(template,ers)
		try:
			L = []
			sysP = sys.path
			sys.path.insert(0,d)
			preppy.getModule(template,directory=d).run(dictionary,__write__=L.append)
			sys.path = sysP
			return string.join(L,'\n')
		except:
			sys.path = sysP
			raise
	except:
		return None

def _getSampleList(*pats):
	R = filter(lambda x: rl_isfile(x),reduce(operator.add,map(lambda p: rl_glob(p),pats)))
	R.sort(lambda a,b: cmp(os.path.basename(a),os.path.basename(b)))
	return R

class RmlPageCounter:
	"Pass to rml2pdf.go and it counts pages and output PDF size"
	def __init__(self):
		self.pageCount = 0
		self.fileSize = 0
	def __call__(self, typ, value):
		if typ == 'PAGE':
			self.pageCount = value
		elif typ == 'PDF_SIZE':
			self.fileSize = value


class DataFilterProcessor:
	'''A mixin class that handles filtering of the input data'''

	def deflate_filter(self,data):
		from reportlab.lib.utils import import_zlib
		zlib = import_zlib()
		if not zlib: raise ValueError, "Can't gzip_encode, zlib not available"
		return zlib.compress(data)

	def flate_filter(self,data):
		from reportlab.lib.utils import import_zlib
		zlib = import_zlib()
		if not zlib: raise ValueError, "Can't deflate, zlib not available"
		return zlib.decompress(data)

	def base64_filter(self,data):
		from base64 import encodestring
		return encodestring(data)

	def unbase64_filter(self,data):
		from base64 import decodestring
		return decodestring(data)

	def md5_filter(self,data):
		import md5
		m = md5.new()
		m.update(data)
		return data+m.hexdigest()

	def checkmd5_filter(self,data):
		import md5
		cksum, data = data[-32:], data[:-32]
		m = md5.new()
		m.update(data)
		assert m.hexdigest()==cksum, 'md5 checksum failure'
		return data

	def _dataFilter(self,DF,data):
		for df in DF.split(','):
			data = getattr(self,df+'_filter')(data)
		return data

	def dataFilter(self,D,keepOld=None):
		DF = D.get('datafilters',None)
		if DF:
			data = D['data']
			if keepOld: D[keepOld] = data
			D['data'] = self._dataFilter(DF,data)

	def actionGoopHTML(self,D):
		'''make HTML goop (de-flated, MD5, checksummed base64) from the data'''
		stdout = sys.stdout
		html = context(D)=='html'
		if not D.has_key('datafilters'):
			D['datafilters'] = 'deflate,md5,base64'
		self. dataFilter(D)
		data = D['data']
		if html:
			stdout.write("Content-Type: text/html\r\n")
			stdout.write("Content-Length: %d\r\n" % len(data))
			stdout.write("\r\n")
		stdout.write(data)
		return ''

class BasicProcessor:
	template = "rmltemplate"
	if sys.hexversion<0x2020000:
		def getActions(self,plus=()):
			"return tuple of valid actions for this project"
			A = ["PdfLink","Pdf","PdfText","Rml","PlainRml","Rml2Pdf","Echo","getImage","Test"]
			for a in plus:
				if a not in A: A.append(a)
			A.sort()
			return tuple(A)
	else:
		def getActions(self):
			"return tuple of valid actions for this instance"
			from inspect import getmembers, ismethod
			A = map(lambda x: x[0][6:],filter(lambda x: x[0][:6]=='action',getmembers(self,ismethod)))
			A.sort()
			return tuple(A)

	def getActionsDict(self, dictionary):
		actions = self.getActions()
		result = {}
		html = context(dictionary)=='html'
		for a in actions:
			methodname = "action" + a
			method = getattr(self, methodname)
			doc = method.__doc__
			doc = doc is not None and doc or 'No documentation for this method'
			if html: doc = quoteValue(doc)
			result[a] = doc
		return result

	def getUniqueFileName(self, prefix='out',suffix=None):
		# builds one from timestamp
		if prefix and prefix[-1]!='-': prefix += '-'
		if not suffix: return '%s%0.0f.pdf' % (prefix,	1000 * time.time()) # no decimals...
		if not hasattr(self,'_fnc'): self._fnc = 0
		else: self._fnc = self._fnc+1
		return prefix+suffix+str(self._fnc)+'.pdf'

	def description(self, dictionary):
		fn = "description.txt"
		addendum = " [processor %s]" % self.__class__.__name__
		try:
			desc = open(fn).read()
		except:
			desc = "No description.txt found"
		return desc+addendum

	def go(self, dictionary):
		from rlextra.ers.webapp import include
		if not dictionary.has_key('include'): dictionary['include'] = include
		dictionary["DESCRIPTION"] = self.description(dictionary)
		if not dictionary.has_key("outfile"):
			dictionary["outfile"] = self.getUniqueFileName(dictionary.get('SCRIPT_PFX',''),dictionary.get('currentid',None))
		if not dictionary.has_key("action"):
			action = dictionary["action"] = "Test"
		else:
			action = dictionary["action"]
		if dictionary.get('LIVE',0) and action=='Test':
			raise ValueError, "Illegal action '%s'" % action
		#1. try for a method with a matching name
		if action in self.getActions():
			actionname = "action"+action
			actionmethod = getattr(self, actionname)
			return actionmethod(dictionary)
		else:
			#2. try for a preppy view
			viewName = 'views' + os.sep + action
			if (os.path.isfile(viewName + '.pyc') or
				os.path.isfile(viewName + '.prep')):
				print "Content-Type: text/html"
				print
				m = preppy.getModule(viewName + '.prep')
				print m.getOutput(dictionary)
				return ""
			else:
				raise ValueError, "Unknown action '%s'" % action
			
			
	
	def actionTest(self, dictionary):
		"acquire sample data or run sample data if available"
		assert context(dictionary)=='html'
		print "Content-Type: text/html"
		print
		if not dictionary.has_key("sample"):
			dictionary["SAMPLES"] = self.getSampleList(dictionary)
			dictionary["FORMAT"] = "dictionary"
			m = preppy.getModule("chooseSample", directory=ersPath("html"), savefile=0)
		else:
			dictionary["SAMPLEDICTIONARY"] = self.getSampleDictionary(dictionary)
			dictionary["ACTIONS"] = self.getActionsDict(dictionary)
			m = preppy.getModule("runDictionary", directory=ersPath("html"), savefile=0)
		m.run(dictionary)
		return ""

	def getSampleList(self, dictionary): 
		return _getSampleList('./sample*.py','./samples/*.py')

	def getSampleDictionary(self, dictionary):
		"return text of sample file"
		fn = dictionary["sample"]
		text = string.strip(open(fn).read())
		dict = eval(text, {}, {})
		return dict

	def actionPdfLink(self, dictionary):
		"make PDF and return a link to the generated file"
		L = []
		a = L.append
		outfile = self.makePdf(dictionary)
		if context(dictionary)=='html':
			a('Content-Type: text/html')
			a('')

			location = os.path.join(dictionary['HTTP_OUTDIR'],outfile)
			dictionary["outfile"] = outfile
			dictionary["DOWNLOAD_LOCATION"] = location
			dictionary["SUCCESS"] = 1

			# look for a template.	If found, this can display both
			# success and error details
			content = getTemplate(dictionary,'linktemplate')
			if content:
				a(content)
			else:
				a('<a href="%s.cgi">Start Over</a><br>'%dictionary['SCRIPT_PFX'])
				a('The PdfLink option creates a PDF document from the data input and returns a link to the document.<br>')
				a("<h1>Your PDF is ready...</h1>")
				a("download PDF: <a href=%s>%s</a>" % (location, outfile))
				a("<p>%(PDF_PAGE_COUNT)d pages generated in %(PDF_GENERATION_TIME)0.4f seconds, PDF size %(PDF_FILE_SIZE)d bytes" % dictionary)
				  
		else:
			location = os.path.abspath(os.path.join(dictionary['OUTDIR'],outfile))
			a(location)
		return string.join(L, "\n")

	def actionPdf(self, dictionary):
		"make PDF and return the name of the generated file"
		if context(dictionary)=='html':
			print "Content-Type: text/html"
			print
		print self.makePdf(dictionary), "generated"
		return ""
	
	def actionPdfText(self, dictionary):
		"""return PDF in HTTP response.
		   Needs a late RML2PDF and may not work
		   when delivered to some browser configurations."""
		fn = "%s.pdf" % os.path.splitext(os.path.basename(dictionary.get('outfile','xmlforms')))[0]
		outfile = getStringIO()
		stdout = sys.stdout
		dictionary["outfile"] = outfile #this causes RML errors unless the outfile name is
											#handled well in the template
		self.makePdf(dictionary)
		pdfout = outfile.getvalue()
		if context(dictionary)=='html':
			stdout.write("Content-Type: application/pdf\r\n")
			stdout.write("content-length: %d\r\n" % len(pdfout))
			stdout.write("content-disposition: inline; filename=\"%s\"\r\n" % fn)
			stdout.write("\r\n")
		stdout.write(pdfout)
		return ""

	def actiongetImage(self,dictionary):
		'''returns an image in the response
		needs document=None&action=getImage&image=iiii'''
		stdout = sys.stdout
		html = context(dictionary)=='html'
		image = dictionary.get('image',None)
		if image: image = findFile(image,D=['images','html','.'])
		if not image: raise ValueError('image not specified')
		try:
			data = open(image,'rb').read()
		except:
			raise ValueError('image %s not found cwd=%s' % (image,os.getcwd()))
		if html:
			ext = os.path.splitext(image)[1][1:]
			if not ext: ext='unknown'
			stdout.write("Content-Type: image/%s\r\n" % ext)
			stdout.write("content-length: %d\r\n" % len(data))
			stdout.write("\r\n")
		stdout.write(data)
		return ''

	def makePdf(self, dictionary):
		"make pdf from xml"
		return self.processRml(dictionary, self.makeRml(dictionary))

	def actionRml(self, dictionary):
		"translate XML to RML and return HTML escaped RML text"
		rmlout = self.makeRml(dictionary)
		if context(dictionary)=='html':
			print "Content-Type: text/html"
			print
			print "<pre>"
			print quoteValue(rmlout)
			print "</pre>"
		else:
			print rmlout
		return ""

	def actionRml2Pdf(self, dictionary):
		"from RML input convert directly to PDF"
		rmlout = dictionary["data"]
		if context(dictionary)=='html':
			print "Content-Type: text/html" # ???
			print
		return self.processRml(dictionary, rmlout)

	def actionPlainRml(self, dictionary):
		"translate XML to RML and return unescaped RML text (not viewable in some browsers)"
		rmlout = self.makeRml(dictionary)
		if context(dictionary)=='html':
			print "Content-Type: data/x-rml" # ???
			print
		print rmlout
		return ""

	def actionEcho(self, dictionary):
		"display the CGI input parameters"
		#return dummy("Echo")
		L = []
		a = L.append
		keys = dictionary.keys()
		keys.sort()
		if context(dictionary)=='html':
			a('Content-Type: text/html')
			a('')
			a('<a href="%s.cgi">Start Over</a><br>' % dictionary['SCRIPT_PFX'])
			a('The echo option simply displays the CGI parameters sent to the server and adjoined parameters.<br>')
			a("<h1>echo</h1>")
			for k in keys:
				a("<b>%s</b> =<br>" % k)
				v = str(dictionary[k])
				v = string.replace(v, "&", "&amp;")
				v = string.replace(v, "<", "&lt;")
				v = string.replace(v, ">", "&gt;")
				a("<blockquote><pre>%s</pre></blockquote><br>" % v)
			a('<a href="%s.cgi">Start Over</a><br>' % dictionary['SCRIPT_PFX'])
		else:
			for k in keys:
				a("%s =" % k)
				a(str(dictionary[k]))
		return string.join(L,"\n")

	def getTemplateInfo(self, dictionary):
		return findTemplate(dictionary.get("template", self.template))

	def makeRml(self, dictionary):
		(templatedir, templatename) = self.getTemplateInfo(dictionary)
		# add to head of search path so import will find this one if template is absent
		sys.path.insert(0, templatedir)
		outfile = getStringIO()
		m = preppy.getModule(templatename, directory=templatedir, savefile=0)
		m.run(dictionary, outputfile=outfile)
		rmlout = outfile.getvalue()
		return rmlout

	def processRml(self, dictionary, rmlout):
		# try to suppress irritating warnings
		started = time.clock()
		try:
			from reportlab import rl_config
			rl_config.PIL_WARNINGS=0
		except:
			pass
		outfile = dictionary["outfile"]
		from rlextra.rml2pdf.rml2pdf import go
		outDir = dictionary.get('OUTDIR','')
		if outDir and not os.path.isdir(outDir):
			raise ValueError('No such directory "%s"' % outDir )
		if dictionary.get('SAVE_LASTEST_RML',None): open('lastest.rml','w').write(rmlout)
		mon = RmlPageCounter()
		go(rmlout, outDir=outDir, outputFileName=outfile, progressCallBack=mon)
		dictionary['PDF_PAGE_COUNT'] = mon.pageCount
		finished = time.clock()
		dictionary['PDF_GENERATION_TIME'] = finished - started
		dictionary['PDF_FILE_SIZE'] = mon.fileSize  # to be completed
		return outfile

class BasicXMLProcessor(DataFilterProcessor,BasicProcessor):
	"""This looks for an attribute called 'data'
	and parses it using the MasterDetail processor"""

	def _parser(self,xmltext):
		from reportlab.lib.rparsexml import parsexml
		return parsexml(xmltext)

	if sys.hexversion<0x2020000:
		def getActions(self,plus=()):
			"return tuple of valid actions for this project"
			return BasicProcessor.getActions(self,plus=("Pprint","GoopHTML")+plus)

	def parseXml(self, dictionary):
		self.dataFilter(dictionary)
		basic = self._parser(dictionary["data"])
		firsttag = basic[2][0]
		D = masterDetail(firsttag)
		for k, v in dictionary.items():
			if not D.has_key(k): D[k] = v
		return D

	def actionPprint(self, dictionary):
		"translate data to internal parse format and return pprint of the parsed format"
		import pprint
		parsed = self.parseXml(dictionary)
		if context(dictionary)=='html':
			print "Content-Type: text/html"
			print
			print "<h1>pprint of parsed representation</h1><pre>"
		pprint.pprint(parsed)
		if context(dictionary)=='html':
			print "</pre>"
		return ""
	
	def actionTest(self, dictionary):
		"acquire sample data or run sample data if available"
		assert context(dictionary)=='html'
		print "Content-Type: text/html"
		print
		dictionary["SAMPLES"] = self.getSampleList(dictionary)
		if not dictionary.has_key("sample"):
			dictionary["FORMAT"] = "XML"
			m = preppy.getModule("chooseSample", directory=ersPath("html"), savefile=0)
		else:
			dictionary["SAMPLETEXT"] = self.getSampleText(dictionary)
			dictionary["ACTIONS"] = self.getActionsDict(dictionary)
			m = preppy.getModule("runSample", directory=ersPath("html"), savefile=0)
		m.run(dictionary)
		return ""

	def getSampleList(self, dictionary):
		"return a list of sample names for this instance"  
		return _getSampleList('./sample*.xml','./samples/*.xml')

	def getSampleText(self, dictionary):
		"return text of sample file"
		return open(string.strip(dictionary["sample"])).read()

	def makeRml(self, dictionary):
		parsedxml = self.parseXml(dictionary)
		(templatedir, templatename) = self.getTemplateInfo(dictionary)
		# add to head of search path so import will find this one if template is absent
		sys.path.insert(0, templatedir)
		outfile = getStringIO()
		m = preppy.getModule(templatename, directory=templatedir, savefile=0)
		m.run(parsedxml, outputfile=outfile)
		rmlout = outfile.getvalue()
		#open('/tmp/latest.rml','w').write(rmlout)
		return rmlout
	
class BasicNonValidatingXMLProcessor(BasicXMLProcessor):
	"""This looks for an attribute called 'data'
	and parses it using the MasterDetail processor"""
	def _parser(self,xmltext):
		from reportlab.lib.rparsexml import parsexmlSimple
		return parsexmlSimple(xmltext)

def dummy(name):
	"dummy method implementation for testing"
	print "Content-Type: text/html"
	print
	print repr(name)
	return ""

class GenericRecord:
	def __init__(self, dict):
		for k in dict.keys():
			setattr(self, k, dict[k])
	def __repr__(self):
		L = ["GenericRecord("]
		d = self.__dict__
		for x in d.keys():
			L.append("%s=%s, " %(x,d[x]))
		L.append(")")
		return string.join(L, "")

def masterDetail(parsedTag):
	(tagName, attributes, contentList, misc) = parsedTag
	# basic results are the attributes of the tag
	if attributes:
		result = attributes.copy()
	else:
		result = {}
	# set content elements
	import types
	if contentList is not None:
		for c in contentList:
			if type(c) is types.StringType:
				cs = string.strip(c)
				if cs:
					raise ValueError, "no non-white string content permitted in masterDetail format %s" % repr(cs[:10])
			elif c is not None:
				(tagName1, attributes1, contentList1, misc1) = c
				taglistname = "LIST_OF_"+tagName1
				if result.has_key(taglistname):
					taglist = result[taglistname]
				else:
					taglist = result[taglistname] = []
				cdict = masterDetail(c)
				crec = GenericRecord(cdict)
				taglist.append(crec)
	return result

class WrappedXMLProcessor(BasicXMLProcessor):
	"""This looks for an attribute called 'data', parses it and puts a TagWrapper around it"""

	def parseXml(self, dictionary):
		from rlextra.radxml.xmlutils import escapeNakedAmpersands, xml2doctree
		shouldValidate = (dictionary.get("validating", "yes")=='yes')
		result = dictionary.copy()
		self.dataFilter(dictionary)
		result["data"] = xml2doctree(escapeNakedAmpersands(dictionary["data"]), validating=shouldValidate)
		return result

class TemplateSelectingWrappedXMLProcessor(WrappedXMLProcessor):
	"""Checks top level tags for a template tag"""
	templatePrefix = 'rmltemplate'
	def parseXml(self, D):
		R = WrappedXMLProcessor.parseXml(self,D)
		suffix = R['data']._value('template',None)
		self.template = self.templatePrefix+(suffix and ('_'+suffix) or '')
		return R

class ParsedXMLProcessor(BasicXMLProcessor):
	"""This looks for an attribute called 'data' and puts a TagWrapper around it"""

	def parseXml(self, dictionary):
		from rlextra.radxml.xmlutils import TagWrapper
		result = dictionary.copy()
		self.dataFilter(dictionary)
		result["data"] = TagWrapper(dictionary["data"])
		return result

class HTMLProcessor(WrappedXMLProcessor):
	"""Same as WrappedXMLprocessor but can do HTML forms too, given a template"""
	def actionHTML(self, dictionary):
		"make HTML"
		L = []
		a = L.append
		if context(dictionary)=='html':
			a('Content-Type: text/html')
			a('')
		dictionary = self.parseXml(dictionary)
		from rlextra.preppy import preppy
		m = preppy.getModule('htmltemplate')
		m.run(dictionary, __write__=a)
		return string.join(L, "\n")

	if sys.hexversion<0x2020000:
		def actions(self):
			"return tuple of valid actions for this project"
			return WrappedXMLProcessor.actions(self)+('HTML',)

if __name__=='__main__': #noruntests
	raise NotImplementedError("%s is not an application!" % os.path.basename(sys.argv[0]))
