"""This is a 'mix-in' for WebApp which provides
a basic user authentication and timeout mechanism.

This is used as a dynamic mix-in within a WebApp,
so that WebApps acquire all its attributes and methods
if authentication is needed.

By default is will use a database connection and
look for userName/ecnryptedPassword fields in
a given table to do the initial validation of a login.
Thereafter, it generates a token embedding the
uservarious facts including the machine it was generated
on and the time generated, which can be used for
subsequent ones.

You can override the _valid(user, password) method
to do any other authenticaion scheme you want.

The main 'authenticate' method takes a WebRequest object
and looks in its params dictionary for either (a)
a username/password pair, or (b) a token.  If these
are found and pass, it proceeds.  If authentication fails,
it will 'bounce' to a login screen by setting the 'action'
parameter in the request to 'showLogin', but will store the
action you were trying to get to, so that after logging
in you cxan resume.  Example: you were midway through
editing a big form and went for coffee and timed out.
You will have to re-log-in but after that your
new record should be saved.



"""

import base64, cPickle, md5
#from __main__ import _DBG

def _getRotor():
    import rotor
    return rotor.newrotor('bingo-bongo')

def _get_remote_addr():
    from os import environ
    return environ.get('REMOTE_ADDR','[no-remote-addr]').replace('-','_')

class GateKeeper:
    timeout = 600
    userCol = 'loginId'
    loginTable = 'LoginTable'
    encryptedPasswordCol = 'encryptedPassword'
    _uid = ''
    
    def authenticate(self, request):
        #_dbg = open('/tmp/dbg.txt','w')
        if not getattr(self,'needAuthentication',None): return
        params = request.params
        token = params.get('token',None)
        if token and self._check_token(token):
            params['token'] = self._make_token()
            return
        if token: del params['token']
        #print >> _dbg, 'gatekeeper.validate -- no valid token'
        lsp = params.get('login_saved_params',None)
        if lsp:
            token = self._get_token(request)
            lsp = cPickle.loads(base64.decodestring(lsp))
            del params['login_saved_params']
            #print >>_dbg,'<!-- params',params,'-->'
            #print >>_dbg,'<!-- lsp',lsp,'-->'
            #print >>_dbg,'<!--self.__dict__',self.__dict__,'-->'
            OUSER = lsp.get('USER','') or self._uid
            if token and self._uid!=OUSER:
                self._uid = ''
                params['action'] = 'error'
                params['ERROR_MESSAGE'] = 'You\'re not '+OUSER
                return
            #print >> _dbg, 'gatekeeper.validate -- lsp:', lsp
            try:
                del params['action']
            except:
                pass
            params.update(lsp)
            if token:
                params['token'] = token
                try:
                    del params['login_count']
                except:
                    pass
                return
            params['login_count'] = 1+params.get('login_count',0)
            if params['login_count']>=3:
                params['action'] = 'tooManyLogins'
                return
        else:
            params['login_count'] = 0
        params['USER'] = self._uid
        params['login_saved_params'] = base64.encodestring(cPickle.dumps(params,1))[:-1]
        params['action'] = 'showLogin'
        return

    def _get_token(self,request):
        #check
        try:
            user = request.params['USER']
            password = request.params['PASSWORD']
            del request.params['USER']
        except KeyError:
            return None
        if not self._valid(user,password): return None
        #generate my token
        return self._make_token()

    def _make_token(self, timeout=None):
        from time import time
        from random import random
        timeout = timeout or self.timeout
        token = base64.encodestring(_getRotor().encrypt('%s-%s-%s-%s-%s' % (_get_remote_addr(),time(),random(),timeout,self._uid)))
        return 'aaa-'+token.replace('\n','')

    def _check_token(self,token):
        if token[:4]!='aaa-': return 0
        token = _getRotor().decrypt(base64.decodestring(token[4:])).split('-')
        if len(token)!=5: return 0
        if _get_remote_addr()!=token[0]: return 0
        try:
            t = float(token[1])
            timeout = float(token[3])
        except:
            return 0
        from time import time
        self._uid = token[4]
        return (time()-t)<=timeout

    def _valid(self,user,password):
        hashPass = md5.md5(password).hexdigest()
        sql = "select %s from %s where %s='%s' and %s='%s'" % (self.userCol,self.loginTable,self.userCol,user,self.encryptedPasswordCol,hashPass)
        result = self.query(sql)
        if result:
            self._uid = result[0][0]
            return 1
        return 0


class DummyGateKeeper(GateKeeper):
    """Subclass for unit testing.  No database connection,
    supports the same few users"""
    passwords = {'alice':'alice',
                 'bob':'bob',
                 'charlie':'charlie',
                 'danny':'danny'
                 }

    def _valid(self, user, password):
        pwd = self.passwords.get(user, None)
        if pwd == password:
            return 1
        else:
            return 0
        
