#copyright ReportLab Inc. 2001
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/graphics/guiedit/guiedit.py
#$Header: /rl_home/xxx/repository/rlextra/graphics/guiedit/guiedit.py,v 1.75 2003/11/22 14:36:20 robin Exp $
__version__=''' $Id: guiedit.py 14580 2004-09-09 11:07:45Z rgbecker $ '''
'''The GUI editor main script'''
import sys, os, Tkinter, re, tkFileDialog, tkSimpleDialog, tkMessageBox, string
import traceback, pprint, tkFont
import rlextra.graphics.guiedit.mutils as mutils
from guidialogs import AttrListBox, NameDialog, EditListDialog, upColor, RL_ScrolledText, RL_MenuBar, PMDisplay
from types import StringType, FloatType, InstanceType
from time import time
from reportlab.lib.utils import getStringIO
from codeop import compile_command

_MAIN_METHODS={'go':'()','save':"(formats=['pdf'],outDir='.',fnRoot=None)"}

# Application information
title = 'ReportLab Drawing Editor'
import Pmw
Pmw.aboutversion(string.split('$Revision: 1.75 $')[1])
Pmw.aboutcopyright('Copyright ReportLab Inc. 2001\nAll rights reserved')
Pmw.aboutcontact('''Contact: Robin Becker
Phone: +44 208 545 7271
email: robin@reportlab.com
'''
)

root = None
config=None
_MAGIC='#Autogenerated by ReportLab guiedit do not edit'
_escLF=re.compile(r'\\\n',re.MULTILINE)	#re to find escaped line ends
_classNamePat=re.compile(r'^class\s+([^\(\s]+)\s*\(\s*_DrawingEditorMixin\s*,\s*([^\)\s]+)\s*\)\s*:')
_excNamePat=re.compile(r"^\s*(?:global\s*|)name '(?P<name>[a-zA-Z_]\w*)' is not defined$")
_initPat=re.compile(r"^\tdef __init__\(self,width=([^,]+),height=([^,]+),\*args,\*\*kw\):$")
_getContentsPat=re.compile(r"^\tdef getContents\(self\):$")
_bnPat=re.compile(r"_[0-9]+$")
_ifmainPat = re.compile('^if\\s+(__name__\\s*==\\s*([\'\"])__main__[\"\']|([\'\"])__main__[\'\"]\\s*==\\s*__name__)\s*:')
_geom_re=re.compile(r'^(\d+)x(\d+)([+-]\d+)([+-]\d+)',re.I)
_testAttrFilter_re=re.compile(r'^self\s*\.\s*test\s*=\s*\d+\s*$')
editor=None
_known=None
_shapes={'Circle': '(20,20,10)',
 		'Ellipse': '(0,0,5,10)',
		'Group': '()',
		'Line': '(0,0,10,10)',
		'Path': '()',
		'PolyLine': '()',
		'Polygon': '()',
		'Rect':'(10,10,30,40)',
		'String':'(10,10,\'text\')',
		'Wedge':'(20,20,10,85)',
		}

def _hasattr(obj,name):
	try:
		eval('obj.'+name,locals())
		return 1
	except:
		return 0

def _save(obj,rootName,f):
	for k,v in obj.__dict__.items():
		if hasattr(v,'__dict__'):
			_save(v,rootName+k+'.',f)
		else:
			f.write('%s = %s\n' % (rootName+k,repr(v)))

class SubConfig:
	def __init__(self,**kw):
		self.__dict__.update(kw)

class Config:
	def __init__(self,path,**kw):
		self._path = path
		self.__dict__.update(kw)
		self.restore()

	def restore(self,path=None):
		if path is None: path = self._path
		if os.path.isfile(path):
			for l in open(path,'r').readlines():
				i = l.find('=')
				if i>=0:
					name = l[:i].strip()
					if _hasattr(self,name):
						try:
							old = self.getDeepAttr(name)
							exec 'self.'+l
							if type(old) is type({}):
								new = self.getDeepAttr(name)
								if type(new) is type({}):
									for k in old.keys():
										if not new.has_key(k): new[k] = old[k]
						except:
							traceback.print_exc()
							print >> sys.stderr, 'failed to restore config.%s'%name

	def getDeepAttr(self,name):
		P = string.split(name,'.')
		o = self
		for p in P:
			o = getattr(o,string.strip(p),None)
			if o is None: break
		return o
		

	def save(self,path=None):
		opath = self._path
		del self._path
		if path is None: path = opath
		_save(self,'',open(path,'w'))
		self._path = opath

def _collectGeom(widget,name):
	if config:
		config.popupGeom[name] = widget.winfo_geometry()
		#if config.debug: writeLog('debug',"%s %s %s %s" % (widget, name, widget.geometry(), widget.winfo_geometry()))

def _savedGeom(name,pos=None):
	geom = None
	if config:
		if config.popupGeom.has_key(name):
			geom = config.popupGeom[name]
	if pos:
		if not geom: return pos
		m = _geom_re.match(geom)
		if m:
			geom = "%sx%s%s" % (m.group(1),m.group(2), pos)
		else:
			geom = pos
	return geom

def _popup(d,cbn,pos=None):
	d['deactivatecommand']=lambda d=d, cbn=cbn:_collectGeom(d,cbn)
	try:
		geom = _savedGeom(cbn,pos)
		if geom:
			r = d.activate(geometry=geom)
		else:
			r = d.activate()
	except:
		exceptionLog()
		r = None
	return r

def getBasesStr(C):
	B = []
	for b in C.__bases__:
		B.append(b.__name__)
	return '('+string.join(B,',')+')'

def obj2str(obj):
	"""Used to format object values for the attribute list"""
	t = type(obj)
	if t is FloatType: return str(obj)
	if t is InstanceType:
		if _known:
			i = id(obj)
			if _known.varsI.has_key(i):
				return _known.varsI[i][0][0]

		if hasattr(obj,'__repr__'):
			return repr(obj)
		elif hasattr(obj,'__str__'):
			return str(obj)
		else:
			C = obj.__class__
			n = C.__name__
			if n=='TypedPropertyCollection' and hasattr(obj,'_value'):
				C = obj._value.__class__
				return '%s(%s%s)' % (n,C.__name__,getBasesStr(C))
			else:
				return '%s%s' % (n,getBasesStr(C))
	return repr(obj)

def _checkName(n):
	try:
		exec n.strip() + '=None'
		return 1
	except:
		return 0

def _findLHS(e):
	j = 0
	while 1:
		i = e.find('=',j)
		if i<0: return None
		lhs = e[:i].strip()
		try:
			compile_command(lhs,'<test>','eval')
			return lhs
		except:
			j = i+1

def _stripSub(e):
	'''remove a leading []. and return the LHS part'''
	j = 0
	while 1:
		i = e.find('].',j)
		if i<0: return None
		j = i+2
		lhs = _findLHS(e[j:])
		if lhs: return e[1:i], lhs

def normalizedReadLines(f):
	'''reads from f and returns a list of unescaped lines (without LFs)'''
	if type(f) is StringType: text = open( f,'r').read()
	else: text = f.read()
	return string.split(_escLF.sub('',text),'\n')

def nullMethod(*args,**kw):
	pass

_initLogText=[]
def writeLog(kind,message,flush=0):
	if editor and editor.log:
		editor.log.write(message,kind=kind,flush=flush)
	else:
		_initLogText.append((kind,message))
		if config:
			n = len(_initLogText)-config.log.maxlines
			if n>0:	del _initLogText[:n]

def exceptionLog(txt=None):
	try:
		S = getStringIO()
		if txt:
			x,v,tb = sys.exc_info()
			n,l,f,t = traceback.extract_tb(tb)[-1]
			del tb
			T = string.split(txt,'\n')
			l -= 1
			txt = (n=='<string>' and l>=0 and l<len(T)) and T[l]+'\n' or ''
			writeLog('warn',"%s%s" % (txt, v))
		traceback.print_exc(None,S)
		writeLog('debug',S.getvalue(),flush=1)
		if not editor or not editor.log:
			pass
		if root and config.errorBells: root.bell()
		S.close()
	except:
		pass

class Project:
	def __init__(self,notify=nullMethod):
		self.notify = notify
		self.clear()
		self._badModules=[]

	def clear(self):
		self._modified = 0
		self.className = self.path = self.baseName = self.sample = None
		self.edits = []
		self.dedits = []
		self.imports = []
		self.froms = {}
		self.medits = []
		self._sys_argv = None
		self._txt = None
		self.notify(self,'clear')

	def _addBadModule(self,moduleName):
		if moduleName not in self._badModules:
			self._badModules.append(moduleName)

	def _checkBadModule(self):
		if _known:
			_known.reloadModule(self._badModules)
			self._badModules = []

	def new(self,className,baseName,moduleName):
		self.clear()
		self._checkBadModule()
		self.className = className
		self.baseName = baseName
		self.addFrom(moduleName,baseName)
		try:
			D = mutils.getModuleClassMethodDefaults(moduleName,baseName,'__init__',
				width=400, height=200,
				)
			self._baseWidth, self._baseHeight = D['width'], D['height']
		except:
			self._addBadModule(moduleName)
			writeLog('warn',"Can't determine class __init__ args: %s %s"% sys.exc_info()[:2])
			self.clear
			return
		try:
			self.notify(self,'redraw')
		except:
			t, m = sys.exc_info()[:2]
			writeLog('warn',"Error in initial draw: %s %s" % (t,m))
			self.clear()
			self._addBadModule(moduleName)
			L = traceback.extract_tb(sys.exc_traceback)
			L.reverse()
			for tb in L:
				n = mutils._normpath(tb[0])	# the filename
				if n=='<string>': break
				if _known.checkPath(n):
					#we know the file probably
					for m in _known.modules:
						if mutils._normpath(m.__file__)==n:
							self._addBadModule(m)
							break
			raise

	def save(self):
		if self._modified:
			if not self.path: self.saveAs()
			else: self.writeout(self.path)

	def saveAs(self):
		fn=self.path and os.path.basename(self.path) or ''
		iDir = self.path and os.path.dirname(self.path) or config.prevDirs.saveAs or ''
		path = tkFileDialog.asksaveasfilename(initialfile=fn,initialdir=iDir,filetypes=[('Python files','*.py')])
		if path=='': return
		if path[-3:]!='.py' and os.path.splitext(os.path.basename(path))[-1]=='':
			path += '.py'
		if not os.path.isabs(path): path = os.path.join(os.getcwd(),path)
		config.prevDirs.saveAs = os.path.dirname(path)
		self.path = path
		self.writeout(path)
		self.notify(self,'redraw')

	def exportAs(self):
		D = getattr(self,'sample',None)
		if not D: return
		fn = self.path and os.path.splitext(os.path.basename(self.path))[0] or ''
		iDir = self.path and os.path.dirname(self.path) or config.prevDirs.export or ''
		path = tkFileDialog.asksaveasfilename(initialdir=iDir,initialfile=fn,defaultextension='pdf',filetypes=[
			('EPS','*.eps'),
			('PS','*.ps'),
			('GIF','*.gif'),
			('JPEG','*.jpg'),
			('PDF','*.pdf'),
			('PNG','*.png'),
			('BMP','*.bmp'),
			('Python','*.py'),
			('TIFF','*.tif'),
			])
		if path=='': return
		if not os.path.isabs(path): path = os.path.join(os.getcwd(),path)
		config.prevDirs.export = os.path.dirname(path)
		try:
			ext = os.path.splitext(path)[1][1:]
			if not ext:
				tkMessageBox.showerror('Export',"Export Failed!\nPlease add an extension!")
				return
			D.save(fnRoot=path,formats=[ext],verbose=0)
		except:
			tkMessageBox.showerror('Export',"Export Failed")
			exceptionLog()

	def open(self,path=None):
		if not self._checkModified('Open'): return
		if not path:
			path = tkFileDialog.askopenfilename(initialdir=config.prevDirs.open or '',
					filetypes=[('Python files','*.py')])
		if not path: return
		self.clear()
		text = normalizedReadLines(path)
		if text[0][:len(_MAGIC)]!=_MAGIC:
			raise ValueError, "This file wasn't created by ReportLab guiedit"
		try:
			if not os.path.isabs(path): path = os.path.join(os.getcwd(),path)
			self.path = path
			self.findImports(text)
			self.findNamesAndEdits(text)
			self.notify(self,'redraw')
			config.prevDirs.open = os.path.dirname(path)
		except:
			exceptionLog(self._txt)
			self.clear()

	def writeout(self,path):
		t = self.buildText(filter=self._writeFilter)
		open(path,'w').write(t)
		self._modified = 0
		self.notify(self,'modulecheck')

	def _checkModified(self,title):
		return (not self._modified) or tkMessageBox.askyesno(title,"Ignore unsaved changes?")

	def addImport(self,I):
		imports = self.imports
		if type(I) is StringType: I = [I]
		for i in I:
			if i and i not in imports: imports.append(i)

	def addFrom(self,n,I):
		froms = self.froms
		if type(I) is StringType: I = [I]
		if not froms.has_key(n): froms[n] = I
		else:
			F = froms[n]
			for i in I:
				if i and i not in F: F.append(i)

	def delFroms(self,addedFroms):
		'''remove all froms matching pairs (I,name)'''
		froms = self.froms
		imports = self.imports
		for I,name in addedFroms:
			try:
				if name:
					froms[I].remove(name)
					if froms[I]==[]: del froms[I]
				else:
					imports.remove(I)
			except:
				pass

	def findImports(self,text):
		'''scan a bunch of lines looking for import and from name import statements'''
		for t in text:
			s = string.split(t)
			if len(s)<=1: continue
			if s[0]=='import':
				self.addImport(string.split(string.join(s[1:],''),','))
			elif s[0]=='from':
				assert s[2]=='import', "Corrupt from statement in python source\n'%s'" % t
				self.addFrom(s[1],string.split(string.join(s[3:],''),','))

	def findNamesAndEdits(self,text):
		state = 0
		for t in text:
			if string.strip(t)=='': continue
			if state ==0 and t[:5]=='class':
				m = _classNamePat.match(t)
				if not m:
					raise ValueError, 'Can\'t find the class name'
				g = m.groups()
				self.className = g[0]
				self.baseName = g[1]
				state = 1
			elif state==1:
				m = _initPat.match(t)
				if m:
					self._baseWidth = m.group(1)
					self._baseHeight = m.group(2)
					state = 2
					m = '\t\tapply(%s.__init__,(self,width,height)+args,kw)' % self.baseName
			elif state==2:
				if m==t:
					state = 3
					edits = self.edits
				else:
					raise ValueError, 'after def __init__ unexpected "%s"' % t
			elif state in [3,6]:
				if t[:2]=='\t\t':
					if state!=6 or not rcpat.match(t):
						edits.append(string.strip(t))
				else:
					if _ifmainPat.match(t):
						state = 4
						m = 0
						edits = self.medits
					elif state==3 and _getContentsPat.match(t):
						state = 6
						rcpat=re.compile(r'^\t\treturn\s+apply\(\s*%s\s*.\s*getContents\s*,\s*\(\s*self\s*,\s*\)\s*\)\s*$'%self.baseName)
						edits = self.dedits
					else:
						raise ValueError, 'Corrupt edit "%s"' % t
			elif state==4:
				i = t[0]=='\t'
				if i:
					t = t[1:]
					if t[0] in [' ','\t']: i = 0
				if not i:
					raise ValueError, 'Indentation error in __main__ section "%s"' % t
				if t[0]!='#': m = m + 1
				edits.append(string.strip(t))

		if state not in [3,4,6]:
			raise ValueError, 'File corrupted'
		if state == 4 and m==0:
			raise ValueError, 'Zero length __main__ section'

	def has_method(self,name):
		for k, v in self.froms.items():
			if self.baseName in v:
				return mutils.getModuleClassMethod(k,self.baseName,name) is not None

	def _writeFilter(self,text):
		if config.testAttrFilter and _testAttrFilter_re.match(text):
			from rlextra.graphics.guiedit.datacharts import DataAwareDrawing
			if issubclass(self.sample.__class__,DataAwareDrawing): return 1
		return 0

	def buildText(self,annotate=None,filter=lambda text: 0):
		self._txt = None
		L = []
		L.append('%s'%_MAGIC)
		if self.imports: L.append('import %s'% string.join(self.imports,', '))
		self.addFrom('reportlab.graphics.shapes','_DrawingEditorMixin')
		for n, i in self.froms.items():
			L.append('from %s import %s'% (n,string.join(i,', ')))
		L.append('')
		L.append('class %s(_DrawingEditorMixin,%s):' % (self.className,self.baseName))
		if annotate is not None:
			n = len(L)
			j1 = 6+len(self.className)
			j2 = j1 + len('(_DrawingEditorMixin,')
			annotate.append(('classname',("%d.%d" % (n,6), "%d.%d" % (n,j1))))
			annotate.append(('basename',("%d.%d" % (n,j2), "%d.%d" % (n,j2+len(self.baseName)))))
		L.append('\tdef __init__(self,width=%s,height=%s,*args,**kw):' % (self._baseWidth,self._baseHeight))
		L.append('\t\tapply(%s.__init__,(self,width,height)+args,kw)' % self.baseName)
		for e in self.edits:
			if filter(e): continue
			L.append('\t\t%s' % e )
			if annotate is not None:
				n = len(L)
				annotate.append(('edit',("%d.%d"%(n,2), "%d.end"%n)))
		dedits = self.dedits
		if dedits:
			L.append('')
			L.append('\tdef getContents(self):')
			for e in dedits:
				if filter(e): continue
				L.append('\t\t%s' % e )
				if annotate is not None:
					n = len(L)
					annotate.append(('dedit',("%d.%d"%(n,2), "%d.end"%n)))
			L.append('\t\treturn apply(%s.getContents,(self,))' % self.baseName)
		if self.medits==[]:
			for n, e in _MAIN_METHODS.items():
				if self.has_method(n):
					self.medits = ['%s().%s%s' % (self.className, n, e)] 
					self._modified = 1
					break
		if self.medits!=[]:
			L.append('')
			L.append('if __name__=="__main__": #NORUNTESTS')
			for e in self.medits:
				L.append('\t%s' % e)
				if annotate is not None:
					n = len(L)
					annotate.append(('medit',("%d.%d"%(n,1), "%d.end"%n)))
		self._txt = string.join(L,'\n')
		return self._txt

	def getSample(self):
		txt = self.buildText()
		try:
			if self.path:
				import __main__
				__main__._guiedit_sample_path = self.path
				cwd = os.getcwd()
				D = os.path.normpath(os.path.dirname(self.path))
				xD = D not in sys.path
				if xD: sys.path.insert(0,D)
				os.chdir(D)
			if self._sys_argv:
				oargv = sys.argv
				sys.argv = self._sys_argv
			exec "%s\n_x=%s()\n" % (txt,self.className) in locals()
		finally:
			if self.path:
				del __main__._guiedit_sample_path
				os.chdir(cwd)
				if xD: sys.path.remove(D)
			if self._sys_argv:
				sys.argv = oargv
		self.sample = _x

def _getClassList(filt):
	C = filter(filt,_known.classObjs)
	r = map(lambda x: (x.__name__,x),C)
	r.sort()
	d = {}
	for n,c in r:
		d[n] = d.has_key(n)
	N = []
	C = []
	for n,c in r:
		if d[n]: n = c.__module__+'.'+n
		N.append(n)
		C.append(c)
	return N, C

class GUIEditor:
	def __init__(self, top):
		self._attrStack = []
		# Create the Balloon.
		class GUIEditorBalloon(Pmw.Balloon):
			def _showBalloon(self, widget, help, isItem):
				Pmw.Balloon._showBalloon(self, widget, help, isItem)
				self.after(config.root.balloonDelay,self._leave,None)

		balloon = self._balloon = GUIEditorBalloon(top)

		self.project = Project(self.redraw)
		self._state = self.log = self.main = None
		self._editMode = 0

		menu = RL_MenuBar(top,
				hull_relief = 'raised',
				hull_borderwidth = 1,
				balloon = balloon)
		self.menu = menu

		# Add some buttons to the MenuBar.
		menu.addmenu('File', 'File related Actions')
		menu.addmenuitem('File', 'command', 'Close project',
				command = self.closeProject,
				label = 'Close')
		menu.addmenuitem('File', 'command', 'Create new project',
				command = self.newProject,
				label = 'New')
		menu.addmenuitem('File', 'command', 'Open existing project',
				command = self.projectOpen,
				label = 'Open')
		menu.addmenuitem('File', 'command', 'Save project',
				command = self.projectSave,
				label = 'Save')
		menu.addmenuitem('File', 'command', 'Save under new name',
				command = self.projectSaveAs,
				label = 'SaveAs')
		menu.addmenuitem('File', 'command', 'Export under new name',
				command = self.project.exportAs,
				label = 'Export')
		menu.addmenuitem('File', 'separator')
		menu.addmenuitem('File', 'command', 'Exit the application',
				command = self.exit,
				label = 'Exit')
		self._setupRecentFiles()

		menu.addmenu('Options', 'Set user preferences')
		menu.addFlipButton('Options', 'Edit getContents on/off',
				label = 'Edit getContents',
				value = self._editMode,
				command = self._editModeToggle,
				)
		menu.addFlipButton('Options', 'Private Attributes on/off',
				label = 'Private Attributes',
				value = config.privateAttrs,
				command = self._privateAttrsToggle,
				)
		menu.addFlipButton('Options', 'test attr. filter on/off',
				label = 'Filter test Attr',
				value = config.testAttrFilter,
				command = self._testAttrFilterToggle,
				)
		menu.addmenuitem('Options', 'separator')
		menu.addFlipButton('Options', 'Error Bells on/off',
				label = 'Error Bells',
				value = config.errorBells,
				command = self._errorBellsToggle,
				)
		menu.addFlipButton('Options', 'Debug on/off',
				label = 'DEBUG',
				value = config.debug,
				command = self._debugToggle,
				)
		menu.addFlipButton('Options', 'Button3 Binding',
				label = 'Button3',
				value = config.button3,
				command = self._button3,
				)
		menu.addFlipButton('Options', 'Tooltips on/off',
				label = 'Tooltips',
				value = config.root.balloon,
				command = self._balloonToggle,
				)
		menu.addmenuitem('Options', 'separator')
		menu.addmenuitem('Options', 'command', 'Searched Packages',
				command = self.editSearch,
				label = 'Searched Packages')
		menu.addmenuitem('Options', 'command', 'Package Exclusion Patterns',
				command = self.editExclude,
				label = 'Package Exclusions')

		menu.addmenu('Actions', 'Editor Actions')
		menu.addmenuitem('Actions', 'command', 'Add Widget',
				command = self.addWidget,
				label = 'Add Widget')
		menu.addmenuitem('Actions', 'command', 'Reload Modules',
				command = self.reloadModules,
				label = 'Reload Modules')
		menu.addmenuitem('Actions', 'command', 'Set Sample sys.argv',
				command = self._set_sample_sys_argv,
				label = 'Set Sample sys.argv')

		menu.addmenu('Windows', 'Window Actions')
		menu.addFlipButton('Windows', 'Log on/off',
				label = 'Log Window',
				value = config.log.on,
				command = self._logToggle,
				)
		menu.addFlipButton('Windows', 'Src on/off',
				label = 'Src Window',
				value = config.src.on,
				command = self._srcToggle,
				)
		menu.addmenuitem('Windows', 'separator')
		menu.addmenuitem('Windows', 'command', 'Clear Log Window',
				command = self.clearLog,
				label = 'Clear Log Window')
		menu.addmenuitem('Windows', 'separator')
		menu.addmenuitem('Windows', 'command', 'Zoom Set',
				command = self.zoomSet,
				label = 'Zoom Set')

		menu.addmenu('Help', 'User manuals')#, side = 'right')

		self._about = Pmw.AboutDialog(top, applicationname = title)
		self._about.withdraw()
		astr = 'About '+title
		menu.addmenuitem('Help', 'command', astr, command=self._about.show, label=astr)

		alm = self.attrListMenu = Tkinter.Menu(top,tearoff=0)
		alm.add_command(label="Up",command=self.attrListContract,foreground=upColor)
		self.project.clear()
		top.configure(menu=menu)
		top.protocol("WM_DELETE_WINDOW",self.exit)

	def attrListPopup(self,event):
		from reportlab.graphics.widgetbase import TypedPropertyCollection
		alm = self.attrListMenu
		n = alm.index('end')
		if n>=1: alm.delete(1,'end')
		alm.entryconfigure(0,state=self._attrStack==[] and 'disabled' or 'normal')
		try:
			self._popup_event = event
			al = self.attrList
			lb = al._AttrListBox__listbox
			al.selectitem(lb.nearest(event.y))
			name = self.getFullObjectName()
			aname = al._lastsetentry[1]
			unsub = aname[0]!='['
			lhs = _findLHS(name+(unsub and '.' or '')+aname)
			if lhs!=None: lhs = lhs[5:]
			if not unsub:
				sname, aname = _stripSub(aname)
				sname = '[%s].%s' % (sname,aname)
			else:
				sname = aname = _findLHS(aname)
			obj = self.getSampleAttr(self.attrRootName)
			if isinstance(obj,TypedPropertyCollection):
				obj = getattr(obj,'_value',obj)
			am = getattr(obj,'_attrMap',None)
			inMap = am and am.has_key(aname) or 0
			if inMap:
				from reportlab.lib.validators import isBoolean, OneOf
				from guidialogs import _addFlipButton
				ame = am[aname]
				desc = getattr(ame,'desc')
				if desc:
					f = tkFont.Font(alm,alm.cget('font'))
					s = int(f.cget('size'))
					if s>=10: f.configure(size=s-2)
					alm.add_command(label=desc,foreground='darkgreen',font=f)
				validate = ame.validate
				if validate is isBoolean:
					alm.add_separator()
					value = self.getSampleAttr(lhs)
					#var = Tkinter.IntVar()
					#var.set(value)
					#alm.add_checkbutton(label=' '+sname,variable=var,
					#	indicatoron=1,foreground=(value and 'darkgreen' or 'darkred'),
					#	command=lambda v=var,n=sname,cmd=self.attrChange: cmd(new='%s=%d'%(n,v.get())))
					_addFlipButton(alm,label=sname,value=value, flip=['is off', 'is on'],
						command=lambda v,n=sname,cmd=self.attrChange: cmd(new='%s=%d'%(n,v)))
					writeLog('debug',str(alm.configure()))
					writeLog('debug',str(alm.entryconfigure(0)))
				elif isinstance(validate,OneOf):
					value = self.getSampleAttr(lhs)
					alm.add_separator()
					for s in validate._enum:
						alm.add_command(label=s,foreground=(value==s and 'darkgreen' or 'darkred'),
							command=lambda v=s,n=sname,cmd=self.attrChange: cmd(new='%s=%s'%(n,repr(v))))

			alm.post(event.x_root, event.y_root)
		finally:
			del self._popup_event

	def saveConfig(self):
		main = self.main
		config.root.geometry = root.winfo_geometry()
		self.pack_forget()
		config.save()

	def pack_forget(self):
		self.menu.pack_forget()
		main = self.main
		if main:
			for i in main.panes():
				getattr(config,i).size=main._size[i]
				w=getattr(self,i)
				w.pack_forget()
				w.destroy()
				setattr(self,i,None)
			main.pack_forget()
			self.main = None
			main.destroy()

	def pack(self):
		main = self.main = Pmw.PanedWidget(root)
		if self._state=='splash':
			self.splash = Tkinter.Label(main.add('splash',size=config.splash.size),
										text = 'ReportLab GUI Drawing Editor',
										background = 'black', foreground = 'white',
										padx = 30, pady = 30)
		else:
			self.pmd = PMDisplay(main.add('pmd',size=config.pmd.size),
				labelpos = 'n',
				label_text = None,
				usehullsize = 0,
				hull_width = 200,
				hull_height = 200,
				vertflex='fixed',
				horizflex='fixed',
				)

			self.attrList = attrList = AttrListBox(main.add('attrList',size=config.attrList.size),
						labelpos = 'n',
						label_text='Attributes',
						)
			attrList.configure(
					dblclickcommand=self.attrListExpand,
					dncommand=self.attrListExpand,
					upcommand=self.attrListContract,
					)
			ef = attrList._entryfield
			balloon = self._balloon
			balloon.bind(ef,'Make changes here\nand press return')
			self._button3(config.button3,1)
			attrList._entryfield.configure(command=self.attrChange)
			w = attrList.component('label')
			w.bind('<Double-1>',self.attrListContract)
			w.configure(foreground=upColor)
			balloon.bind(w,'Double-click for\nparent attributes')
			balloon.bind(attrList._ub,'Double-click for\nparent attributes')
			if config.src.on:
				self.src = RL_ScrolledText(main.add('src',size=config.src.size),
							autoclear=1, colors=config.src.colors,
							labelpos = 'n', label_text = 'Source')
		if config.log.on:
			self.log = RL_ScrolledText(main.add('log',size=config.log.size),
				initText = _initLogText,
				colors=config.log.colors,
				maxlines = config.log.maxlines,
				labelpos = 'n', label_text = 'Log')
		for i in self.main.panes():
			getattr(self,i).pack(padx=1,pady=2,expand=1,fill='both')
		#self.menu.pack(fill = 'x')
		main.pack(pady=2,fill='both',expand=1)

	def setMain(self,state):
		if self._state==state: return
		self.pack_forget()
		self._state = state
		self.pack()

	def closeProject(self):
		if self.project._checkModified('Close'):
			self.project.clear()

	def newProject(self):
		if not self.project._checkModified('New'): return
		self.findObjects()
		from reportlab.graphics.shapes import Drawing
		N, C = _getClassList(lambda c,D=Drawing: issubclass(c,D))

		d = NameDialog(
					title = 'New Project Parameters',
					buttons = ('OK', 'Cancel'),
					defaultbutton = 'OK',
					combobox_labelpos = 'n',
					label_text = 'Base Class Name?',
					namelabel = 'New Class Name',
					scrolledlist_items = N,
					)
		r = _popup(d,'newProject')
		if r != 'OK': return
		baseName = d.get()
		if not baseName in N:
			tkMessageBox.showerror('New',"Base Class must be in list")
			return
		className = d._nameField.get()
		if not className:
			i = _bnPat.search(baseName)
			base = i and baseName[:i.start()] or baseName
			for i in xrange(1000000):
				className = '%s_%03d' % (base, i)
				if className not in N: break
		base = C[N.index(baseName)]
		baseName = base.__name__
		moduleName = base.__module__
		for n in (baseName, className):
			if not n or not _checkName(n):
				tkMessageBox.showerror('New',"Invalid class name '%s'"%n)
				return
		self.project.new(className,baseName,moduleName)

	def projectOpen(self,path=None):
		self.project.open(path=path)
		self._setupRecentFiles()

	def projectSave(self):
		self.project.save()
		self._setupRecentFiles()

	def projectSaveAs(self):
		self.project.saveAs()
		self._setupRecentFiles()

	def exit(self):
		if self.project._checkModified('Exit'):
			root.wm_protocol("WM_DELETE_WINDOW", None)
			root.update()
			try:
				self.saveConfig()
			except:
				traceback.print_exc()
			root.destroy()

	def redraw(self,project=None,action='redraw'):
		if	hasattr(self,'main'):
			if project is None: project = self.project
			if action=='clear':
				self.setMain('splash')
				self.attrRootName = ''
				self._editMode = 0
			elif action=='redraw':
				src = getattr(self,'src',None)
				if src:
					src._storedViews = [src.xview()[0],src.yview()[0]]
				self.setMain('display')
				self.newDrawing()
				self.newAttrList()
				if not config.src.on: return
				src = self.src
				A = []
				src.write(self.project.buildText(annotate=A),'info')
				k = getattr(src,'_storedViews',None)
				if k:
					src.xview_moveto(k[0])
					src.yview_moveto(k[1])
					delattr(src,'_storedViews')
				for k, rg in A:
					src.tag_remove('info',rg[0],rg[1:])
					src.tag_add(k,rg[0],rg[1:])
				B3 = config.button3 and '<Button-3>' or '<Shift-Button-1>'
				src.tag_bind('basename', B3, lambda event,self=self: self._changeClassName(event,0))
				src.tag_bind('classname', B3, self._changeClassName)
				src.tag_bind('edit', B3, self._changeEdit)
				src.tag_bind('dedit', B3, lambda x,self=self: self._changeEdit(x,mode='dedit'))
				src.tag_bind('medit', B3, lambda x,self=self: self._changeEdit(x,mode='medit'))
			elif action=='modulecheck':
				if _known.checkPath(self.project.path):
					self.findObjects(force=1)

	def _set_sample_sys_argv(self):
		p = self.project
		old0 = p._sys_argv
		old = old0 or sys.argv
		d = Pmw.TextDialog(None,
				title = 'Sample sys.argv',
				label_text = 'Sample sys.argv elements one per line',
				scrolledtext_labelpos = 'n',
				defaultbutton=0,
				buttons = ('OK', 'Cancel'),
				scrolledtext_text_width= min(120,max(max(map(len,old)),30)),
				)
		d.unbind('<Return>')
		d.insert('end','\n'.join(old))
		r=_popup(d,'__set_sample_sys_argv')
		if r!='OK': return
		new = filter(None,d.get().split('\n'))
		if len(new)<1 or new==sys.argv: new = None
		if new==old0: return
		try:
			p._sys_argv = new
			if p.className: self.redraw()
			return
		except:
			p._sys_argv = old0
			self._showerror('Alter Sample sys.argv', "Can't change sample.sys.argv:\n%s" % new)
			if p.className: self.redraw()

	def _changeEdit(self,event,mode='edit'):
		p = self.project
		e = getattr(p,mode+'s')
		src = self.src
		x = event.x
		i = int(float(src.index("@%d,%d" % (x,event.y))))-int(float(src.index(mode+".first")))
		if i < 0:
			i = 0
		elif i>=len(e):
			i = len(e)-1
		old = e[i]
		d = Pmw.PromptDialog(None,
				title = 'Alter%s Edit' % (mode=='dedit' and ' Dynamic ' or (mode=='medit' and ' __main__' or '')),
				label_text = 'edit:',
				entryfield_labelpos = 'w',
				defaultbutton = 0,
				buttons = ('OK', 'Remove', 'Cancel'),
				entryfield_entry_width=min(120,len(old)),
				)
		#d.overrideredirect(1)
		d.setentry(old)
		while 1:
			r=_popup(d,'_changeEdit',pos="+%d+%d" % (event.x_root-x, event.y_root+20))
			if r not in ('OK','Remove'): return
			new = d.get()
			r = r=='Remove'
			if r:
				del e[i]
			else:
				addedFroms = []
			while 1:
				try:
					if not r: e[i] = new
					self.redraw()
					self.project._modified = 1
					return
				except:
					if r:
						exceptionLog()
						e.insert(i,old)
						t, m = sys.exc_info()[:2]
						O = string.split(self.getFullObjectName(),'.')
						if (t==AttributeError and string.find(str(m),repr(O[-1]))>=0) \
							or (t==IndexError and 'contents' in O):
								writeLog('info',"Are you deleting what's being inspected?")
						self.redraw()
					else:
						e[i] = old
						if r:
							m = "Can't delete %s:\n%s" % (mode,e[i])
						else:
							m = "Can't change %s:\n%s" % (mode,new)
						if self._exceptionFixup(addedFroms,'Alter Edit', m): continue
					break


	def _changeClassName(self,event,c=1):
		if c:
			aName, bName = 'className', ''
		else:
			addedFroms = []
			aName, bName = 'baseName', 'Base '
		title = bName + 'Class Name'
		p = self.project
		old = getattr(p,aName)
		d = Pmw.PromptDialog(None,
				title = title,
				label_text = title+':',
				entryfield_labelpos = 'w',
				defaultbutton = 0,
				buttons = ('OK', 'Cancel'),
				)
		d.setentry(old)
		#d.overrideredirect(1)
		r= _popup(d,'_changeClassName', pos="+%d+%d" % (event.x_root-event.x, event.y_root+20))
		if r!='OK': return
		new = d.get()
		while 1:
			try:
				setattr(p,aName,new)
				self.redraw()
				self.project._modified = 1
				return
			except:
				setattr(p,aName,old)
				if c or not self._exceptionFixup(addedFroms,
					title, "Could not rename %sclass %s to class %s" % (bName, old,new)): break

	def newDrawing(self):
		project = self.project
		self.project.getSample()
		self.pmd.newDrawing(config.pmd.zoom, project.sample,
			'File: %s, class %s(...,%s)' % (project.path or '<unknown>', project.className,project.baseName))

	def getFullObjectName(self,arn=None):
		if not arn: arn = self.attrRootName
		if arn !='':
			arn = 'self.'+arn
		else:
			arn = 'self'
		return arn

	def newAttrList(self,active=None):
		attrList = self.attrList
		arn = self.attrRootName
		obj = self.getSampleAttr(arn)
		text = "%s Attrs "% self.getFullObjectName(arn)
		newText = attrList['label_text']!=text
		if active is None:
			if newText:
				active = 0
			else:
				n = attrList.curselection()
				if len(n):
					active = int(n[0])
				else:
					active= attrList._lastsetentry[0]

		attrList['label_text'] = text
		A = mutils.getAttributes(obj,config.privateAttrs)
		n = max(map(lambda x: len(x[0]),A))
		attrList._n = n
		attrList.setlist(map(lambda x,fmt="%-"+str(n)+"s = %s": fmt % (x[0],obj2str(x[1])), A))
		attrList.selectitem(active)

	def attrListExpand(self,S):
		attrList = self.attrList
		if S:
			if S[attrList._n] == '=':
				S = [S[:n],S[n+1]]
			else:
				S = string.split(S,'=')
			S = map(string.strip,S)
			s = S[0]
			a = self.attrRootName
			if a: a = a+(s[0]!='[' and '.' or '')+s
			else: a=s
			try:
				obj = self.getSampleAttr(a)
			except:
				writeLog('debug','getSampleAttr(%s) failed'%repr(a))
				exceptionLog()
				return
			if hasattr(obj,'__class__'):
				self._attrStack.append((self.attrRootName,int(attrList.curselection()[0])))
				self.attrRootName = a
				try:
					self.newAttrList()
				except:
					self.attrListContract(None)
					writeLog('warn','failed to expand '+a)
					raise

	def attrListContract(self,S=None):
		stack = self._attrStack
		attrList = self.attrList
		if stack==[]:
			writeLog('warn','Can\'t contract!')
			if not len(attrList.curselection()):
				attrList.selectitem(0)
			return
		self.attrRootName, active = stack.pop()
		self.newAttrList(active)

	def attrChange(self, new=None, name=None):
		'''this is called when the entry field is changed'''
		attrList = self.attrList
		if not new:
			new = string.strip(attrList._entryfield.get())
		active, current = attrList._lastsetentry
		addedFroms = []
		project = self.project
		if self._editMode:
			edits = project.dedits
		else:
			edits = project.edits
		if new != current:
			while 1:
				try:
					name = name or self.getFullObjectName()
					e = name+(new[0]!='[' and '.' or '')+new
					if edits:
						en = _findLHS(e)
						if en:
							eh = edits[-1]
							if en==_findLHS(eh):
								en = eh
								del edits[-1]
							else:
								en = None
					else:
						en = None
					edits.append(e)
					self.redraw()
					project._modified = 1
					return
				except:
					del edits[-1]
					if en: edits.append(en)
					if not self._exceptionFixup(addedFroms,'Atribute Change','Couldn\'t set %s' % e): break

	def _exceptionFixup(self,addedFroms, title=None, msg=None):
		'''Return 1 if we can possibly fix up and exception by adding a new import'''
		ex_type, ex_value = sys.exc_info()[:2]
		if ex_type==NameError:
			m = _excNamePat.match(str(ex_value))
			if m:
				m = m.group('name')
				K = self.findObjects(m)
				if len(K)==1:
					if K.keys()[0]=='':
						# it's a pure module
						I,m = m,''
						self.project.addImport(I)
					else:
						I = K.values()[0][0].__name__
						self.project.addFrom(I,m)
					addedFroms.append((I,m))
					return 1
		self.project.delFroms(addedFroms)
		return self._showerror(title,msg)

	def _showerror(self,title,msg,sample=1):
		exceptionLog(sample and self.project._txt or None)
		self.redraw(action='redraw')
		if title and msg:
			tkMessageBox.showerror(title, msg )
			writeLog('warn', msg)
		return 0

	def getSampleAttr(self,name):
		if name and name[0]!='.': name = '.'+name
		return eval('self.project.sample'+name)

	def findObjects(self,name=None,force=0):
		global _known
		if force or not _known:
			t0 = time()
			writeLog('info','making list of known objects','flush')
			pmwMods = mutils._checkPmw()
			_known = mutils.KnownNames((config.package.search,
				config.package.exclude+['*.guiedit.guiedit','*.guidialogs']))
			mutils._uncheckPmw(pmwMods)
			writeLog('info','finished list of _known objects %.2f"'%(time()-t0))
			writeLog('info','#M=%d #F=%d #C=%d #VN=%d #VI=%d'%(len(_known.modules),len(_known.funcs.keys()),
						len(_known.classes.keys()),len(_known.varsN.keys()),len(_known.varsI.keys())))
		if name: return _known.findName(name)

	def findObjectsFirst(self):
		self._mN = filter(lambda x: sys.modules[x],sys.modules.keys())
		self.findObjects()

	def clearLog(self):
		if self.log: self.log.clear()

	def zoomSet(self):
		d = Pmw.PromptDialog(None,
				title = 'Image Zoom',
				label_text = 'Zoom:',
				entryfield_labelpos = 'w',
				defaultbutton = 0,
				buttons = ('OK', 'Cancel'),
				)
		s = "%.2f" % (100*config.pmd.zoom)
		if s[-1]=='0': s=s[:-1]
		if s[-1]=='0': s=s[:-1]
		if s[-1]=='.': s=s[:-1]
		s0=s = s+'%'
		d.setentry(s)
		r=_popup(d,'zoomSet')
		if r!='OK': return
		r0 = r = string.strip(d.get())
		if r[-1]=='%': r = r[:-1]
		try:
			z = float(r)/100.
			if z<=0.001 or z > 1600: raise ValueError, "Invalid Zoom"
		except:
			writeLog('warn','Cannot make a zoom out of %s' % r0)
			return
		r = config.pmd.zoom - z
		if r<0: r = -r
		if r>0.01:
			writeLog('info', "Zoom changed from %s to %s" % (s0,r0))
			config.pmd.zoom = z
			if self.project.className:
				self.redraw()

	def editSearch(self):
		'''Allow user to change the searched packages'''
		d = EditListDialog(
					title = 'Packages to Search',
					buttons = ('OK', 'Cancel'),
					defaultbutton = 'OK',
					combobox_labelpos = 'n',
					label_text = 'Packages',
					scrolledlist_items = config.package.search,
					balloon=self._balloon,
					)
		r=_popup(d,'editSearch')
		if r != 'OK': return
		L = list(d.getList())
		if L!=config.package.search:
			config.package.search = L
			self.findObjects(force=1)

	def editExclude(self):
		'''Allow user to change the searched packages'''
		d = EditListDialog(
					title = 'Package Exclusion Patterns',
					buttons = ('OK', 'Cancel'),
					defaultbutton = 'OK',
					combobox_labelpos = 'n',
					label_text = 'Exclusion Patterns',
					scrolledlist_items = config.package.exclude,
					balloon=self._balloon,
					)
		r=_popup(d,'editExclude')
		if r != 'OK': return
		L = list(d.getList())
		if L!=config.package.exclude:
			config.package.exclude = L
			self.findObjects(force=1)

	def reloadModules(self):
		'''Reload most modules'''
		project = self.project
		redraw = self._state=='display' and project.sample is not None
		B = self._mN+['rlextra.graphics.guiedit.mutils','copy','tempfile','zipfile']
		if redraw: project.sample = None
		M = filter(lambda m, B=B: m not in B and hasattr(sys.modules[m],'__file__') and sys.modules[m].__file__[-4:].lower()!='.pyd',sys.modules.keys())
		try:
			_known.reloadModule(M)
		except:
			exceptionLog()
		if redraw: self.redraw()

	def addWidget(self):
		N = self._state=='display' and self.project.sample is not None
		if not N: return
		self.findObjects()
		from reportlab.graphics.widgetbase import Widget
		from reportlab.graphics.shapes import Shape
		_sk = _shapes.keys()
		N, C = _getClassList(lambda c, draw=Widget.draw, S=Shape, W=Widget, _sk=_sk: \
									(issubclass(c,W) and c.draw!=draw) or \
										(issubclass(c,S) and  c.__name__ in _sk))
		for i in xrange(len(N)):
			n = N[i]
			if issubclass(C[i],Shape) and n in _sk:
				N[i] = n+_shapes[n]
		d = NameDialog(
					title = 'Add New Widget',
					buttons = ('OK', 'Cancel'),
					defaultbutton = 'OK',
					combobox_labelpos = 'n',
					label_text = 'Widgets?',
					namelabel = 'New Widget Name',
					scrolledlist_items = N,
					)
		r=_popup(d,'addWidget')
		if r != 'OK': return
		wName = d.get()
		if wName[-1]!=')':
			wName = wName + '()'
		aName = d._nameField.get() or None
		objName = self.getFullObjectName()
		self.attrChange(new="_add(%s,%s,name=%s,validate=None,desc=None)" % (objName, wName, repr(aName)),
						name='self')

	def _setupRecentFiles(self,newpath=None):
		RFL = config.recent_files
		newpath = newpath or self.project.path
		if newpath:
			if newpath in RFL:
				if RFL[0]==newpath: return
				RFL.remove(newpath)
			RFL.insert(0,newpath)
		while len(RFL)>5: del RFL[-1]
		menu = self.menu
		try:
			menu.deletemenuitems('File',menu.component('File').index('Export')+1,'end')
		except:
			pass
		if not RFL: return
		menu.addmenuitem('File', 'separator')
		for i in xrange(len(RFL)):
			menu.addmenuitem('File', 'command', 'Open '+RFL[i],
				command = lambda self=self, path=RFL[i]: self.projectOpen(path),
				label = '%d: %s'%(i+1,os.path.basename(RFL[i])))

	def _srcToggle(self,v):
		config.src.on = v
		if self.project.className:
			self._state = None
			self.redraw()

	def _logToggle(self,v):
		config.log.on = v
		self._state = None
		if self.project.className:
			self.redraw()
		else:
			self.redraw()
		writeLog('info','Logging turned %s' % (v and 'on' or 'off'))

	def _editModeToggle(self,v):
		self._editMode = v
		writeLog('info','Edit getContents turned %s' % (v and 'on' or 'off'))

	def _testAttrFilterToggle(self,v):
		config.testAttrFilter = v
		writeLog('info','testAttrFilter turned %s' % (v and 'on' or 'off'))

	def _errorBellsToggle(self,v):
		config.errorBells = v
		writeLog('info','Error Bells turned %s' % (v and 'on' or 'off'))

	def _debugToggle(self,v):
		config.debug = v
		writeLog('info','DEBUG turned %s' % (v and 'on' or 'off'))

	def _privateAttrsToggle(self,v):
		config.privateAttrs = v
		if self._state=='display' and self.project.sample is not None: self.newAttrList()
		writeLog('info','privateAttrs turned %s' % (v and 'on' or 'off'))

	def _balloonToggle(self,v):
		config.root.balloon = v
		self._balloon.configure(state=v and 'balloon' or 'none')
		writeLog('info','Tooltips turned %s' % (v and 'on' or 'off'))

	def _button3(self,v,which=3):
		config.button3 = v

		B1 = '<Shift-Button-1>'
		B3 = '<Button-3>'
		if not v:
			B1, B3 = B3, B1

		if 1 & which:
			lb = self.attrList._AttrListBox__listbox
			self._balloon.bind(lb,'Double click to expand\nUse %s for info' % (v and 'right' or 'shift'))
			lb.bind(B1,None)
			lb.bind(B3,self.attrListPopup)
		if (2 & which) and self.project.className:
			self._state = None
			self.redraw()

def _getConfig(fn=None):
	global config
	if not config:
		if fn is None:
			fn = os.path.join(os.path.dirname(sys.argv[0]),'_%s.ini' % os.path.splitext(os.path.basename(sys.argv[0]))[0])
		config = Config(fn,
					package = SubConfig(
						search = [	'reportlab.lib','reportlab','rlextra'],
						exclude = [	'reportlab.demo.*','reportlab.test*',
									'reportlab.doc*','reportlab.utils.*','rlextra.preppy.flintstones*',
									'reportlab.lib.graphdocpy',
									'reportlab.lib.docpy', 'reportlab.tools*',
									],
						),
					main=SubConfig(
						size=400,
						),
					log = SubConfig(
						on=1,
						maxlines=200,
						colors = {
								'info': {'foreground':'blue'},
								'debug': {'foreground':'red'},
								'warn': {'foreground':'orange','background':'grey50'},
								'error': {'foreground':'red','background':'grey50'},
								'security': {'foreground':'blue','background':'red'},
								},
						size = 100,
						max = 0.25,
						min = 0,
						),
					root = SubConfig(
						geometry = '400x500+100+100',	#width X height +x+y
						balloonDelay = 800,				# milliseconds
						balloon=1,
						),
					pmd = SubConfig(
						on=1,
						size = 200,
						min = 100,
						max = 1.0,
						zoom = 1.0,
						),
					splash = SubConfig(
						on=1,
						size = 200,
						min = 0,
						max = 1.0,
						zoom = 1.0,
						),
					attrList = SubConfig(
						on=1,
						size = 200,
						min = 100,
						max = 1.0,
						),
					src = SubConfig(
						on = 1,
						size = 200,
						min = 100,
						max = 1.0,
						colors ={
								'info': {'foreground':'blue'},
								'classname': {'foreground': 'red'},
								'basename': {'foreground': '#ff00ff'},
								'edit': {'foreground': 'darkred'},
								'dedit': {'foreground': '#804000'},
								'medit': {'foreground': '#808000'},
								},
						),
					prevDirs = SubConfig(
							saveAs = None,
							export = None,
							open = None,
							),
					popupGeom = {},
					button3 = 1,
					debug=0,
					testAttrFilter=1,
					privateAttrs=0,
					errorBells=0,
					recent_files=[],
				)
	return config
######################################################################
# Create demo in root window for testing.
def mainApp():
	#sys.stdout = sys.stdin = open(os.path.join(os.path.dirname(sys.argv[0]),'_guiedit.out'),'w')
	global root, editor, config
	writeLog('info','Configuration Starting')
	_getConfig()
	root = Tkinter.Tk()
	#Pmw.initialise(root, fontScheme = 'pmw1')
	Pmw.initialise(root, fontScheme = None )
	root.title(title)
	root.geometry(config.root.geometry)
	editor = GUIEditor(root)
	root.after(500,editor.findObjectsFirst)
	writeLog('info','Configuration Finished')
	if len(sys.argv)!=1: editor.projectOpen(sys.argv[1])
	root.focus_set()
	root.mainloop()

if __name__ == '__main__':
	mainApp()
