#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/pageCatcher/pageCatcher.py
"capture pages from PDF files, SEE script_usage FOR USAGE."


## EXAMPLE RUN (FEB 2001)
##C:\reportlab\repository\rlextra\pageCatcher>pageCatcher.py MakeForms testfile.pdf dump.data prefix --test testoutput.pdf 1 2
##Makeforms from testfile.pdf to dump.data page numbers [1, 2]
##
##
## placing pages [1, 2] from testfile.pdf into dump.data
##storing to dump.data
##making test grid from dump.data into testoutput.pdf
##now unpickling
##checking license
##license check failed
##form prefix1
##form prefix2
##wrote test file testoutput.pdf
##
##created forms ['prefix1', 'prefix2'] in dump.data
##
##C:\reportlab\repository\rlextra\pageCatcher>

import md5

UnresolvedRefError = "UnresolvedRefError"

from reportlab.rl_config import pageCompression, verbose, invariant

DECOMPRESSFILTERS = 1 # set to auto decompress on input (recommended!)
COMPACT = pageCompression # set to compress on output

import sys, string, types, cPickle

from pprint import pprint

from reportlab.pdfbase import pdfdoc, pdfutils
from reportlab.pdfgen import canvas
from reportlab.lib.utils import getStringIO, open_and_read

# hack to support both kinds of x objects we have defined
if not hasattr(pdfdoc,'xObjectName'): pdfdoc.xObjectName = pdfdoc.formName

CHECK = {}

RERAISE = 1 # SET THIS TO GET A FULL TRACEBACK FROM main()

PAGESNAME = pdfdoc.PDFName("Pages")
PAGENAME = pdfdoc.PDFName("Page")

ESCAPEDICT = {
    "\\" : "\\",
    "(" : "(",
    ")" : ")",
    "r" : "\r",
    "n" : "\n",
    "t" : "\t",
    "b" : "\b",
    "f" : "\f",
    }
OCTALDIGITS = "01234567"

def unescapePDFText(t):
    # first nuke all whitespace
    t = string.strip(t)
    t = string.join(string.split(t), "")
    L = list(t)
    if L[0]!='<' and L[-1]!='>':
        raise ValueError, "string must be of form '<..>'"
    del L[0]; del L[-1]
    outL = []
    if len(L)%2!=0:
        #print t
        raise ValueError, "string content must have even number of digits"
    while L:
        hexdigits = L[0]+L[1]
        del L[:2]
        try:
            outord = string.atoi(hexdigits, 16)
        except:
            #print t
            raise
        outch = chr( outord )
        outL.append(outch)
    return string.join(outL, "")

def unescapePDFString(s):
    #print; print "unescaping", s; stop
    if "\\" not in s: # shortcut, no tranlate needed
        if s[0]!="(" or s[-1]!=')':
            raise ValueError, "all PDF strings must be bounded by parens (..)"
        return s[1:-1]
    inlist = list(s)
    outlist = []
    if inlist[0]!='(' or inlist[-1]!=')':
        raise ValueError, "all PDF strings must be bounded by parens (..)"
    cursor = 1
    last = len(inlist)-1
    while cursor<last:
        thischar = inlist[cursor]
        #print "unescape at", repr(thischar)
        cursor = cursor+1
        if thischar=="\\":
            #print "escape!"
            next = inlist[cursor] # must have a next, otherwise not well formed
            cursor = cursor+1
            if ESCAPEDICT.has_key(next):
                #print "escapedict", repr(next)
                #print cursor,len(outlist), "alpha escape", thischar+next, "-->", repr(ESCAPEDICT[next])
                thischar = ESCAPEDICT[next]
                #cursor = cursor+1
            elif next in "\r\n": # skip end of line
                thischar = "" # no char
                # must have next, else not well formed
                next = inlist[cursor]
                if next in "\r\n":
                    # skip other eoln char
                    cursor = cursor+1
            elif next in OCTALDIGITS:
                #print "octal1", repr(next)
                octalstring = next
                if inlist[cursor] in OCTALDIGITS:
                    octalstring = octalstring+inlist[cursor]
                    #print "octal2", octalstring
                    cursor = cursor+1
                    if inlist[cursor] in OCTALDIGITS:
                        octalstring = octalstring+inlist[cursor]
                        #print "octal3", octalstring
                        cursor = cursor+1
                octal = string.atoi(octalstring, 8)
                thischar = chr(octal)
                #print cursor, len(outlist), "octal char", repr(octalstring), "-->", repr(thischar)
            else:
                raise ValueError, "don't understand string escape sequence %s %s" % (repr(thischar), repr(next))
        else:
            #print cursor, len(outlist), "passthru", repr(thischar), "-->", repr(thischar)
            pass # no translation needed
        outlist.append(thischar)
    result = string.join(outlist, "")
    #print "UE", repr(s), "-->", repr(result)
    return result

def unescapePDFCharacterSequence(s):
    s = str(s) # in case it is an object, convert to string
    s0 = s[0]
    if s0=="(":
        return unescapePDFString(s)
    elif s0=="<":
        return unescapePDFText(s)
    else:
        raise ValueError, "bad pdf char sequence starter %s" % repr(s0)

def getPdfEncrypt():
    # we now no longer do anything to do with encryption...
    raise ImportError, "\nThis file appears to be encrypted. Please save an unencrypted version of this file and try using pageCatcher on the unencrypted file."

class PDFFileCompilation:
    def __init__(self, prefix):
        self.prefix = prefix
        self.sequence = []
        # mapping name --> reference
        self.references = {}
        # mapping name --> tree
        self.trees = {}
        # mapping name --> pdfdoc object
        self.objects = {}
        # mapping name --> list of name
        self.uses = {}
        self.catalogName = None
        self.pageList = []
        self.lastObjectNumber = self.lastGenerationNumber = \
        self.encrypt = self.Ownerkey = self.IDarray1 = self.Permissions = None
        self.plainText = {}
    def encryptParameters(self, password, OwnerKey, UserKey, IDarray1, Permissions):
        if self.encrypt:
            raise ValueError, "encryptParameters set twice (not permitted)."
        getPdfEncrypt()
    def pagesAsForm(self, pagenumbers, formprefix, all=None,fformname=None):
        """get dictionary of name-->object including form streams for the pagenumbers
           page numbers start at 0 and page 4 will have the pdfgen form
           name formprefix4 (if formprefix=="formprefix")
        """
        # result of function: object reference name --> pdfdoc object
        # special entry None --> list of EXTERNAL names of forms created (NOT INTERNALS!)
        nameToObject = {}
        nameToObject[None] = formNamesList = []
        #print "self.uses", self.uses
        import copy
        uses = nameToObject[0] = self.uses
        # collect all referents in pages (but don't trace parent, don't include pages themselves)
        requiredNames = {}
        forbiddenNames = {}
        # contents streams and arrays will have to be removed later, remember them.
        pageContentNames = {}
        # if all is set then do all pages
        if all:
            pagenumbers = range(len(self.pageList))
        for n in pagenumbers:
            # collect all references in the page dictionaries...
            pagename = self.pageList[n]
            forbiddenNames[pagename] = pagename
            pagerefs = list(self.uses[pagename]) # make a copy for modification
            page = self.objects[pagename]
            if page.dict.has_key("Parent"):
                raise ValueError, "parent not removed"
            if page.dict.has_key("Annots"):
                raise ValueError, "annotation not removed"
            for name in pagerefs:
                requiredNames[name] = name
            compiledpagerep = self.trees[pagename]
            (ind, dict) = compiledpagerep
            dict = dict.copy()
##            del dict["Parent"] # don't trace the parent link
##            for (ind2, thing) in dict.values():
##                if ind2=="R": # reference: trace it
##                    name = self.internalName(thing)
##                    requiredNames[name] = name
            # also, remember the references to content streams, and any indirect references to them
            # we don't need to include them twice in the output!
            c = dict["Contents"]
            # resolve it if it is a reference
            while 1: # we already know there are no loops (previous steps)
                (ind, thing) = c
                if ind=="R":
                    c = thing
                    name = self.internalName(thing)
                    #print "name=", name
                    pageContentNames[name] = name
                else:
                    break # stop at first non-reference
            # if content is an array, loop over contents
            (ind, thing) = c
            if ind=="array":
                for (ind2, thing2) in thing:
                    if ind2=="R": # I assume the reference is direct... (some extra garbage included if not)
                        name = self.internalName(thing2)
                        #print "name=", name
                        pageContentNames[name] = name
        # transitively close the required references
        check = requiredNames.keys()
        uses = self.uses
        while check:
            currentname = check[-1]
            del check[-1]
            requiredNames[currentname] = currentname
            for othername in uses[currentname]:
                if forbiddenNames.has_key(othername):
                    #print "forbidden", othername, "from", currentname
                    raise ValueError, "forbidden object encountered "+repr((othername, currentname))
                if not requiredNames.has_key(othername):
                    check.append(othername)
        # now remove from the references names of page content arrays and streams
        # (these must be *traced* but should not be included in the output, since they are converted)
        for name in pageContentNames.keys():
            del requiredNames[name]
        # store the required elements in result
        rk = requiredNames.keys()
        for name in rk:
            ob = self.objects[name]
            nameToObject[name] = ob
        # construct the forms
        for n in pagenumbers:
            externalFormName = fformname or formprefix+repr(n)
            fformname = None
            formName = pdfdoc.xObjectName(externalFormName)
            formNamesList.append(externalFormName)
            pagename = self.pageList[n]
            # remember objects required for this form
            #print formName, "uses", uses[pagename]
            uses[formName] = list(uses[pagename])
            compiledpagerep = self.objects[pagename]
            # get the content for the page
            contents = compiledpagerep.dict["Contents"]
            contents = self.resolve(contents)
            # there are two cases, single stream or array of stream
            filters = None
            if isinstance(contents, pdfdoc.PDFStream):
                filters = contents.dictionary.dict.get("Filter", None)
                contentstring = contents.content
            elif isinstance(contents, pdfdoc.PDFArray):
                contentlist = []
                for s in contents.sequence:
                    s = self.resolve(s)
                    # if it has an unexpanded filter give up!
                    if s.dictionary.dict.has_key("Filter"):
                        raise ValueError, "unexpanded filter in page array stream: giving up "+\
                              repr(s.dictionary.dict["Filter"])
                    contentlist.append(s.content)
                contentstring = string.join(contentlist, "")
            else:
                raise ValueError, "bogus contents %s" % repr(contents)
            if not COMPACT and not filters:
                contentstring = string.replace(contentstring, "\n", "\r\n")
            self.plainText[n] = contentstring

            theform = pdfdoc.PDFStream()
            sdict = theform.dictionary
            # this attempt at security failed since it would require back patching the Resources Dict...
            # the idea was to add "evaluation copy" directly on top of the form.
##            if contentadjustment:
##                if filters:
##                    raise "Unpaid evaluation copy cannot handle filters: " + repr(filters)
##                # mess up the form
##                contentstring = ("q\n%s\n" %contentstring)+contentadjustment
            # do compression
            if filters is not None:
                sdict["Filter"] = filters # HOW DID THAT GET COMMENTED?
                if verbose: print "filters is",filters
                pass
            elif COMPACT:
                if verbose: print "FlateDecode set"
                sdict["Filter"] = pdfdoc.PDFName("FlateDecode")
                contentstring = pdfdoc.PDFZCompress.encode(contentstring)
            theform.content = contentstring
            sdict["Type"] = pdfdoc.PDFName("XObject")
            sdict["Subtype"] = pdfdoc.PDFName("Form")
            sdict["FormType"] = 1
            sdict["BBox"] = compiledpagerep.dict["MediaBox"]
            sdict["Matrix"] = pdfdoc.PDFArray([1,0,0,1,0,0]) # XXXX IS THIS ALWAYS RIGHT???
            sdict["Resources"] = compiledpagerep.dict["Resources"]
            nameToObject[formName] = theform
        return nameToObject
    def getReference(self, idAndVersion):
        name = self.internalName(idAndVersion)
        return self.references[name]
    def doTranslations(self, catalogInfo):
        "translate all elements to pdfdoc structures, after all names have been added"
        for name in self.sequence:
            #print "NAME IS", name
            #stop
            (prefix, pair) = name
            (self.lastObjectNumber, self.lastGenerationNumber) = pair # for encryption
            tree = self.trees[name]
            object = self.translateTree(tree)
            self.objects[name] = object
        self.catalogName = self.internalName(catalogInfo)
    def sanitizePages(self, save=("Type", "Contents", "MediaBox", "Resources")):
        "for all page dictionaries, cut out attributes not saved (for form translation)"
        #print "sanitizing"
        for name in self.sequence:
            tree = self.trees[name]
            (ind, val) = tree
            if ind=="dictionary":
                (ind2, typ) = val.get("Type", (None, None))
                #print " dictionary", `typ`
                if typ=="Page":
                    for name in val.keys():
                        if name not in save:
                            #print "  cutting", name
                            del val[name]
    def resolve(self, ob):
        "find object referred to if reference (after translations)"
        seen = {}
        #ob = self.objects[name]
        while isinstance(ob, pdfdoc.PDFObjectReference):
            name = ob.name
            if seen.has_key(name):
                raise ValueError, "infinite reference loop %s" % repr(name)
            seen[name] = 1
            ob = self.objects[name]
        return ob
    def populatePageList(self):
        "create list of names for pages self.pageList, after doTranslations"
        catalog = self.objects[self.catalogName]
        # note catalog is a pdfdoc.pdfDictionary, not a pdfdoc.pdfcatalog
        pagesref = catalog.dict["Pages"]
        pagesname = pagesref.name
        pages = self.resolve(pagesref)
        self.expandPages(pagesref)
    def expandPages(self, pagesref, inheritancedict=None,
                    #inheritableattributes = ("MediaBox", "Resources", "CropBox", "Rotate", "Dur", "Hid", "Trans")
                    # for these purposes we are only interested in mediabox and resources (others may mess us up)
                    inheritableattributes = ("MediaBox", "Resources")
                    ):
        # note: to make pageToForm conversion work, attributes "inherited" down
        # the Pages tree are explicitly copied down and references are patched.
        ### NOTE: THIS MAKES THE PAGE DICTIONARY PARSE TREES NO LONGER IN SYNC WITH THE
        ###  THE COMPILED PDFDOC DICTIONARY OBJECTS
        if inheritancedict is None:
            inheritancedict={}
        else:
            inheritancedict = inheritancedict.copy() # no sharing across recursions
        pagesName = pagesref.name
        pagesDictionary = self.resolve(pagesref)
        # paste in inheritable attributes if absent (so page dicts are self contained)
        pd = pagesDictionary.dict
        for a in inheritableattributes:
            if inheritancedict.has_key(a):
                if not pd.has_key(a): # force down the inherited value
                    inheritedobject = inheritancedict[a]
                    pd[a] = inheritedobject
                    # patch the uses list!
                    if type(inheritedobject) is types.InstanceType:
                        # add any references in the inherited parse tree to the uses list for page dictionary
                        self.findReferences(pagesName, inheritedobject.__parse_tree__)
            elif pd.has_key(a):
                inheritancedict[a] = pd[a] # remember this inherited value
        dictType = pagesDictionary.dict["Type"]
        if dictType==PAGENAME:
            # it is a page. record its name
            self.pageList.append(pagesName)
        elif dictType==PAGESNAME:
            # it is a pages tree: expand it
            kids = self.resolve(pagesDictionary.dict["Kids"]) # should be an array
            for p in kids.sequence:
                self.expandPages(p, inheritancedict)
    def translateTree(self, tree):
        #dieHere
        # Note: decryption herein will "decode" the encryption dictionary strings.
        #   turning them into garbage (since they should not be decrypted).  But this doesn't
        #   matter, since they have already been extracted in order to enable encryption....
        (indicator, val) = tree
        if indicator=="R":
            name = self.internalName(val)
            result = self.references[name]
        elif indicator in ("true", "false", "null", "number"):
            result = val
        elif indicator=="text":
            if not self.encrypt:
                #result = "<%s>" % val # note: hexification was not undone in this parser ! BUG!
                value = unescapePDFText("<%s>" % val)
                result = pdfdoc.PDFText(value)
            else:
                pdfencrypt = getPdfEncrypt()
        elif indicator=="string":
            if not self.encrypt:
                # must decode and reencode in case of encryption (?)
                value = unescapePDFString("(%s)" % val)
                result = pdfdoc.PDFString(value)
            else:
                pdfencrypt = getPdfEncrypt()
        elif indicator=="name":
            result = pdfdoc.PDFName(val)
        elif indicator=="array":
            elts = []
            for v in val:
                e = self.translateTree(v)
                elts.append(e)
            result = pdfdoc.PDFArray(elts)
        elif indicator=="dictionary":
            d = {}
            for k in val.keys():
                v = val[k]
                e = self.translateTree(v)
                d[k] = e
            result = pdfdoc.PDFDictionary(d)
        elif indicator=="stream":
            (dict, content) = val
            test = len(content)
            result = pdfdoc.PDFStream()
            ### decode content here, if desired...
            if self.encrypt:
                pdfencrypt = getPdfEncrypt()
            filters = dict.get("Filter", None)
            if verbose: print 'filters=', filters
            if DECOMPRESSFILTERS and filters is not None:
                originalcontent = content
                originalfilters = dict["Filter"]
                backout = 0 # set this on error
                del dict["Filter"]
                (ind, fx) = filters
                if ind == "array":
                    filters = fx
                else:
                    filters = [filters]
                # do decoding in order given
                for fm in filters:
                    (n, f) = fm
                    if n!="name":
                        raise ValueError, "filter not named %s" % repr(fm)
                    if f=="FlateDecode":
                        if verbose>=5: print 'FlateDecode'
                        try:
                            content = pdfdoc.PDFZCompress.decode(content)
                        except:
                            if verbose: print "WARNING: unable to zcompress"
                            backout=1
                            break
                    elif f=="ASCII85Decode":
                        if verbose>=5: print 'ASCII85Decode'
                        try:
                            content = pdfdoc.PDFBase85Encode.decode(content)
                        except:
                            if verbose: print "WARNING: unable to base 85 decode"
                            backout=1
                            break
                    else:
                        #raise ValueError, "can't handle filter %s" % repr(f)
                        backout = 1
                        break
                if backout:
                    # for example got DCTdecode, which we don't do yet
                    content = originalcontent
                    dict["Filter"] = originalfilters
            if verbose>=9: print content
            dicttuple = ("dictionary", dict)
            pd = self.translateTree(dicttuple)
            result.dictionary = pd
            result.content = content
            #### debugging!!!
            #result.content = ""
        else:
            raise ValueError, "don't understand %s indicator" % repr(indicator)
        if type(result) is types.InstanceType:
            # STORE THE PARSE TREE
            result.__parse_tree__ = tree
        return result
    def internalName(self, externalName):
        return (self.prefix, externalName)
    def addParse(self, idAndVersion, tree):
        name = self.internalName(idAndVersion)
        sequence = self.sequence
        if name in sequence:
            raise ValueError, "name not new %s" % repr(name)
        self.sequence.append(name)
        self.uses[name] = []
        self.trees[name] = tree
        self.references[name] = pdfdoc.PDFObjectReference(name)
        self.uses[name] = []
        #if findrefs:
        #    self.findReferences(name, tree)
        # self.object translated in later pass.
    def findAllReferences(self):
        sequence = self.sequence
        trees = self.trees
        for name in sequence:
            tree = trees[name]
            self.findReferences(name, tree)
    def findReferences(self, name, tree):
        (indicator, val) = tree
        if indicator in ("true", "false", "null", "number", "text", "name", "string"):
            pass # no references in these
        elif indicator=="R":
            # found a reference!
            rname = self.internalName(val)
            uses = self.uses[name]
            if rname not in uses:
                uses.append(rname)
        elif indicator=="array":
            for v in val:
                self.findReferences(name, v)
        elif indicator=="dictionary":
            for v in val.values():
                self.findReferences(name, v)
        elif indicator=="stream":
            (dict, content) = val
            for v in dict.values():
                self.findReferences(name, v)
        else:
            raise ValueError, "don't understand indicator "+repr(indicator)


class PDFParseContext:
    decompressFilters = 1
    encrypt = None

    # class global dispatch dictionary for first char parse lookup...
    dispatch_dictionary = {}

    def __init__(self, text, prefix="no prefix"):
        from string import split
        self.compilation = PDFFileCompilation(prefix)
        self.cursor = 0
        self.stack = []
        self.text = text
        self.splittext = split(text)
        self.splitcursor = 0
        self.objectdict = {} # for collecting objects (number, version)-->object
        self.objectoffsets = {} # for collecting byte offests (number, version)-->object
        self.align()
        self.trailer = None

    def parse(self):
        """parse the file by looking at the end, then the cross reference, etc.
        """
        from string import rfind
        text = self.text
        # look for the xref
        endoftext = text[-50:]
        from string import find, strip, atoi
        s = "startxref"
        f = rfind(text, s)
        if f<0: raise ValueError, "can't find %s at end of file %s" % (repr(s), repr(endoftext))
        f = f+len(s)
        f2 = find(text, "%%EOF", f)
        if f2<0: raise ValueError, "eof marker not found at file end %s" % repr(endoftext)
        startxref = atoi(strip(text[f:f2]))
        catalog = None
        # repeatedly parse all cross reference tables
        while startxref is not None:
            #print "parsing xref at", startxref
            self.cursor = startxref
            self.splitcursor = 0
            self.cursorcheck()
            if self.cursor!= startxref:
                raise ValueError, "alignment error"
            self.splitcursor =  self.splitcursor+1 # move past "xref" keyword
            self.cursorcheck()
            # parse the entries
            self.parse_xref_entries()
            #print self.objectoffsets
            # now parse the trailer
            if self.splittext[self.splitcursor]!="trailer":
                raise ValueError, "trailer not found after xref %s" % self.cursor
            self.splitcursor = self.splitcursor+1 # move past "trailer"
            self.cursorcheck()
            (marker, trailerdict) = self.getdictionary()
            # fail if encryption is specified
            if trailerdict.has_key("Encrypt"):
                # XXX
                raise ValueError, "Secure PDF Encrypted files cannot be parsed by pageCatcher"
                #print 'WARNING: ENCRYPTED'
            if trailerdict.has_key("ID"):
                # remember the id (needed for decryption)
                self.PDFID = trailerdict["ID"]
            # do we need to repeat?
            if trailerdict.has_key("Prev"):
                (mark, startxref) = trailerdict["Prev"]
            else:
                startxref = None # no more xrefs
            if catalog is None:
                self.catalog = catalog = trailerdict["Root"] # object reference
                self.trailerdict = trailerdict
            # XXXX for now ignore other info in trailerdict
        # now parse ALL objects (even ones that can't be used anymore, possibly)
        objectoffsets = self.objectoffsets.copy()
        lastlen = len(objectoffsets)
        dd = self.dispatch_dictionary
        # THIS LOOP MUST REPEAT UNTIL ALL OBJECT REFS CAN RESOLVE (REFERENCES MAY NOT BE IN FILE ORDER)
        while objectoffsets:
            #print "looking for", lastlen, "objects"
            keys = objectoffsets.keys()
            keys.sort()
            for k in keys:
                #print "trying to resolve", k
                offset = objectoffsets[k]
                self.cursor = offset
                self.splitcursor = 0
                self.cursorcheck()
                stack = self.stack = []
                # first find x y obj
                for n in (0,1):
                    numinfo = self.getnumber()
                    stack.append(numinfo)
                    self.skipwhitespace()
                idinfo = self.getobjectID()
                stack.append(idinfo)
                self.skipwhitespace()
                # now find some object (maybe fail)
                c = self.cursor
                firstchr = text[c]
                unboundparsemethod = dd.get(firstchr, None)
                failed = 0
                if unboundparsemethod is None:
                    raise ValueError, "can't parse at %s %s" % (c, `text[c:c+20]`)
                try:
                    parsepair = unboundparsemethod(self)
                except UnresolvedRefError, v:
                    #print "resolution failed", k
                    failed = 1 # need to try again for this one
                else:
                    #print "::: got :::", idinfo
                    #pprint(parsepair)
                    stack.append(parsepair)
                    self.skipwhitespace()
                    # are we at a stream keyword
                    if text[self.cursor]=="s":
                        try:
                            streaminfo = self.getstream()
                        except UnresolvedRefError, v:
                            #print "resolution failed", k
                            failed = 1 # need to try again for this one
                        else:
                            stack.append(streaminfo)
                # now we should be at the endobject keyword
                if not failed:
                    self.skipwhitespace()
                    self.getindirectObject() # this records the object in self
                    # we're done with this one!
                    #print "resolved", k
                    del objectoffsets[k]
            newlen = len(objectoffsets)
            if newlen==lastlen: raise ValueError, "no progress resolving objects %s" % repr(objectoffsets.keys())
            lastlen = newlen
        # XXXX NOW ARRANGE THINGS AS IF IT HAD BEEN DONE BY THE PROCESS METHOD
        #self.objectdict[pair] = object
        od = self.objectdict
        keys = od.keys()
        keys.sort
        stack = []
        iddict = {}
        for k in keys:
            (id, version) = k
            #if iddict.has_key(id): raise ValueError, "multiple object versions not yet supported... %s" % id
            #iddict[id]=id
            ob = od[k]
            stack.append(ob)
        stack.append(("trailer", self.trailerdict))
        self.stack = stack
        return stack


    def parse_xref_entries(self):
        "parse xref entries starting at cursor, add to self.objectoffsets (cursor *after* xref keyword)"
        sc = self.splitcursor
        text = self.text
        splittext = self.splittext
        # HACK motivated by acrobat: if we are at the startxref keyword then realign at the xref
        # (problem observed by arw)
        thispoint = splittext[sc]
        #print "parse_xref_entries at", splittext[sc:sc+4]
        if thispoint=="startxref":
            # find where the xref REALLY STARTS!
            newcursor = int(splittext[sc+1])
            # go there
            self.cursor = newcursor
            self.splitcursor = 0
            self.cursorcheck()
            sc = self.splitcursor+1
        from string import atoi, find, split
        done = None
        objectoffsets = self.objectoffsets
        while done is None:
            # get section head info
            #sc = self.splitcursor
            #print "section head at", splittext[sc:sc+6]
            firstentrynumber = int(splittext[sc])
            nentries = int(splittext[sc+1])
            sc = self.splitcursor = sc+2
            #self.cursorcheck()
            entrynumber = firstentrynumber
            #print "nentries", nentries
            for i in range(nentries):
                # get this entry (don't check eoln requirements too carefully)
                #print "near", splittext[sc:sc+5]
                [offset, generation, indicator] = splittext[sc:sc+3]
                sc = sc+3
                #print "advanced to", splittext[sc: sc+3]
                offset = atoi(offset, 10) # decimal! not octal!
                generation = atoi(generation, 10)
                if indicator=="n":
                    # in use: record it
                    objectkey = (entrynumber, generation)
                    if objectoffsets.has_key(objectkey): # apparently this is not an error!
                        #raise ValueError, "already defined %s" %repr(objectkey)
                        pass # ignore it: use the "more recent" value which occurs later in the file (for acrobat written files)
                    else:
                        objectoffsets[ objectkey ] = offset
                elif indicator=="f":
                    # free: ignore it
                    pass
                else:
                    raise ValueError, "bad refentry indicator %s" % repr(indicator)
                entrynumber = entrynumber+1
            # are we done?
            try:
                test = atoi(splittext[sc])
                #print "not done near", splittext[sc:sc+6]
            except:
                done = 1 # yep!
                #print "done near", splittext[sc:sc+6]
        self.splitcursor = sc
        self.cursorcheck()

    def resolve(self, thing):
        """resolve thing, or signal unresolved."""
        od = self.objectdict
        limit = len(od)
        count = 0
        #print "thing", thing
        (mark, rep) = thing
        while mark=="R":
            count = count+1
            if count>limit+2:
                raise ValueError, "infinite object loop "+repr(thing)
            test = od.get(rep,None)
            if test is None:
                raise UnresolvedRefError, rep # pass the rep up as the value
            thing = (mark, rep) = test
        return thing

    def getindirectObject(self):
        t = self.text; c = self.cursor
        test = self.gettrue("endobj")
        # pop the obid and object off the stack
        stack = self.stack
        [(m1,pair), object] = stack[-2:]
        if m1!="obj":
            raise ValueError, "expect objectID before object before 'endobj' got %s  at %s %s" % (m1, c, `t[c:c+20]`)
        del stack[-2:]
        self.objectdict[pair] = object
        # record in compilation
        self.compilation.addParse(pair, object)
        return ("indirectobject", (pair, object))

    dispatch_dictionary["e"] = getindirectObject

    def getobjectref(self, keyword="R"):
        t = self.text; c = self.cursor
        test = self.gettrue(thing=keyword)
        # pop off the two numbers from the stack
        stack = self.stack
        [(m1,n1), (m2,n2)] = stack[-2:]
        from types import IntType
        if type(n1) is not IntType or type(n2) is not IntType:
            raise ValueError, "expected 2 ints before %s %s %s" % (keyword, c, `t[c:c+20]`)
        del stack[-2:]
        return (keyword, (n1,n2))


    def getobjectID(self):
        return self.getobjectref(keyword="obj")

    dispatch_dictionary["R"] = getobjectref
    dispatch_dictionary["o"] = getobjectID

    def gettrue(self, thing="true"):
        lt = len(thing)
        t = self.text; c = self.cursor
        endkw = c+lt
        if t[c:endkw]!=thing:
            raise ValueError, "`%s` keyword not found %s %s" % (thing, c, `t[c:c+20]`)
        self.cursor = endkw
        self.cursorcheck()
        return (thing, thing)

    dispatch_dictionary["t"] = gettrue

    def getfalse(self):
        return self.gettrue(thing="false")

    dispatch_dictionary["f"] = getfalse

    def getnull(self):
        return self.gettrue(thing="null")

    dispatch_dictionary["n"] = getnull

    def getstream(self):
        t = self.text; c = self.cursor
        LF = "\n"; CR = "\r"
        if t[c: c+6]!="stream":
            raise ValueError, "not at stream keyword at %s %s" % (c, `t[c:c+20]`)
        if t[c+6] == LF:
            streamstart = c+7
        elif t[c+6]==CR and t[c+7]==LF:
            streamstart = c+8
        else:
            raise ValueError, "invalid stream keyword end of line %s %s" % (c, `t[c:c+20]`)
        # pop the dictionary off the stack
        stack = self.stack
        (mark, dict) = stack[-1]
        if mark!="dictionary":
            raise ValueError, "stream not preceded by dictionary %s %s %s" %(mark, c, `t[c:c+20]`)
        del stack[-1]
        # find the length
        streamlen = dict.get("Length", None)
        if streamlen is None:
            raise ValueError, "stream with no length key %s %s" % (c, `t[c:c+20]`)
        # streamlen may be an indirect object!
        streamlen = self.resolve(streamlen)
        (tag, slen) = streamlen
        streamend = streamstart+slen
        streamcontent = t[streamstart:streamend]
        # should be at "endstream"
        if t[streamend: streamend+9]!="endstream":
            if verbose: print "EXPECTED STREAM LENGTH", slen
            #print "text length", len(text)
            # try to tolerate whitespace, with a warning
            if verbose: print "WARNING stream length doesn't end at endstream keyword %s %s" % (
                                  streamend, `t[streamend:streamend+20]`)
            endstreamkw = "endstream"
            lkw = len(endstreamkw)
            f = string.find(t, endstreamkw, streamend)
            if f<0:
               pass
            elif string.strip(t[streamend:f])=="":
               streamend = f
            streamend2 = streamend+lkw
            if t[streamend: streamend2]!="endstream":
                raise ValueError, "stream length doesn't end at endstream keyword %s %s" % (
                                  streamend, `t[streamend:streamend+20]`)
        self.cursor = streamend+9
        self.cursorcheck()
        return ("stream", (dict, streamcontent))

    dispatch_dictionary["s"] = getstream

    def getdictionary(self):
        t = self.text; c = self.cursor
        (dummy, stack) = self.getarray(startmark="<<", endmark=">>")
        if len(stack)%2:
            #print stack
            raise ValueError, "odd number of elements in dictionary contents %s %s" %(c, t[c:c+20])
        stackcopy = list(stack)
        D = {}
        while stackcopy:
            [k,v] = stackcopy[:2]
            del stackcopy[:2]
            (kmark, kname) = k
            # strip out the marker, key assumed to always be a name
            if kmark!="name":
                raise ValueError, "dictionary key must be name %s %s %s" % (kmark, c, `t[c:c+20]`)
            D[kname] = v
        return ("dictionary", D)

    # getdictionary called by dispatchanglebracket below

    def skiptokeyword(self, keyword):
        # does this handle (missing) whitespace correctly always?
        done = None
        sc = self.splitcursor
        st = self.splittext
        lk = len(keyword)
        while done is None:
            sc = sc+1
            ss = st[sc]
            if ss[:lk]==keyword: done = 1
        self.splitcursor = sc = sc+1
        self.cursor = self.splitindices[sc]

    def skiptostring(self, st):
        #print "skiptostring", st
        t = self.text
        c = self.cursor
        c2 = string.find(t, st, c)
        if c2<0: raise ValueError, "can't find %s past %s" % (repr(st), c)
        self.cursor = c2+len(st) # actually skip PAST!
        self.cursorcheck()

    def getarray(self, startmark="[", endmark="]", verbose=0):
        start = c = self.cursor
        text = self.text
        if startmark and text[c:c+len(startmark)]!=startmark:
            raise ValueError, "getname not starting at %s %s %s"%(`startmark`, c, `text[c:c+20]`)
        if startmark:
            self.cursor = c+len(startmark)
        save_stack = self.stack
        stack = self.stack = []
        done = None
        ltext = len(text)
        lendmark = 1
        if endmark:
            lendmark = len(endmark)
        while done is None:
            self.skipwhitespace()
            c = self.cursor
            if verbose: print "scanning array", repr(text[c:c+20])
            if c>ltext:
                if endmark:
                    raise ValueError, "unclosed array at %s %s" % (start, `text[start:start+20]`)
                self.cursor = ltext
                done = 1
            elif text[c:c+lendmark]==endmark:
                if verbose: print repr(endmark), "endmark found"
                self.cursor = c+lendmark
                done = 1
            else:
                # dispatch
                dd = self.dispatch_dictionary
                firstchr = text[c]
                unboundparsemethod = dd.get(firstchr, None)
                # unboundparsemethod
                if unboundparsemethod is None:
                    linecount = len(string.split(text[:c], "\n"))
                    raise ValueError, "can't parse at line %s %s %s%s" % (
                        linecount,  firstchr,repr(text[c-50:c]),repr(text[c:c+50]),)
                parsepair = unboundparsemethod(self)
                stack.append(parsepair)
                if verbose:
                    print parsepair[0], "at", len(stack)
        sequence = self.stack
        self.stack = save_stack
        self.cursorcheck()
        return ("array", sequence)

    dispatch_dictionary["["] = getarray

    def getnumber(self, text=None, cursor=None):
        """get a number at cursor or error"""
        # really simple implementation.. probably should use re instead...
        if text is None:
            text = self.text
        if cursor is None:
            cursor = self.cursor
        startnum = endnum = cursor
        ltext = len(text)
        while endnum<ltext and text[endnum] in "+-0123456789.":
            endnum = endnum+1
        if startnum==endnum:
            raise ValueError, "bad call to getnumber at %s %s" % (cursor, `text[cursor:cursor+20]`)
        numtext = text[startnum:endnum]
        from string import atoi, atof
        try:
            num = atoi(numtext)
        except:
            num = atof(numtext)
        self.cursor = endnum
        self.cursorcheck()
        return ("number", num)

    for c in "+-0123456789.":
        dispatch_dictionary[c] = getnumber

    def getstring(self, text=None, cursor=None, startmark="(", closemark=")"):
        """find a string in parens at cursor, or error
           NOTE: does not mess with content (yet) XXX???"""
        # start/closemark must have length 1 in this implementation!!!
        #print "getstring at", self.cursor
        ### in IRS form f1000.pdf I find the string (Form 1000 (Rev. August 1997))
        ### SO: if you find an open ( not preceded by \, must look for matching unescaped )
        ### I guess (this is undocumented, or at least poorly documented!)
        from string import find, replace
        if cursor is None:
            c = self.cursor
        else:
            c = cursor
        if text is None:
            text = self.text
        if text[c]!=startmark:
            raise ValueError, "getstring not starting at %s %s %s"%(`startmark`, c, `text[c:c+20]`)
        #cursor = c
        cursor = startstring = c+1
        # look for close paren not preceded by '\'
        #closed = None
        depth = 1
        while depth>0:
            #print "at", repr(text[cursor:cursor+40]), depth
            closed = 0
            closestring = cursor
            # look for a close
            while not closed:
                closestring = find(text, closemark, closestring)
                if closestring<0:
                    raise ValueError, "unclosed %s at %s %s"%(`startmark`, c, `text[c:c+20]`)
                if text[closestring-1:closestring]!="\\" or text[closestring-2:closestring]=="\\\\":
                    closed = 1
                    depth = depth-1
                else:
                    closestring = closestring+1
            #print "closestring", closestring, `text[closestring: closestring+10]`
            #depth = depth-1
            # BUT increment for every unescaped ( before it!
            foundallopens = 0
            openstring = cursor
            while not foundallopens:
                openstring = find(text, startmark, openstring)
                if openstring<0 or openstring>=closestring:
                    foundallopens = 1
                elif text[openstring-1:openstring]!="\\" or text[openstring-2:openstring]=="\\\\":
                    #print "openstring", openstring, `text[openstring: openstring+10]`
                    depth = depth+1
                openstring = openstring+1 # keep looking until all found
            cursor = closestring
            if depth>0:
                cursor = cursor+1 # look to next
        s = text[startstring:closestring] # string left in PDF escaped format
        #print open("text.txt", "wb").write(s)
        self.cursor = closestring+1
        self.cursorcheck()
        #print "string", s
        return ("string", s)

    dispatch_dictionary["("] = getstring

    def getname(self, text=None, cursor=None):
        "find /Name at cursor or error.  Do convert the #30 encoded hex characters"
        # really stupid implementation for now (no point in optimizing [yet]
        startmark = "/"
        if cursor is None:
            c = self.cursor
        if text is None:
            text = self.text
        if text[c]!=startmark:
            raise ValueError, "getname not starting at %s %s %s"%(`startmark`, c, `text[c:c+20]`)
        startname = c+1
        endname = startname
        thischar = text[endname]
        while 0x21<=ord(thischar)<=0x7e and thischar not in "%()<>{}[]/":
            endname = endname+1
            if endname>len(text): break # should be an error?
            thischar = text[endname]
        if endname==startname:
            if verbose: print "empty name at %s %s" % (c, `text[c:c+20]`)
        s = text[startname: endname]
        # advance cursor
        self.cursor = endname
        self.cursorcheck()
        # look for hex chars #20
        index = 0
        from string import atoi
        while index<len(s): # len(s) may change!
            if s[index]=="#":
                hexchar = s[index+1: index+3]
                try:
                    hexord = atoi(hexchar, 16)
                    char = chr(hexord)
                    s = "%s%s%s" %(s[:index], char, s[index+3:])
                except:
                    # this wasn't 2 digits after the hash, so wasn't a number
                    # so let it through.
                    pass
            index = index+1
        return ("name", s)

    dispatch_dictionary["/"] = getname

    def gettext(self,text=None,cursor=None):
        """find text in <angle brackets> at cursor, or error, don't unhex the contents (yet)"""
        #if cursor is None: cursor = self.cursor
        #raise ValueError, "gettext %s" % repr(self.text[cursor: cursor+20])
        (dummy, s) = self.getstring(text=None, cursor=None, startmark="<", closemark=">")
        #print "found text", s
        return ("text", s)

    #dispatch_dictionary["<"] = gettext
    def dispatchanglebracket(self):
        c = self.cursor
        t = self.text
        if t[c]!="<":
            raise ValueError, "didn't find < at "+repr(text[c:c+20])
        if t[c:c+2]=="<<":
            return self.getdictionary()
        else:
            return self.gettext()

    dispatch_dictionary["<"] = dispatchanglebracket

    aligned = None

    def align(self):
        """find indices for starts of splitted text"""
        from string import find
        splittext = self.splittext
        text = self.text
        splitindices = self.splitindices = splittext[:]
        cursor = i = 0
        for st in splittext:
            cursor = string.find(text, st, cursor)
            splitindices[i] = cursor
            cursor = cursor + len(st)
            i = i+1
        splitindices.append(len(self.text))
        self.aligned = 1
        self.cursor = 0
        self.splitcursor = 0
        # invariant:
        #  self.splitindices[self.splitcursor]<=self.cursor<=self.splitindices[self.splitcursor+1]
        #  self.cursor<=len(self.text) only = when done

    def cursorcheck(self):
        "make sure self.cursor and self.splitcursor refer to same virtual point in document"
        # actually, splitcursor can refer to the end of last whitespace *before* cursor
        if not self.aligned:
            if verbose:
                print "warning: cannot cursorcheck before alignment"
        cursor = self.cursor
        splitcursor = self.splitcursor
        if splitcursor>=len(self.splittext): # eof case
             self.cursor = len(self.text)
             self.splitcursor = len(self.splittext)
             return
        splitindices = self.splitindices
        [thisindex,nextindex] = splitindices[splitcursor:splitcursor+2]
        if cursor>=nextindex:
            # adjust the splitcursor forward
            from bisect import bisect
            self.splitcursor = bisect(splitindices, cursor)-1
            return
        if cursor<thisindex:
            self.cursor = thisindex
            return

    def skipwhitespace(self, whitespace=string.whitespace):
        c0 = c = self.cursor
        text = self.text
        #print "skipwhite at", `text[c:c+20]`
        splitindices = self.splitindices
        aligned = self.aligned
        splitcursor = self.splitcursor
        done = None
        count = 1
        while done is None:
            #print "skipwhite", count
            count = count+1
            try:
                cc = text[c]
            except:
                raise ValueError, "skip white to eof starting at %s" % c0
            if cc=="%":
                # in comment: skip to eoln
                #print "comment at", `text[c:c+20]`
                from string import find
                c2 = find(text, "\n", c)
                if c2<=0:
                    c2 = find(text, "\r", c)
                if c2<=0:
                    raise ValueError, "comment not closed by newline at %s" % c
                c = c2+1
                while splitindices[splitcursor]<c:
                    splitcursor = splitcursor+1
                c = splitindices[splitcursor]
                #print "skipped to", `text[c:c+20]`
            # treat startxref...%%EOF as whitespace too
            elif cc=="s" and text[c:c+9]=="startxref":
                self.skiptostring("%%EOF")
                c = self.cursor
                splitcursor = self.splitcursor
                #print "skipped startxref..%%EOF from", c0, "to", c, "total", c-c0
            elif cc in whitespace:
                # skip to next nonwhite
                splitcursor = splitcursor+1
                c = splitindices[splitcursor]
            else:
                done = 1
        self.splitcursor = splitcursor
        self.cursor = c

def rewrite(frompdffile, topdffile):
    print "\n\n reformatting", frompdffile, "into", topdffile
    text = open_and_read(frompdffile)
    p = PDFParseContext(text, frompdffile)
    p.parse()
    (ind, catinfo) = p.catalog
    c = p.compilation
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    doc = pdfdoc.PDFDocument()
    doc.Catalog = catalog
    # put all elements into file in order found
    print "\n\n\n Now installing objects in document"
    c.sequence.reverse() # REVERSE SEQUENCE (TESTING)
    for name in c.sequence:
        ob = c.objects[name]
        print name, ob
        doc.Reference(ob, name)
    # now write the file
    print "\n\nnow writing", topdffile
    doc.SaveToFile(topdffile, None) # no canvas needed, no outline
    print "rewrite done"


def combinePdfsInMemory(list_of_pdfs):
    """Take a list of PDf files 'in memory' and combine.

    Each item in the list is a PDF file held in memory as
    a string. This will use PageCatcher techniques to make
    one PDF file.
    """
    buf = getStringIO()
    canv = canvas.Canvas(buf)
    for pdfContent in list_of_pdfs:
        copyPagesInMemory(pdfContent, canv)
    canv.save()
    return buf.getvalue()

def combinePdfs(combinedFileName, list_of_pdf_filenames):
    "Merge list of PDF files into one.  Uses lots of memory :-)"
    list_of_pdf_content = []
    for filename in list_of_pdf_filenames:
        stuff = open_and_read(filename)
        list_of_pdf_content.append(stuff)

    bigPdf = combinePdfsInMemory(list_of_pdf_content)
    open(combinedFileName, 'wb').write(bigPdf)

def copyPages(frompdffile, tocanvas, withoutline=1):
    """copy all pages from pdf file into the Reportlab canvas."""
    pdfFileContent = open_and_read(frompdffile)
    copyPagesInMemory(pdfFileContent, tocanvas, withoutline)

def copyPagesInMemory(pdfFileContent, tocanvas, withoutline=1):
    """copy all pages from pdf content into the Reportlab canvas"""
    #need a doc-specific prefix
    p = PDFParseContext(pdfFileContent, md5.md5(pdfFileContent).hexdigest())
    p.parse()
    (ind, catinfo) = p.catalog
    # this will add some extra gunk (non used catalog, for instance)
    c= p.compilation
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    # copy in all elements of the file
    docObject = tocanvas._doc
    from reportlab.pdfbase.pdfdoc import PDFZCompress, PDFStream
    for name in c.sequence:
        ob = c.objects[name]
        if isinstance(ob, PDFStream):
            # compress all streams
            ob.filters = [PDFZCompress]
        #print name, ob
        docObject.Reference(ob, name)
    # now add the pages
    c.populatePageList()
    pages = docObject.Pages
    pagesref = docObject.Reference(pages)
    for pageName in c.pageList:
        pageObject = c.objects[pageName]
        pageObject["Parent"] = pagesref
        pages.addPage(pageObject)
    # add outline entries, if requested
    if withoutline:
        #print "catalog", catalog
        #print "catinfo", catinfo, ind
        catobj = c.resolve(catalog)
        #print "catobj", catobj
        if catobj.dict.has_key("Outlines"):
            outlines = c.resolve(catobj.dict["Outlines"])
            #print "outlines", outlines
            #print "outlines dict", outlines.dict
            count = outlines.dict["Count"]
            if count:
                first = outlines.dict["First"]
                installOutlines(first, c, tocanvas)
                #stop
        for otherattribute in string.split("""Dests Threads AcroForm Names OpenActions URI
                ViewerPreferences PageLabels PageLayout JavaScript StructTreeRoot SpiderInfo"""):
            if catobj.dict.has_key(otherattribute):
                print otherattribute
                value = catobj.dict[otherattribute]
                setattr(docObject.Catalog, otherattribute, value)
    #they might still be using the canvas, update its page number
    tocanvas._pageNumber = len(tocanvas._doc.Pages.pages) + 1
    return tocanvas

def installOutlines(node, compilation, canvas, level=0, seen=None):
    c = compilation
    node = c.resolve(node)
    if seen is None: seen={}
    if seen.has_key(node):
        raise ValueError, "infinite loop traversing outlines"
    seen[node] = 1
    dict = node.dict
    title = dict["Title"].s # get the title string
    #print "   "*level, title
    if dict.has_key("A"):
        #print "warning: pagecatcher does not process actions in outlines"
        return
    count = 0
    if dict.has_key("Count"):
        count = dict["Count"]
    if dict.has_key("Dest"):
        Dest = dict["Dest"]
        name = "destination"+repr(id(Dest))
        canvas._destinations[name] = Dest
        #print "name, dest", (name, Dest)
        closed = 0
        if count<0: closed = 1
        canvas.addOutlineEntry(title, name, level, closed)
    # recursively do the children
    if dict.has_key("First"):
        childnode = dict["First"]
        installOutlines(childnode, compilation, canvas, level+1, seen)
    # recursively do the syblings
    if dict.has_key("Next"):
        siblingnode = dict["Next"]
        installOutlines(siblingnode, compilation, canvas, level, seen)

# specialized info structures for recording product information

class PDFInfoL(pdfdoc.PDFInfo):
    """PDF documents can have basic information embedded, viewable from
    File | Document Info in Acrobat Reader.  If this is wrong, you get
    Postscript errors while printing, even though it does not print."""
    def __init__(self, oldinfo, user, number):
        self.title = oldinfo.title
        self.author = oldinfo.author
        self.subject = oldinfo.subject
        self.producer = "(%s,%s) PageCatcher http://www.reportlab.com" % (repr(user), repr(number))
        #now = time.localtime(time.time())
        #self.datestr = '%04d%02d%02d%02d%02d%02d' % tuple(now[0:6])

    def format(self, document):
        D = {}
        PDFString = pdfdoc.PDFString
        D["Title"] = PDFString(self.title)
        D["Author"] = PDFString(self.author)
        D["CreationDate"] = pdfdoc.PDFDate(invariant=invariant)
        D["Producer"] = PDFString(invariant and "PageCatcher http://www.reportlab.com" or self.producer)
        D["Subject"] = PDFString(self.subject)
        PD = pdfdoc.PDFDictionary(D)
        return PD.format(document)

class PDFInfoU(pdfdoc.PDFInfo):
    def format(self, document):
        D = {}
        PDFString = pdfdoc.PDFString
        D["Title"] = PDFString("PageCatcher Evaluation Test Document")
        D["Author"] = PDFString("PageCatcher Evaluator")
        D["CreationDate"] = pdfdoc.PDFDate(invariant=invariant)
        D["Producer"] = PDFString("Evaluation copy of PageCatcher http://www.reportlab.com")
        D["Subject"] = PDFString(
            "This document was created with an evaluation version of ReportLab PageCatcher.\n"
            "Please contact ReportLab to obtain a license for production use.")
        PD = pdfdoc.PDFDictionary(D)
        return PD.format(document)

def storeForms(frompdffile, storagefile,
               pagenumbers=None, prefix="PageForms", all=None,
               verbose=0, BBoxes=0,
               fformname=None):
    "returns names of forms stored"
    if pagenumbers is None:
        pagenumbers = [0] # default to first page
    if verbose: print "\n\n placing pages %s from %s into %s" % (pagenumbers, frompdffile, storagefile)

    pdfContent = open_and_read(frompdffile)
    result = storeFormsInMemory(pdfContent, pagenumbers, prefix, all,
                verbose, BBoxes, #extractext=0,
                fformname=fformname)

    if BBoxes:
        #return value is a dict of bounding boxes;
        #preserve the old API
        (bboxInfoDict, pickledData) = result
        return bboxInfoDict
    else:
        (names, pickledData) = result
        if hasattr(storagefile, 'write'):
            storagefile.write(pickledData)
        else:
            if verbose: print "storing to", storagefile
            open(storagefile, 'wb').write(pickledData)
        return names

def storeFormsInMemory(pdfContent, pagenumbers=None, prefix="PageForms", all=None,
               verbose=0, BBoxes=0, extractText=0, fformname=None):
    """Stores Forms without touching disk.  You pass in the PDF file
    contents as a string.  It normally returns a 2-tuple with the list of form names
    (for backwards compatibility) and the content of the storage file, which
    is essentially a pickle of forms objects. If you select the BBoxes option,
    the first element is a dictionary mapping form names to page bounding boxes,
    so you can extract the page sizes."""

    c, nameToObj = _storeFormsInDict(pdfContent, pagenumbers, prefix, all, verbose, fformname=fformname)
    pickledData = cPickle.dumps(nameToObj, 1)
    formnames = nameToObj[None]
    result = (formnames, pickledData)
    if BBoxes:
        # if the user requested bboxes then return dictionary name-->bbox
        bboxInfo = {}
        for formname in formnames:
            xobjname = pdfdoc.xObjectName(formname)
            form = nameToObj[ xobjname ]
            #print dir(firstform)
            dict = form.dictionary.dict
            #print dict
            bbox = dict["BBox"].sequence
            #print bbox
            bboxInfo[formname] = bbox
        result = (bboxInfo, pickledData)
    elif extractText:
        result = c.plainText
    return result

def storeFormsInDict(pdfContent, pagenumbers=None, prefix="PageForms", all=None, verbose=0, fformname=None):
    """Stores Forms in a dict.  You pass in the PDF file
    contents as a string.  It returns a dict of form objects.
    To get form ie page names use the returnedValue[None]
    """
    return _storeFormsInDict(pdfContent, pagenumbers=pageNumbers, prefix=prefix, all=all,
                verbose=verbose, fformname=fformname)[0]

def _storeFormsInDict(pdfContent, pagenumbers=None, prefix="PageForms", all=None, verbose=0, fformname=None):
    # no security in this one... always store forms the same.
    p = PDFParseContext(pdfContent, prefix=prefix)
    p.parse()
    (ind, catinfo) = p.catalog
    c = p.compilation
    if p.encrypt: getPdfEncrypt()
    c.sanitizePages()
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    c.populatePageList()
    return c, c.pagesAsForm(pagenumbers, prefix, all=all,fformname=fformname)

def restoreForms(storagefilename, canv, verbose=0, formnames=None,
                 allowDuplicates=0):
    if verbose: print "now restoring forms"
    content = open_and_read(storagefilename)
    return restoreFormsInMemory(content, canv, verbose=verbose, formnames=formnames, allowDuplicates=allowDuplicates)

def restoreFormsInMemory(content, canv, verbose=0, formnames=None, allowDuplicates=0):
    """Adds the named forms into the canvas and underlying PDF, with
    all the stuff they reference.
    If one attemtps to add forms twice, normally a ValueError is raised.
    If allowDuplicates == 1, it will test if any of the form names are present
    up front, and succeed silently (with few cycles) if they are.  this
    makes it safe to call it several times with the same form, which can
    be convenient.
    """
    if verbose: print "now restoring forms"
    return restoreFormsFromDict(cPickle.loads(content), canv, verbose=verbose, formnames=formnames, allowDuplicates=allowDuplicates)

def restoreFormsFromDict(nameToObj, canv, verbose=0, formnames=None, allowDuplicates=0):
    '''restores the forms from the dict nameToObj'''
    if nameToObj.has_key(0):
        uses = nameToObj[0]
    else:
        uses = {}
    # get the form names
    formNamesAll = nameToObj[None]
    if formnames is None:
        formnames = formNamesAll
    else:
        for fn in formnames:
            if fn not in formNamesAll:
                raise ValueError, "no such form %s in %s, only found %s" % (
                    repr(fn), repr(storagefilename), formNamesAll)
    if verbose: print "formNames are", formnames


    #new mode allows us to safely call this repeated times, with
    #no effects - assuming form name is unique and reliable.
    if allowDuplicates:
        if verbose: print 'up-front check forms are already present'
        doc = canv._doc
        discard = []
        for name in formnames:
            if doc.idToObject.has_key(pdfdoc.xObjectName(name)):
                discard.append(name)
                if verbose: print '%s already defined, not processing' % name
        for d in discard:
            formnames.remove(d)
        if formnames == []:
            if verbose: print 'no new forms, returning'
            return

    # now if the user has no license mess up the forms.
    # magic string:
    SS = " PageCatcher is a licensed product of ReportLab http://www.reportlab.com. "
    ok = 0
    nagPlus = "Contact www.reportlab.com to purchase"
    try:
        sys.path.insert(0, ".")
        if verbose:
            print "checking license"
        try:
            import PageCatcherLicense
        except:
            if verbose:
                print "error0", sys.exc_type, sys.exc_value
                print sys.path
            raise
        if verbose:
            print "from", PageCatcherLicense.__file__
        number = str(PageCatcherLicense.number)
        user = PageCatcherLicense.user
        if len(number)<11:
            if verbose: print "error1"
            raise ValueError
        key = PageCatcherLicense.key
        expire = PageCatcherLicense.expire
        misc = PageCatcherLicense.misc
        try:
            nagPlus = misc[0]
        except:
            pass
        Sn = SS+number+SS+user+SS+repr(expire)+SS+repr(misc)
        #print "commentme sn=", repr(Sn)
        from md5 import new
        code = new(Sn).digest()
        #print "commentme code=", type(code)
        #print "commentme", map(ord, code)
        code85 = pdfutils._AsciiBase85Encode(code)
        if code85!=key:
            if verbose:
                if len(code85)!=len(key) and verbose:
                    print "error 2.1", (code85[len(key):], key[len(code85):]), len(key), len(code85)
                for i in range(max(len(code85),len(key))):
                    if code85[i:i+1]!=key[i:i+1]:
                        print "error 2.2", i, (code85[i:i+1], key[i:i+1])
            if verbose: print "error2", (code85,key), (type(code85), type(key))
            raise "security fails!"
        import time
        if expire is not None and expire<time.time():
            if verbose:
                print "license expired %s" % repr(time.localtime(expire))
            raise "license expired %s" % repr(time.localtime(expire))
    except:
        #import sys
        #print sys.exc_type, sys.exc_value
        canv._doc.info = PDFInfoU() # mark file as unlicensed
        if verbose:
            print "license check failed"
        # license failure: mess up the forms...
        # no comments in the output file
        pdfdoc.DoComments = 0
        #pdfdoc.DoComments = 1 # temp
        # encrypt all streams, make it hard to figure out how the encoding works.
        #canv._doc.defaultStreamFilters = [pdfdoc.PDFZCompress]
        #canv._doc.defaultStreamFilters = None #temp
        count = 0
        BBoxes = {}
        newNames = {}
        for formName in formnames:
            count = count+1
            # create "wrapper" forms
            h = repr(hash(formName))
            # make a unique new name that's hard to understand
            newName = "x%s%s%s" % (h[0:3], count, h[3:6])
            newNames[formName] = newName
            oldinternalname = pdfdoc.xObjectName(formName)
            # use the cryptic name for the form
            theform = nameToObj[ oldinternalname ]
            thenewinternalname = pdfdoc.xObjectName(newName)
            #print "new", repr(thenewinternalname)
            nameToObj[ thenewinternalname ] = theform
            uses[ thenewinternalname ] = uses.get(oldinternalname, [])
            # print thenewinternalname
            BBoxes[ newName ] = theform.dictionary.dict["BBox"].sequence
            del nameToObj[oldinternalname]
##            # add marked form in place of the old one.
##            canv.beginForm(formName) # moved below
##            canv.doForm(newName)
##            canv.drawString(20, 20, "This page used an unlicensed copy of PageCatcher, http://www.reportlab.com.")
##            canv.endForm()
    else:
        # license succeeded. put the license number into the info
        ok = 1
        canv._doc.info = PDFInfoL(canv._doc.info, user, number) # mark file as unlicensed
        if verbose: print "license checks out", number, user
    doc = canv._doc
    # always compress streams
    if not ok or COMPACT:
        if verbose: print 'ok=',ok,' COMPACT=',COMPACT
        canv._doc.defaultStreamFilters = [pdfdoc.PDFZCompress]
    #print "\n\n\n Now installing objects in document"
    keys = nameToObj.keys()
    keys.sort()
    #print "keys", keys
    #print "uses", nameToObj[0]
    # install only objects used by these forms
    required = []
    if 0 and nameToObj.has_key(0):
        uses = nameToObj[0]
        #print uses.keys()
        for fn in formnames:
            efn = pdfdoc.xObjectName(fn)
            required.extend(uses[efn])
        installed = {}
        while required:
            require = required[-1]
            del required[-1]
            if not installed.has_key(require):
                if nameToObj.has_key(require):
                    ob = nameToObj[require]
                    doc.Reference(ob, require)
                    required.extend(uses[require])
                    #print 'installed', ob
                else:
                    #print 'CAN\'T FIND OBJECT', require
                    pass # if it can't be found it might be a "forbidden" object
                installed[require] = 1
    else:
        # if its using an old data file just install everything
        for name in keys:
            #print "installing", name
            ob = nameToObj[name]
            if name:
                #print name, ob
                doc.Reference(ob, name)
    # add form name translations
    if not ok:
        for (formName, newName) in newNames.items():
            # add marked form in place of the old one.
            [lx, ly, ux, uy] = BBoxes[newName]
            canv.beginForm(formName, lx, ly, ux, uy)
            canv.doForm(newName)
            canv.drawString(20, 20, "Evaluation or expired copy of ReportLab's PageCatcher. " + nagPlus)
            canv.endForm()
            thenewinternalname = pdfdoc.xObjectName(newName)
            ob = nameToObj[thenewinternalname]
            doc.Reference(ob, thenewinternalname)
    else:
        for fn in formnames:
            efn = pdfdoc.xObjectName(fn)
            ob = nameToObj[efn]
            doc.Reference(ob, efn)
    return formnames

def page1(frompdffile, topdffile):
    "\n\n placing page 1 of", frompdffile, "in", topdffile
    storagefile = "dump.pickle"
    prefix = "myPrefix"
    names = storeForms(frompdffile, storagefile, [1], prefix)
    canv = canvas.Canvas(topdffile)
    names = restoreForms(storagefile, canv, verbose=verbose, formnames=[prefix+"1"])
    # for testing, put a form in a form
    canv.beginForm("testform")
    canv.doForm(prefix+"1")
    canv.drawString(50, 500, "HELLO "*5)
    canv.endForm()
    canv.doForm("testform")
    print "\n\nnow writing", topdffile
    canv.save()
    print "page1 done"

def do_test_grids(storagefile, testfile):
    print "making test grid from", storagefile, "into", testfile
    canv = canvas.Canvas(testfile, pageCompression=0)
    names = restoreForms(storagefile, canv, verbose=verbose)
    for name in names:
        print "form", name
        canv.doForm(name)
        canv.setFont("Helvetica", 6)
        from reportlab.lib.units import cm
        from reportlab.lib.colors import green, cyan
        canv.setStrokeColor(cyan)
        canv.setFillColor(green)
        lineend = cm*100 # too many cm, but what the...
        textoffset = 1.2*cm
        for count in range(100):
            y = count*cm
            canv.line(0, y, lineend, y)
            canv.drawString(textoffset, y, "y=%scm" % count)
        canv.rotate(90)
        for count in range(100):
            x = count*cm
            ty = -x
            canv.line(0, ty, lineend, ty)
            canv.drawString(textoffset, ty, "x=%scm" % count)
        canv.showPage()
    canv.save()
    print "wrote test file", testfile

def fourPage(storagefile, testfile, scalefactor = 0.5):
    print "placing forms from", storagefile, "into", testfile, "four to a page"
    from reportlab.pdfgen import canvas
    canv = canvas.Canvas(testfile)
    (width, height) = canv._pagesize
    names = restoreForms(storagefile, canv, verbose=verbose)
    while names:
        for (xoff, yoff) in [ (0,1), (1,1), (0,0), (1,0) ]:
            thisname = names[0]
            print thisname,
            canv.saveState()
            (x,y) = (xoff*width/2.0, yoff*height/2.0)
            canv.translate(x,y)
            canv.scale(scalefactor, scalefactor)
            canv.doForm(thisname)
            canv.restoreState()
            del names[0]
            if not names: break
        canv.showPage()
        print
    canv.save()
    print "wrote", testfile

def cb_annotate(out, pdfin, cb=None):
    '''split out this callback version so others can get at the annotation themselves'''
    from rlextra.utils.proxyfile import MemFile
    storage = MemFile('')
    n = storeForms(pdfin, storage, [0], "annotate", verbose=verbose)
    canv = canvas.Canvas(out)
    n = restoreForms(storage, canv, verbose=verbose)
    canv.doForm(n[0])
    if cb: cb(canv)
    canv.save()
    print "wrote", out

def annotate(out, pdfin, x,y, font, fontSize, text):
    def _cb(canv,x=x,y=y,font=font,fontSize=fontSize,text=text):
        canv.setFont(font, fontSize)
        canv.drawString(x,y,text)
    cb_annotate(out, pdfin, cb=_cb)

def test():
    import sys
    sstdout = sys.stdout
    fn = "dump.txt"
    #sys.stdout = open(fn, "w")
    sourcefile = "rfp_sample.pdf"
    page1(sourcefile, "onepage.pdf")
    #rewrite("translation1.pdf", "translation.pdf")
    sys.stdout = sstdout
    print "wrote", fn


#if __name__=="__main__": test()
#stop

script_usage = """
PAGECATCHER SCRIPT PREPROCESSOR USAGE:

In script mode the page catcher prepares the contents of one or more pages of
a PDF file for use in other PDF files.

Line mode usage:

% PageCatcher makeforms pdffile [-s storagefile] [-p prefix ] [-f formname]
                                [--test pdftestfile] [--all] [pagenumber]*

  Capture the pages from pdffile and place them in storagefile for later use.
  If the test option is used then the captured pages are reimported and placed
  in the test file, overlayed with a centimeter grid.

  Note! Pagenumbers start at 0 (zero)
  (with no necessary relation to the pagenumber shown by a PDF viewer such as Acroread).
  If the pagenumbers are omitted the first page of the document will be made into
  a form.

  The "form names" for the forms derived from the pages will be prefix0 for the first page,
  prefix1 for the second, prefix2 for the third, and so forth.
  If the -f formname option is used then the first form only will be called formname.

  For example

  % pageCatcher makeforms picture.pdf -s pic.data -p pict --test pictest.pdf 0 2

  extracts the first and third page from picture.pdf, archiving them in pic.data
  for later use, giving them the form names pict0 and pict2, respectively.
  The test file pictest.pdf will display the captured forms overlayed with a
  centimeter grid.

  If the storagefile is omitted it defaults to "storage.data".
  If the prefix is omitted it defaults to PF (for "page form").
  If the --all option is used then all pages of the document are captured.

USING THE DERIVED STORAGE FILES:

  To use the extracted forms later from a Python program proceed as follows:

    canvas = reportlab.pdfgen.Canvas(...)
    ...
    import pageCatcher
    # import pict0 and pict2 stored in pic.data by the command line above, put in canvas.
    storagefilename = "pic.data"
    formnames = pageCatcher.restoreForms(storageFileName, canvas)
    ...
    canvas.doForm(formnames[0]) # use the first form (as a backdrop for a page, for example).
    # or equivalently: canvas.doForm("pict0")
    ...
    canvas.save()

The restoreForms(filename, canvas) operation extracts the forms previously
stored in filename and installs them for later use in the pdfgen canvas object.

This is a licensed product.  You may use this module for evaluation purposes
but the forms will be visibly marked as "evaluation use of PageCatcher" and the
document will be interally annotated as a test document.

Licensed use of PageCatcher also results in the document being invisibly marked
with the license information.

DEMO SCRIPT MODES:

There are several built in demo modes:

 % pageCatcher.py help
     Print this message.

 % pageCatcher.py note [pdffile]
     This mode places a text string over the first page of pdffile,
     storing the result in "annotated.pdf".  The font, size and
     string are read interactively from the console.

 % pageCatcher.py 4page pdffile [--scale scalefactor]
                     [--output pdfoutputfilename] [-s storagefilename]
     This mode rewrites the pages of pdffile with 4 pages of the input
     on each page of the output ("save the trees" mode).

 % pageCatcher.py pages storagefile
 print names of the pages.
"""

#if __name__=="__main__":
def scriptInterp():
    try:
        import sys
        argv = list(sys.argv)
        all = None
        if "--all" in argv:
            argv.remove("--all")
            all = 1
        parameters = {"-s": "storage.data",
                      "--scale": 0.5,
                      "-p": "PF",
                      "-f": None,
                      "--test": None,
                      "--output": None}
        for k in parameters.keys():
            if k in argv:
                index = argv.index(k)
                del argv[index]
                value = argv[index]
                del argv[index]
                parameters[k] = value
        known_modes = ["makeforms", "4page", "help", "note", "exec", "pages"]
        try:
            mode = argv[1]
        except:
            print "please provide a 'mode' argument"
            print "one of", known_modes
            return
        mode = string.lower(mode)
        if mode=="help":
            print script_usage
            return
        elif mode=="note":
            def sr(prompt):
                return string.strip(raw_input(prompt))
            print "Annotate the first page of a file (interactive mode only)"
            try:
                pdffile = argv[2]
            except:
                pdffile = sr("Annotate pdf file: ")
            else:
                print "file=", pdffile
            try:
                x = float(sr("at x location (default 100): "))
            except:
                x = 100
            try:
                y = float(sr("at y location (default 400): "))
            except:
                y = 400
            font = sr("font [default=Helvetica]: ")
            if not font: font = "Helvetica"
            try:
                fontsize = float(sr("font size (default 20): "))
            except:
                fontsize = 20
            text = sr("Text: ")
            if not text: text = "Example Annotation"
            print "Attempting annotation..."
            annotate("annotated.pdf", pdffile, x, y, font, fontsize, text)
        elif mode=="4page":
            try:
                pdffile = argv[2]
            except:
                print "4Page requires a pdf file argument. Please provide filename"
                return
            storagefile = parameters["-s"]
            scale = parameters["--scale"]
            output = parameters["--output"]
            if output is None: output = "4page.pdf"
            print "Formatting", pdffile, "as", output, "4 pages of input in each output page."
            print "Makeforms from", pdffile, "to", storagefile, "all pages"
            names = storeForms(pdffile, storagefile, [], "fourpage", 1, verbose=verbose)
            print "generating", output
            fourPage(storagefile, output, scale)
        elif mode=="makeforms":
            try:
                pdffile = argv[2]
            except:
                print "MakeForms requires a pdf file argument. Please provide filename"
                return
            storagefile = parameters["-s"]
            prefix = parameters["-p"]

            testfile = parameters["--test"]
            pagenumbers = argv[3:]
            if (pagenumbers and all):
                raise ValueError, \
                      "Cannot use '--all' parameter together with specified page numbers %s" % repr(pagenumbers)
            if not (pagenumbers or all):
                print "defaulting to first page"
                pagenumbers = [0]
            elif not all:
                pagenumbers = map(int, pagenumbers)
            else:
                pagenumbers = "all pages"
            print "Makeforms from", pdffile, "to", storagefile, "page numbers", pagenumbers
            names = storeForms(pdffile, storagefile, pagenumbers, prefix, all, verbose=verbose , fformname=parameters["-f"])
            if testfile is not None:
                do_test_grids(storagefile, testfile)
            print; print "created forms", names, "in", storagefile
        elif mode == 'exec':
            # execute a script
            try:
                scriptName = argv[2]
            except:
                print 'Exec requires a python script name as the second argument.'
                return
            scriptname = argv[2]
            import os
            if not os.path.isfile(scriptname):
                print 'Script "%s" not found!' % scriptname
                return
            if __name__=='__main__':
                execfile(scriptname)
            else:
                oldname = globals().get('__name__', None)
                globals()['__name__'] = '__main__'
                execfile(scriptname)
                del globals()['__name__']
                if oldname: globals()['__name__'] = oldname
        elif mode == 'pages':
            try:
                P = _listPages(argv[2])
                s = ''
                for p in P:
                    if ' ' in p: p = '"'+p+'"'
                    s = '%s %s' % (s,p)
                print s[1:]
            except:
                print 'pages requires a storage file name as the second argument'
                raise
        else:
            print "unknown mode %s, known modes are %s" %(mode, known_modes)
    except:
        print error_hint
        print "ERROR", sys.exc_type, sys.exc_value
        if RERAISE:
            raise

def _listPagesInMemory(s):
   import cPickle
   return cPickle.loads(s)[None]

def _listPages(fn):
   return _listPagesInMemory(open_and_read(fn))


error_hint = """
PageCatcher has terminated in an error condition.

There are several common situations where this can occur:
  1) The input parameters may be invalid (the pdf file doesn't exist,
     or the pdf file doesn't contain the page you requested, etcetera).
  2) The PDF file may be encrypted with a non null password.  If you have
     the owner password, save it without a password and try again.
  3) The page you requested contains active form elements, such as
     fillable fields or check buttons.  This version of pageCatcher does not
     correctly extract the graphics from these pages.
  4) The page you requested contains an array of streams one of which
     is encrypted using LZW encryption.  This version of pageCatcher does not
     correctly extract the graphics from these pages.
Often it is possible to "work around" problems such as 2 through 4 above
by making use of a "PDF printer emulator" such as Distiller.

If you have Distiller (or similar software) available, try this: open
the file in Acrobat or Acrobat Reader and "Print to PDF" using Distiller.
The resulting file will be visually identical to the original, and in all
cases we have observed it should be acceptible as a PageCatcher input.
"""
if __name__=="__main__": #NO RUNTESTS
    scriptInterp()
