#Copyright ReportLab Europe Ltd. 2000-2004
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/rml2pdf/rml2pdf.py
__version__=''' $Id: rml2pdf.py 14699 2004-09-21 14:40:45Z rgbecker $ '''
import string, reportlab, os, sys, time, re
version="1.0"
compatible_versions = ["0.2", "0.3", "1.0"]

SV = "Security Violation"

verbose = 0
from types import ListType, TupleType, StringType, InstanceType

from reportlab.platypus.flowables import Flowable, Image, Macro, PageBreak, Preformatted, Spacer, XBox, \
                         CondPageBreak, KeepTogether, TraceInfo, FailOnWrap, FailOnDraw, HRFlowable, \
                         PTOContainer
from reportlab.platypus.paragraph import Paragraph, cleanBlockQuotedText, ParaLines
from reportlab.platypus.paraparser import ParaFrag
from reportlab.platypus.tables import Table, TableStyle, CellStyle
from reportlab.platypus.frames import Frame
from reportlab.platypus.doctemplate import BaseDocTemplate, NextPageTemplate, PageTemplate, ActionFlowable, \
                         SimpleDocTemplate, FrameBreak, PageBegin
from reportlab.platypus.xpreformatted import XPreformatted
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle, PropertySet
from reportlab.lib.units import inch, cm
from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT
from reportlab.lib.utils import recursiveImport, recursiveSetAttr, recursiveGetAttr, getStringIO, open_and_read, flatten
from reportlab.pdfbase import pdfdoc, pdfutils
from reportlab.rl_config import defaultPageSize, defaultImageCaching
from reportlab.lib.colors import white, black
from reportlab.lib import colors

# hack to support both kinds of x objects we have defined
if not hasattr(pdfdoc,'xObjectName'): pdfdoc.xObjectName = pdfdoc.formName
try:
    __loader__
except:
    __loader__ = None

def _rml2pdf_locations(D=None):
    if not D: D = []
    elif type(D) not in (type(()),type([])): D = [D]
    try:
        p = __file__
    except:
        from rlextra import rml2pdf
        p = rml2pdf.__file__
    D.append(os.path.dirname(p))
    pyver = '%spython%s.zip%s' % (os.sep,''.join(sys.version.split()[0].split('.'))[:2],os.sep)
    i = p.lower().find(pyver)
    if i>0: D += [p[:i], os.path.join(p[:i],p[i+len(pyver):])]
    return D

class _FS_Checker:
    def __init__(self,dir):
        self._dir = dir

    def __call__(self,target):
        fn = os.path.abspath(os.path.join(self._dir,target))
        if os.path.isfile(fn):
            if sys.platform=='win32':
                return 'file:///'+fn.replace('\\','/')
            else:
                return 'file://'+fn
        return None

class _ZIP_Checker:
    def __init__(self,dir):
        self._dir = dir

    def __call__(self,target):
        try:
            return target, __loader__.get_data(os.path.join(self._dir,target))
        except:
            return None

def _rml_dtd_dirs(exe=None):
    D = []
    for d in _rml2pdf_locations():
        c = __loader__ and d.startswith(__loader__.archive) and _ZIP_Checker or _FS_Checker
        D.append(c(d))

    if exe:
        cwd = os.path.abspath(".")
        for i in 1,0:
            try:
                d = os.path.dirname(sys.argv[i])
                if not os.path.isabs(d): d=os.path.join(cwd,d)
                D.append(_FS_Checker(d))
            except:
                pass
    return D

class CanvasParameters:
    "container for canvas parameter storage (trivial at the moment)"
    pageSize = defaultPageSize
    imageCaching = defaultImageCaching

class DeferredInitialization:
    """used to delay initialization of flowables (for example paragraphs in a frame)
       until bind time, so the names get bound correctly, for example"""
    def __init__(self, klass, *args, **kw):
        self.klass = klass
        self.args = args
        self.kw = kw

    def initialize(self):
        return self.klass(*self.args, **self.kw)

    def wrap(self, *args):
        raise ValueError, "initialization deferred for %s %s %s %s" %(id(self), self.klass, map(str, self.args), self.kw)


ENTITY_SUBSTITUTIONS_DRAWSTRING = [
    ("&pound;", chr(163)),
    ("&amp;", "&"),
    ("&lt;", "<"),
    ("&gt;", ">"),
    ("&apos;", "'"),
    ("&quot;", "\""),
    ]

def unEscapeDrawString(s):
    for (old, new) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
        s = s.replace(old, new).replace(old.upper(), new)
    return s

ENTITY_SUBSTITUTIONS_PARA = [
    ("&pound;", chr(163)),
    ]

RMLError = "RMLError"


### first some lazy object trickery
### (to allow, eg, pageNumbers)

# XXXX even this will not give the right page number for a split paragraph!
# XXXX really need to use the xml tree uniformly!

### maybe these should go in another (or other) module(s)

### probably should also add lazy Table
class LazyStringForm:
    def __init__(self):
        raise ValueError, "virtual superclass"

class LazySetItem(LazyStringForm):
    """set the value when converted to a string!"""
    def __init__(self, dict, item, value):
        self.dict = dict
        self.item = item
        self.value = value
    def __str__(self):
        self.dict[self.item] = str(self.value)
        return ""

class LazyGetItem(LazyStringForm):
    def __init__(self, dict, item, default=None):
        self.dict = dict
        self.item = item
        self.default = default
    def __str__(self):
        if self.default is None:
            result = self.dict[self.item]
        else:
            result = self.dict.get(self.item, self.default)
        return str(result)


### expediency: table wrap seems to be buggy in the x dimension!

class xTable(Table):
    # pretend the table has 0 width!
    def wrap(self, aw, ah):
        (x,y) = Table.wrap(self, aw,ah)
        return (1,y)

# comment to fix
xTable = Table

class _LazyMixin:
    """don't do any initialization until wrap time! (allow lazy string conversion)"""
    def __init__(self, text, style, bulletText = None, frags=None):
        self._initialization_arguments = (text, style, bulletText, frags)
        self._initialized = 0
        #Curing excessive laziness...
        if hasattr(style, 'spaceBefore'):
            #print "before", style.spaceBefore
            self.spaceBefore = style.spaceBefore
            #self.spaceBefore = "GARBAGE"

    def __str__(self):
        raise ValueError, "%s instance has no string conversion method" % self.__class__

    def _Initialize(self):
        if not self._initialized:
            (text, style, bulletText, frags) =  self._initialization_arguments
            if text is not None:
                text = str(text)    # make sure text is a string
                text = quoteText(text)  # HACK STRAY AMPERSANDS
                text = text.strip() # discard extra whitespace
            # store it, (in case of rewrap
            self._initialization_arguments = (text, style, bulletText, frags)
            self.SUPER.__init__(self, text, style, bulletText, frags)
            self._initialized = 1

    def wrap(self, availWidth, availHeight):
        self._Initialize()
        return self.SUPER.wrap(self, availWidth, availHeight)

    def split(self,availWidth, availHeight):
        self._Initialize()
        return self.SUPER.split(self, availWidth, availHeight)

class LazyParagraph(_LazyMixin,Paragraph):
    SUPER=Paragraph

class LazyXPreformatted(_LazyMixin,XPreformatted):
    SUPER=XPreformatted

class LazyPreformatted(Preformatted):
    def __init__(self, text, style, bulletText = None, dedent=0):
        self.style = style
        self.Preformatted_initialization_arguments = (text, style, bulletText, dedent)
        #Curing excessive laziness...
        if hasattr(style, 'spaceBefore'):
            self.spaceBefore = style.spaceBefore

    def wrap(self, availWidth, availHeight):
        (text, style, bulletText, dedent) =  self.Preformatted_initialization_arguments
        # make sure text is a string
        text = str(text)
        Preformatted.__init__(self, text, style, bulletText, dedent)
        return Preformatted.wrap(self, availWidth, availHeight)

class LazyJoin(LazyStringForm):
    def __init__(self, seq, sep=''):
        self.seq = seq; self.sep = sep
    def __str__(self):
        return str(self.sep).join(map(str, self.seq)) # late string conversion

class LazyAttribute(LazyStringForm):
    def __init__(self, thing, attribute):
        self.thing = thing
        self.attribute = attribute
    def __str__(self):
        val = getattr(self.thing, self.attribute)
        return str(val)

class LazySubst(LazyStringForm):
    def __init__(self, format, data):
        self.format = format; self.data = data
    def __str__(self):
        return self.format % self.data

### other utilities
def readBool(text):
    if text.upper() in ("Y", "YES", "TRUE", "1"):
        return 1
    elif text.upper() in ("N", "NO", "FALSE", "0"):
        return 0
    else:
        raise RMLError, "true/false attribute has illegal value '%s'" % text

def readAlignment(text):
    from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
    up = text.upper()
    if up == 'LEFT':
        return TA_LEFT
    elif up == 'RIGHT':
        return TA_RIGHT
    elif up in ['CENTER', 'CENTRE']:
        return TA_CENTER
    elif up == 'JUSTIFY':
        return TA_JUSTIFY

lengthMatcher = re.compile(r'\s*([-+]?[\d.]+)\s*(\S*)')
def readLength(text, percentTotal = None, context=None):
    """
    Read a dimension measurement: accept "3in", "5cm", "72 pt" and so on.

    If percentTotal is specified then percentages can also be specified
    (e.g. "5%") and function will return number/100 * percentTotal
    or if percentTotal is specified as '%' we get back the original string.
    """
    if context: text = context.lengths.get(text,text)

    match = lengthMatcher.match(text)
    if not match:
        if text.strip()=='*' and percentTotal=='%': return text
        raise ValueError, "invalid length attribute '%s'" % text
    number = float(match.group(1))
    units = match.group(2).lower()
    if not units:
        units = 'pt'
    if units == '%':
        if percentTotal is None:
            raise ValueError, "invalid length attribute '%s' - percent not allowed here" % text
        elif percentTotal=='%':
            return text
        multiplier = percentTotal / 100.0
    else:
        try:
            multiplier = {
                'in': 72,
                'cm': 28.3464566929,  #72/2.54; is this accurate?
                'mm': 2.83464566929,
                'pt': 1
            }[units]
        except KeyError:
            raise RMLError, "invalid length attribute '%s'" % text
    return number * multiplier

def lengthSequence(s, converter=readLength):
    """from "(2, 1)" or "2,1" return [2,1], for example"""
    s = s.strip()
    if s.startswith("(") and s.endswith(")"): s = s[1:-1]
    return map(converter, map(string.strip, s.split(",")))

def readColor(text):
    """Read color names or tuples, RGB or CMYK, and return a Color object."""
    if not text or text in ('none','None'): return None
    from reportlab.lib import colors
    if text[0] in string.letters:
        return colors.__dict__[text]
    tup = lengthSequence(text)

    msg = "Color tuple must have 3 (or 4) elements for RGB (or CMYK)."
    assert 3 <= len(tup) <= 4, msg
    msg = "Color tuple must have all elements <= 1.0."
    for i in range(len(tup)):
        assert tup[i] <= 1.0, msg

    if len(tup) == 3:
        colClass = colors.Color
    elif len(tup) == 4:
        colClass = colors.CMYKColor
    return colClass(*tup)

def _getColor(colorName):
    try:
        return readColor(colorName)
    except:
        return reportlab.lib.colors.toColor(colorName)

def getColor(colorName,context):
    try:
        return context.colors[colorName]
    except KeyError:
        return _getColor(colorName)

def readAlign(s):
    s=s.lower()
    if s in ('center','centre','left','right','decimal'): return s
    raise ValueError('Illegal align attribute "%s"' % s)

def readAlignU(s): return readAlign(s).upper()

def readvAlign(s):
    s=s.lower()
    if s in ('top','middle','bottom'): return s
    raise ValueError('Illegal vAlign attribute "%s"' % s)

def readvAlignU(s): return readvAlign(s).upper()

def readString(s): return s #identity

# the default aliases
DEFAULT_ALIASES = {
    "h1.defaultStyle": "style.Heading1",
    "h2.defaultStyle": "style.Heading2",
    "h3.defaultStyle": "style.Heading3",
    "title.defaultStyle": "style.Title",
    "para.defaultStyle": "style.Normal",
    "pre.defaultStyle": "style.Code",
    "li.defaultStyle": "style.Definition"
    }

styles = getSampleStyleSheet()

class NamesArchive:
    "like a dictionary, but keeps record of defaults and updates to remember when to do a multipass build"
    # None is a forbidden value
    def __init__(self, maxResets=None):
        self.archive = {}
        self.maxResets = maxResets
        self.resetCount = -1 # for bootstrapping only
        self.touched = 1 # for bootstrapping only
        self.reset(0)
    def reset(self, abortOnNoProgress=1):
        if not self.touched and abortOnNoProgress:
            raise ValueError, "no progress in resolving %s\n (count=%s)" % (repr(self.defaultEncountered.keys()), self.resetCount)
        self.touched = {}
        x = self.resetCount = self.resetCount+1
        if self.maxResets is not None and self.maxResets<=x:
            raise ValueError, "too many resets %s resolving %s" % (x, repr(self.defaultEncountered.keys()))
        self.defaultEncountered = {}
    def complete(self):
        if self.defaultEncountered:
            return 0
        else:
            return 1
    def __setitem__(self, key, value):
        if value is None:
            raise ValueError, "None is forbidden for names archive setitem"
        a = self.archive
        if not a.has_key(key):
            self.touched[key] = value
        a[key] = value
    def get(self, key, default=None):
        a = self.archive
        if a.has_key(key):
            return a[key]
        else:
            self.defaultEncountered[key] = default
            return default
    def __getitem__(self, key):
        test = self.get(key)
        if test is None:
            raise KeyError, "no entry  for %s" % repr(key)
        return test


class RMLContext:
    """Allow for aliases and redirection of aliases
    should colors be treated specially (like now)?
    eventually this should grow to include shadowing, etc..."""

    filename = None # should be set
    doc = None # the document being constructed...

    def __init__(self, namesarchive=None, permitEvaluations=1):
        self.permitEvaluations = permitEvaluations
        if namesarchive is None:
            namesarchive = NamesArchive()
        self.imageFileMap = {}
        self.pageDrawings = []
        self.canvasInfo = CanvasParameters()
        self.initialize = []
        self.document = None
        self.story = []
        self.compression = None
        self.invariant = None
        namedColors = reportlab.lib.colors.getAllNamedColors()
        self.colors = namedColors.copy() # for paranoia purposes...
        self.lengths = {}
        names = self.names = namesarchive # formerly {}
        #print dir(styles)
        #print styles.byAlias
        #print styles.byName.keys()
        for n in styles.byName.keys():
            s = styles[n]
            names["style."+n] = s
            # change this?
            names[n] = s
        for n in styles.byAlias.keys():
            s = styles[n]
            names["style." +n] = s
            names[n] = s
        self.aliases = {}
        # an example default (ultimatel
        #self.aliases["h1.defaultStyle"] = "style.Heading1"
        for k in DEFAULT_ALIASES.keys():
            self.alias(k, DEFAULT_ALIASES[k])
    def destroy(self):
        "this should (I hope!) break all possible cycles leading from the context"
        d = self.__dict__
        for k in d.keys():
            del d[k]
    def prepare(self):
        # execute delayed actions stored in initialize
        i = self.initialize
        self.initialize = []
        L = []
        for c in i:
            # converting to string will set off delayed actions
            L.append(str(c))
    def __setitem__(self, item, value):
        if value is None:
            raise ValueError, "None is forbidden for context setitem"
        self.names[item] = value
    def alias(self, alias, name):
        self.aliases[alias] = name
    def get(self, item, default=None):
        visited = {}
        aliases = self.aliases
        names = self.names
        item1 = item
        while 1:
            if visited.has_key(item1):
                raise ValueError, "infinite alias loop %s" % repr((item, item1))
            visited[item1] = 1
            # always prefer an alias... (?)
            if not aliases.has_key(item1):
                test = names.get(item1, default)
                if test is None:
                    raise KeyError, "%s not found in context" % repr(item1)
                return test
            else:
                item1 = aliases[item1]
    def __getitem__(self, item):
        return self.get(item)

def joinflatten(e):
    te = type(e)
    if te is StringType:
        return e
    if te in (TupleType, ListType):
        stre = map(joinflatten, e)
        return LazyJoin(stre)
    return e # str conversion handled late by LazyJoin

class MapController:
    """Top level controller"""

    def __init__(self):
        self.nodemappers = {None: None} # default

    def __setitem__(self, item, value):
        self.nodemappers[item] = value

    def __getitem__(self, item):
        return self.nodemappers[item]

    def process(self, text, context, parsed=0):
        """process (possibly preparsed) xml text"""
        # the real result of the process are the side effects to context
        if not parsed:
            from reportlab.lib.rparsexml import parsexml
            elts = parsexml(text,entityReplacer=None)
        else:
            elts = text
        elts = self.clean_elts(elts)
        output = self.processContent(elts, context, top_level=1)
        toutput = type(output)
        output = joinflatten(output)
        return output

    def processContent(self, elts, context, overrides=None, top_level=0):
        if overrides is None:
            overrides = {} # no overrides yet
        telts = type(elts)
        #print "processContent", telts
        if telts is StringType:
            return self.processString(elts, context)
        #print telts
        if telts is TupleType:
            return self.processTuple(elts, context, overrides)
        else:
            L = []
            for e in elts:
                typee = type(e)
                if typee is StringType:
                    e2 = self.processString(e, context)
                else:
                    if typee is not TupleType:
                        raise ValueError, "bad content type %s" % typee
                    e2 = self.processTuple(e, context, overrides)
                L.append(e2)
            return L

    def processTuple(self, e, context, overrides):
        tagname = e[0]
        #print "processTuple", tagname
        nodemappers = self.nodemappers
        if overrides:
            nodemappers = nodemappers.copy()
            # do overrides
            nodemappers.update(overrides)
        defaultmapper = nodemappers.get(None, None)
        processor = nodemappers.get(tagname, defaultmapper)
        if processor is None:
            raise NameError, "no processor for %s" % repr(tagname)
        e2 = processor.translate(e, self, context, overrides)
        return e2
##
##    def joinTranslatedContent(self, L):
##        strL = map(str, L)
##        return (string.join(strL, ""), L)

    def processString(self, elts, context):
        # do entity substitutions
        if "&" in elts:
            for (e, r) in ENTITY_SUBSTITUTIONS_PARA:
                elts = elts.replace(e,r)
        return elts # default, for now

    def clean_elts(self, elts):
        """optionally do stuff like dispose of all white content"""
        return elts

# the one and only controller
Controller = MapController()

class MapNode(MapController):

    def __init__(self, defaults=None):
        self.nodemappers = {}
        if defaults is None:
            defaults = {}
        self.defaults = defaults
        # transforms to apply per attribute to the content (eg make it have 2 digits and put a dollar in front)...
        self.transforms = {}
        self.init1() # for localization in subclasses

    def __str__(self):
        raise ValueError, "%s instance has no string conversion method" % self.__class__

    def init1(self):
        pass # only for use in subclasses

    def addTransform(self, name, t):
        self.transforms[name] = t
##
##    def transformContent(self, t):
##        self.transforms["__content__"] = t

    def translate(self, nodetuple, controller, context, overrides):
        # add own overrides if present
        # "translate", self
        if self.nodemappers:
            overrides = overrides.copy()
            overrides.update(self.nodemappers)
        (tagname, attdict, content, extra) = nodetuple
        #tagname = nodedict[0]
        #content = nodedict.get(1, None)
        if not attdict: attdict = {}
        sdict = attdict.copy() # for modification
        defaults = self.defaults
        for name in defaults.keys():
            if not sdict.has_key(name):
                sdict[name] = defaults[name]
        pcontent = None
        if content is not None:
            pcontent = self.MyProcessContent(content, controller, context, overrides)
            #sdict["__content__"] = pcontent
        transforms = self.transforms
        for name in transforms.keys():
            if sdict.has_key(name):
                t = transforms[name]
                sdict[name] = t(sdict[name])
        result = self.evaluate(tagname, sdict, pcontent, extra, context)
        return result

    def evaluate(self, tagname, sdict, pcontent, extra, context):
        """after all defaults and transformations, evaluate the tag.
        Maybe modify the content. return something to use higher up..."""
        # this superclass is a "pass thru", just returns the tag itself (with substitutions,etc)
        # no side effects to the context
        if tagname=="":
            return pcontent # top level
        L = ["<"+tagname]
        keys = sdict.keys()
        keys.sort()
        for k in keys:
            v = sdict[k] # XXXX maybe need to "escape" v???
            ss = LazySubst(' %s="%s"', (k, v))
            L.append(ss)
        if pcontent is not None:
            pcontent = joinflatten(pcontent)
            ss = LazySubst(">%s</%s>", (pcontent, tagname))
            L.append(ss)
        else:
            L.append("/>")
        return LazyJoin(L)

    def MyProcessContent(self, content, controller, context, overrides):
        """by default ask the global controller to do it"""
        return controller.processContent(content, context, overrides)

class NoWhiteContentMixin:
    def MyProcessContent(self, content, controller, context, overrides):
        """erase whitespace, then, ask the global controller to do it"""
        newcontent = []
        for c in content:
            if type(c) is StringType and not c.strip():
                pass # erase it
            else:
                newcontent.append(c)
        return controller.processContent(newcontent, context, overrides)

# the "top level" mapper (doesn't itself do anything)
Controller[""] = MapNode()

### map nodes for each element
### should define init1 and evaluate where appropriate

class document(MapNode):

    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # context
        context.filename = sdict["filename"]
        context.compression = sdict.get("compression", None)  #None means unknown
        context.invariant = sdict.get("invariant", None)  #None means unknown
        return MapNode.evaluate(self, tagname, sdict, pcontent, extra, context)

Controller["document"] = document()

class PassThru(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return "".join(pcontent)

Controller['<![CDATA['] = PassThru()

# defined below
#TOPLEVELCLASSES = (LazyParagraph, Table, xTable, Spacer, CondPageBreak,
#                   NextPageTemplate, PageBreak, ActionFlowable)
def _story_evaluate(tagname, sdict, pcontent, extra, context):
    for x in flatten(pcontent):
        tx = type(x)
        if x is None or x=="":
            pass # ignore
        elif tx is StringType:
            raise ValueError, "string %s in story content" % repr(x)
        elif tx is InstanceType:
            #print x.__class__
            if isinstance(x, DeferredInitialization):
                x = x.initialize() # no need to defer any longer for story.
            if 1 or x.__class__ in TOPLEVELCLASSES: # sanity test disabled (historical)
                context.story.append(x)
            else:
                raise ValueError, "bad instance in story content %s" % repr((x, x.__class__))

class _LContext:
    def __init__(self):
        self.story = []

def get_local_story(tagname, sdict, pcontent, extra):
    lcontext = _LContext()
    _story_evaluate(tagname, sdict, pcontent, extra, lcontext)
    return lcontext.story

class story(NoWhiteContentMixin, MapNode):
    def evaluate(self,tagname, sdict, pcontent, extra, context):
        _story_evaluate(tagname, sdict, pcontent, extra, context)
Controller["story"] = story()

class setNextTemplate(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        name = sdict["name"]
##        try:
##            template = context[name]
##        except:
##            raise ValueError, "no such page template %s" % repr(name)
##        if template.__class__ is not PageTemplate:
##            raise ValueError, "bad template %s" % repr(template)
        template = name
        return NextPageTemplate(template)

Controller["setNextTemplate"] = setNextTemplate()

class nextFrame(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return FrameBreak

Controller["nextFrame"] = nextFrame()

class nextPage(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return PageBreak

Controller["nextPage"] = nextPage()


class name(MapNode):
    "name a thing, for now just strings"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        value = sdict["value"]
        type = sdict.get("type", "string")
        if type=="string":
            #context[id] = value, not now but later
            pass
        else:
            raise ValueError, "name type %s not allowed" % repr(name)
        # when stringiffied set the value again.
        ### for example when processing the story or in initialization.
        return LazySetItem(context, id, value)

Controller["name"] = name()

class initialize(NoWhiteContentMixin, MapNode):
    "initialize names (eg, for headers and footers)"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # save delayed actions
        context.initialize = pcontent
        return "" # vanish

Controller["initialize"] = initialize()

class docinit(NoWhiteContentMixin, MapNode):
    """initialize anything - security, font setup etc.  Container for other tags"""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # save delayed actions
        context.docinit = pcontent
        return "" # vanish

Controller["docinit"] = docinit()

class alias(MapNode):
    "name a thing, for now just strings"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        value = sdict["value"]
        context.alias(id, value)
        return "" # vanish
Controller["alias"] = alias()

class registerType1Face(MapNode):
    "name a thing, for now just strings"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # if some glyphs are missing, suppress warnings
        import reportlab.rl_config
        reportlab.rl_config.warnOnMissingFontGlyphs = 0
        afmFile = sdict["afmFile"]
        pfbFile = sdict["pfbFile"]
        from reportlab.pdfbase import pdfmetrics
        theface = pdfmetrics.EmbeddedType1Face(afmFile, pfbFile)
        pdfmetrics.registerTypeFace(theface)
        return "" # vanish
Controller["registerType1Face"] = registerType1Face()

class registerFont(MapNode):
    "register an ordinary Type 1 Font"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        name = sdict["name"]
        faceName = sdict["faceName"]
        encName = sdict["encName"]
        from reportlab.pdfbase import pdfmetrics
        thefont = pdfmetrics.Font(name, faceName, encName)
        pdfmetrics.registerFont(thefont)
        return "" # vanish
Controller["registerFont"] = registerFont()

class registerCidFont(MapNode):
    "Register a CID (Asian) font"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        faceName = sdict["faceName"]
        encName = sdict["encName"]
        from reportlab.pdfbase.pdfmetrics import registerFont
        from reportlab.pdfbase.cidfonts import CIDFont
        registerFont(CIDFont(faceName, encName))
        return ""
Controller["registerCidFont"] = registerCidFont()

class registerTTFont(MapNode):
    "Register a TTF font"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        faceName = sdict["faceName"]
        fileName = sdict["fileName"]
        from reportlab.pdfbase.pdfmetrics import registerFont
        from reportlab.pdfbase.ttfonts import TTFont
        registerFont(TTFont(faceName, fileName))
        return ""
Controller["registerTTFont"] = registerTTFont()


class getName(MapNode):
    "name a thing, for now just strings"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        if sdict.has_key("default"):
            default = sdict["default"]
        else:
            default = None
        #val = context[id]
        ### note: this may occur in a HEADER/FOOTER
        ### and the value may change every time!!!
        ### so it's done lazily!
        return LazyGetItem(context,id,default)
Controller["getName"] = getName()

class color(MapNode):
    """define a color in the context"""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        RGB = sdict.get("RGB", None)
        CMYK = sdict.get("CMYK", None)
        alias = sdict.get("alias", None)
        if RGB:
            thecolor = reportlab.lib.colors.toColor(RGB)
            context[id] = RGB # for color in para, eg...
        elif CMYK:
            # XXXXX
            raise ValueError, "cmyk not implemented yet"
        elif alias:
            thecolor = context.colors[alias]
            context[id] = alias # for colors in para, eg...
        else:
            raise ValueError, "underdefined color %s" % repr(id)
        context.colors[id] = thecolor
        return ""
Controller["color"] = color()

class length(MapNode):
    """define a length in the context"""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        value = sdict.get("value",None)
        if not value: raise ValueError, "bad length %s" % repr(id)
        context.lengths[id] = value
        return ""
Controller["length"] = length()

class pageNumber(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return LazyPageNumber(context)

class LazyPageNumber(LazyStringForm):
    def __init__(self, context):
        self.context = context # possible cyclic structure!
    def __str__(self):
        return str(self.context.doc.page)
Controller["pageNumber"] = pageNumber()


# specialized info structures for recording product information

class PDFInfoL(pdfdoc.PDFInfo):
    """PDF documents can have basic information embedded, viewable from
    File | Document Info in Acrobat Reader.  If this is wrong, you get
    Postscript errors while printing, even though it does not print."""
    def __init__(self, oldinfo, user, number):
        self.invariant = 0
        self.title = oldinfo.title
        self.author = oldinfo.author
        self.subject = oldinfo.subject
        self.producer = "(%s,%s) RML2PDF http://www.reportlab.com" % (repr(user), repr(number))
        #now = time.localtime(time.time())
        #self.datestr = '%04d%02d%02d%02d%02d%02d' % tuple(now[0:6])

    def format(self, document):
        D = {}
        PDFString = pdfdoc.PDFString
        D["Title"] = PDFString(self.title)
        D["Author"] = PDFString(self.author)
        if self.invariant:
            D["CreationDate"] = PDFString('19001231000000')
        else:
            D["CreationDate"] = pdfdoc.PDFDate()
        D["Producer"] = PDFString(self.producer)
        D["Subject"] = PDFString(self.subject)
        PD = pdfdoc.PDFDictionary(D)
        return PD.format(document)

class PDFInfoU(pdfdoc.PDFInfo):
    def format(self, document):
        D = {}
        PDFString = pdfdoc.PDFString
        D["Title"] = PDFString("RML2PDF Evaluation Test Document")
        D["Author"] = PDFString("RML2PDF Evaluator")
        if getattr(self,'invariant',None):
            D["CreationDate"] = PDFString('19001231000000')
        else:
            D["CreationDate"] = pdfdoc.PDFDate()
        D["Producer"] = PDFString("Evaluation copy of RML2PDF http://www.reportlab.com")
        D["Subject"] = PDFString(
            "This document was created with an evaluation version of ReportLab RML2PDF.\n"
            "Please contact ReportLab to obtain a license for production use.")
        PD = pdfdoc.PDFDictionary(D)
        return PD.format(document)


class RMLDocTemplate(BaseDocTemplate):
    # primarily added for security purposes
    def beforeDocument(self):
        canv = self.canv
        # do security on the canvas
        alignmentAdjust(canv)

    def _endBuild(self):
        if self._hanging!=[] and self._hanging[-1] is PageBegin:
            del self._hanging[-1]
            self.clean_hanging()
        else:
            self.clean_hanging()
            self.handle_pageBreak()

        if getattr(self,'_doSave',1): self.canv.save()
        if self._onPage: self.canv.setPageCallBack(None)

# note: clever hacker could redefine this. Should add traps...
# this is misnamed for security purposes....
def alignmentAdjust(canv):
    #print "beginform a"
    canv.beginForm(repr(id(alignmentAdjust)))
        # check security.  if it fails complain on all pages
    SS = " RML2PDF is a licensed product from ReportLab http://www.reportlab.com. "
    ok = 0
    nagLine = "Evaluation or expired copy of ReportLab's RML2PDF. "
    #partners can customize/override the second part.
    nagPlus = "Contact ReportLab (www.reportlab.com) to purchase."
    try:
        if verbose:
            print "checking license"
        import RML2PDFLicense
        number = str(RML2PDFLicense.number)
        user = RML2PDFLicense.user
        if len(number)<11: raise ValueError
        key = RML2PDFLicense.key
        expire = RML2PDFLicense.expire
        misc = RML2PDFLicense.misc
        try:
            nagPlus = misc[0]
        except:
            pass
        Sn = SS+number+SS+user+SS+repr(expire)+SS+repr(misc)
        import md5
        code = md5.new(Sn).digest()
        code85 = pdfutils._AsciiBase85Encode(code)
        if code85!=key:
            raise "security fails!"
        import time
        if expire is not None and expire<time.time():
            if verbose:
                print "license expired %s" % repr(time.localtime(expire))
            raise "license expired %s" % repr(time.localtime(expire))
    except:
        #import sys
        #print sys.exc_type, sys.exc_value
        canv._doc.info = PDFInfoU() # mark file as unlicensed
        if verbose:
            print "license check failed"
        # license failure: mess up the form...
        # no comments in the output file
        pdfdoc.DoComments = 0
        #pdfdoc.DoComments = 1 # temp
        # encrypt all streams, make it hard to figure out how the encoding works.
        canv._doc.defaultStreamFilters = [pdfdoc.PDFZCompress]
        canv.drawString(10,10, nagLine + nagPlus)
    else:
        if verbose:
            print "license ok"
        canv._doc.info = PDFInfoL(canv._doc.info, user, number)
    #print "endform a"
    canv.endForm()

# weak hacker checks
AiD = id(alignmentAdjust)
sid = id(RMLDocTemplate)

# for example, this will not foil security:
#RMLDocTemplate = BaseDocTemplate

def pageSizeParse(sdict,defaultPageSize=defaultPageSize):
    ps = sdict.get('pageSize','').strip()
    if ps <> '':
        def err(ps=ps):
            raise ValueError, 'Bad pageSize attribute "%s"' % ps
        try:
            pageSize = lengthSequence(ps)
        except:
            L = ps.upper().split()
            if len(L)>2: err()
            from reportlab.lib import pagesizes
            pageSize = getattr(pagesizes,L[0],None)
            if pageSize:
                if len(L)==2:
                    if L[1]=='LANDSCAPE': pageSize = pageSize[1],pageSize[0]
                    elif L[1]!='PORTRAIT': err()
            else: err()
    else:
        pageSize = defaultPageSize
    return pageSize

class pageInfo(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        context.canvasInfo.pageSize = pageSizeParse(sdict)

Controller["pageInfo"] = pageInfo()

class template(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, deferredPageTemplates, extra, context):
        filename = context.filename
        pageSize = pageSizeParse(sdict)
        # pageTemplates contains DeferredPageTemplates, call them to create
        # the PageTemplate objects (now that we know have access to default pageSize)
        pageTemplates = map(lambda t, pageSize=pageSize: t.create(pageSize), deferredPageTemplates)
        showBoundary=0
        allowSplitting=1
        title=None
        author=None
        leftMargin=rightMargin=topMargin=bottomMargin=inch
        if sdict.has_key("showBoundary"):
            showBoundary = readBool(sdict["showBoundary"])
        if sdict.has_key("allowSplitting"):
            allowSplitting = int(sdict["allowSplitting"])
        if sdict.has_key("title"):
            title = sdict["title"]
        if sdict.has_key("author"):
            author = sdict["author"]
        if sdict.has_key("leftMargin"):
            leftMargin = readLength(sdict["leftMargin"],context=context)
        if sdict.has_key("rightMargin"):
            rightMargin = readLength(sdict["rightMargin"],context=context)
        if sdict.has_key("topMargin"):
            topMargin = readLength(sdict["topMargin"],context=context)
        if sdict.has_key("bottomMargin"):
            bottomMargin = readLength(sdict["bottomMargin"],context=context)

        context.pageSize = pageSize
        T = RMLDocTemplate(filename,
                               pagesize=pageSize,
                               pageTemplates=pageTemplates,
                               showBoundary=showBoundary,
                               leftMargin=leftMargin,
                               rightMargin=rightMargin,
                               topMargin=topMargin,
                               bottomMargin=bottomMargin,
                               allowSplitting=allowSplitting,
                               title=title,
                               author=author)
        pt = sdict.get("firstPageTemplate",None)
        if pt:
            T.handle_nextPageTemplate(pt)
            T._firstPageTemplateIndex = T._nextPageTemplateIndex
            del T._nextPageTemplateIndex
        context.doc = T
        return "" # vanish

Controller["template"] = template()

class CanvasOps:
    """a (delayed) sequence of canvas operations for drawing a page"""
    def __init__(self):
        self.ops = []

    def add(self,op):
        self.ops.append(op)

    def __call__(self, canvas, doc):
        # hacker check, don't make it extremely easy to bypass security
        if id(SWrap)!=WID:
            raise SV, SV
        if doc is not None and id(doc.__class__)!=sid:
            raise SV, SV
        canvas.saveState()
        #print "saveState in CanvasOp"
        for op in self.ops:
            #print op
            if op:
                op(canvas, doc)
        #print "restoreState in CanvasOp"
        canvas.restoreState()

CiD = id(CanvasOps)

class FormOps(CanvasOps):
    """a (delayed) form Xobject creator: only create the form ONCE
        (even if specified in a page template which is "executed" many times)
    """
    def __init__(self, formname):
        CanvasOps.__init__(self)
        self._formname = formname
        self._FormHasBeenCreated = 0
    def __call__(self, canvas, doc):
        if self._FormHasBeenCreated:
            pass # don't create it again
        else:
            #print "beginform", self._formname
            canvas.beginForm(self._formname)
            #CanvasOps.__call__(self, canvas, doc)
            for op in self.ops:
                #print op
                if op: # don't execute, eg None or ""
                    op(canvas, doc)
            canvas.endForm()
            #print "endform", self._formname
            self._FormHasBeenCreated = 1

def noop(*args, **kwargs):
    pass

class SWrap:
    # security marking wrapper: do form 122961 (mark or empty depending on license)
    def __init__(self, wrapped):
        self.w = wrapped
    def __call__(self, canvas, doc):
        #print "swrap", self.w
        #canvas.doForm(repr(id(alignmentAdjust))) # if alignmentadjust hacked, this will break the pdf file
        canvas.saveState()
        self.w(canvas, doc)
        canvas.restoreState()
        canvas.doForm(repr(id(alignmentAdjust))) # if alignmentadjust hacked, this will break the pdf file

# another weak hacker check
WID = id(SWrap)

class pageTemplate(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        content = list(pcontent)
        for c in content:
            if type(c) is not InstanceType:
                raise ValueError, "non instance content in pageTemplate"
        onPage = onPageEnd = noop
        if content[0].__class__==CanvasOps:
            onPage = content[0]
            del content[0]
        if content[0].__class__==CanvasOps:
            onPageEnd = content[0]
            del content[0]
        onPageEnd = SWrap(onPageEnd)
        return DeferredPageTemplate(context, id, content, onPage, onPageEnd, pagesize=pageSizeParse(sdict,None))

class DeferredPageTemplate:
    def __init__(self, context, id, frames, onPage, onPageEnd, pagesize):
        self._context, self._id, self._frames = context, id, frames
        self._onPage, self._onPageEnd, self._pagesize = onPage, onPageEnd, pagesize
    def create(self, defaultPageSize):
        # use default page size if pagetemplate doesn't have its own
        if not self._pagesize:
            self._pagesize = defaultPageSize
        frames = []
        for c in self._frames:
            if c.__class__ !=DeferredFrame:
                raise ValueError, "non frame in page template %s" % repr(c.__class__)
            # Create the actual Frame objects now that we have the page size
            # so can evaluate any % units used in the frame dimensions
            frames.append(c.create(self._pagesize))
        result = PageTemplate(
            self._id, frames, self._onPage, self._onPageEnd, self._pagesize
        )
        self._context[self._id] = result
        return result

Controller["pageTemplate"] = pageTemplate()

class pageGraphics(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        result = CanvasOps()
        for x in pcontent:
            result.add(x)
        return result

Controller["pageGraphics"] = pageGraphics()

class form(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        name = sdict["name"]
        result = FormOps(name)
        for x in pcontent:
            result.add(x)
        return result

Controller["form"] = form()

class doForm(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        name = sdict["name"]
        def doFormOp(canv, doc, name=name):
            canv.doForm(name)
        return doFormOp

Controller["doForm"] = doForm()


class pageDrawing(pageGraphics):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        result = pageGraphics.evaluate(self, tagname, sdict, pcontent, extra, context)
        result = SWrap(result) # security mark it
        context.pageDrawings.append(result)
        return result

Controller["pageDrawing"] = pageDrawing()

class image(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        file = sdict["file"]
        width = height = None
        if sdict.has_key("width"):
            width = readLength(sdict["width"],context=context)
        if sdict.has_key("height"):
            height = readLength(sdict["height"],context=context)
        inline = sdict.has_key('inline')
        if inline:
            inline = readBool(sdict['inline'])
        mask = sdict.get('transparency_mask','auto').strip()
        if mask!='auto': mask = reportlab.lib.colors.toColor(mask)

        if inline:
            name = str(hash(file))
            makeForm = not context.imageFileMap.has_key(name) # form not made yet
            #die
            context.imageFileMap[name] = 1
            #print "width, height", width, height, sdict
            ##  def imageOp(canv, doc, x=x, y=y, file=file, width=width, height=height):
            ##        canv.drawInlineImage(file, x, y, width, height)
            def imageOp(canv, doc, x=x, y=y, file=file, width=width, height=height,
                        makeForm=makeForm,name=name,mask=mask):
                # AR 2001-10-12 - this fails when the image
                # is in a pageGraphics section on a multi
                # page document, as it is called on every
                # page with makeForm = true, and tries
                # to recreate the same form.  Ugly
                # workaround: when it fails,
                # reference the form already there.
                if makeForm: #not internalname:
                    #would prefer canvas.hasForm instead of poking around in here
                    if canv._doc.idToObject.has_key(pdfdoc.xObjectName(name)):
                        # I am being drawn again on page 2, skip.
                        canv.doForm(name)
                    else:
                        canv.beginForm(name)
                        canv.drawInlineImage(file, x, y, width, height,mask=mask)
                        canv.endForm()
                        canv.doForm(name)
                else:
                    canv.doForm(name)
        else:
            def imageOp(canv, doc, x=x, y=y, file=file, width=width, height=height,mask=mask):
                canv.drawImage(file, x, y, width, height,mask=mask)
        return imageOp

Controller["image"] = image()

class plugInGraphic(MapNode):
    """Call a function in a "plug in module" that is on the python path.
    The function must take exactly two arguments:
        f(canvas, datastring)
    where canvas is the pdfgen canvas object and datastring is the data occurring
    between the begin and end tags.
    """
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        try:
            dirName = sdict["baseDir"]
        except:
            dirName = None
        moduleName = sdict["module"]
        functionName = sdict["function"]
        def plugInOp(canv, doc, dirName=dirName, moduleName=moduleName, functionName=functionName, content=pcontent):
            content = map(str, content) # lazy conversion....
            data = "".join(content)
            #print 'plugInOp called: dirName=%s, moduleName=%s, functionName=%s' % (
            #    dirName, moduleName, functionName)
            module = recursiveImport(moduleName, _rml2pdf_locations(dirName))
            f = getattr(module, functionName)
            f(canv, data)
        return plugInOp
Controller["plugInGraphic"] = plugInGraphic()

def _args_kw(*args,**kw):
    return args, kw
class LazyPlugInFlowable(Flowable):
    """defer conversion of content until wrap time (to allow, eg, page numbering)"""
    def __init__(self, dirname, modulename, functionname, content):
        #print "lazy plug in flowable", (modulename, functionname, content)
        self.dirname = dirname # usually None
        self.modulename = modulename
        self.functionname = functionname
        self.content = content
        self.flowable = None
    def initializeFlowable(self):
        # if packagized, import package first
        m = recursiveImport(self.modulename, _rml2pdf_locations(self.dirname))
        f = getattr(m, self.functionname)

        # handle argument list
        if self.content is None:
            self.flowable = f()
        else:
            data = "".join(map(str, self.content))
            kw = {}
            try:
                args = eval(data)   #first try simple eval
                if type(args) is not type(()): args = (args,)
            except:
                try:
                    args, kw = eval('_args_kw(%s)' % data)  #try assuming keyword args
                except:
                    args = (data,)  #just pass as a single argument
            try:
                self.flowable = f(*args,**kw)
            except:
                print 'flowable %s, args=%s kw=%s' % (f, args, kw)
                raise

    def wrap(self, w, h):
        if self.flowable is None:
            self.initializeFlowable()
        return self.flowable.wrapOn(self.canv,w,h)

    def draw(self):
        if self.flowable is None:
            self.initializeFlowable()
        self.flowable._drawOn(self.canv)

    def split(self,aW,aH):
        if self.flowable is None:
            self.initializeFlowable()
        return self.flowable.splitOn(self.canv,aW,aH)

class OutlineEntry(Flowable):
    """make an outline entry (lazily convert content)
       should work either in story or in page graphic mode!
    """
    def __init__(self, level, content, closed):
        self.level = level
        self.content = content
        self.closed = closed
    def wrap(self, w, h):
        return (0,0) # consumes no space
    def draw(self):
        "flowable mode"
        self.doEntry(self.canv)
    def __call__(self, canv, doc):
        "graphic mode"
        self.doEntry(canv)
    def doEntry(self, canv):
        key = "k"+repr(id(self)) # arbitrary unique string
        content = "".join(map(str, self.content))
        for (old, new) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
            content = content.replace(old,new)
        canv.bookmarkPage(key)
        # bogus level may result in error here
        canv.addOutlineEntry(content, key, self.level, self.closed)

class NameFlowable(OutlineEntry):
    "name a string when executed (no visible effect)"
    def __init__(self, id, default, context, pcontent):
        self.id = id
        self.default = default
        self.context = context
        self.pcontent = pcontent
    def doEntry(self, canv):
        try:
            content = "".join(map(str, self.pcontent))
        except:
            if self.default is not None:
                content = self.default
            else:
                raise
        self.context[self.id] = content
        return "" # vanish


class EvaluateOp(LazyStringForm):

    "try to evaluate the content as a python expression, return result"
    def __init__(self, default, pcontent, imports, context):
        # note: in some untrusted environments this might be a very big security hole, so we permit it to be disabled
        if not context.permitEvaluations:
            raise ValueError, "RML2PDF invocation disallows the evalString tag for this instance (security check)"
        self.default = default
        self.pcontent = pcontent
        self.imports = imports
        self.context = context
    def __str__(self):
        content = "".join(map(str, self.pcontent))
        # clean up possible newlines in expression
        content = ' '.join(content.split())
        g = globals()
        l = locals()
        ### NOTE THIS WILL NOT WORK WITH PACKAGE IMPORTS AT THIS TIME!
        imps = self.imports
        if imps:
            implist = ",".split(imps)
            for im in implist:
                im = im.strip()
                try:
                    module = __import__(im)
                except:
                    raise ValueError, "failed to import %s from list %s" % (repr(im), implist)
                l[im] = module
        result = eval(content, g, l)
        return str(result)

class FormsCatcher:
    executed = 0
    def __init__(self, filename,pfx=None,pages=None):
        if not os.path.isabs(filename):
            filename = os.path.normpath(os.path.abspath(filename))
        self.filename = filename
        self.pfx = pfx
        self.pages = pages
    def __call__(self, canv, doc):
        "import the forms (only once!)"
        if not self.executed:
            filename = self.filename
            pfx = self.pfx
            self.executed = 1
            try:
                from rlextra.pageCatcher import pageCatcher
            except ImportError:
                import pageCatcher
            except:
                raise ImportError, "catchForms tag requires the PageCatcher product http://www.reportlab.com"
            if pfx is None:
                pageCatcher.restoreForms(self.filename, canv, allowDuplicates=1)
            else:
                pages = self.pages
                if pages=='all':
                    all = 1
                    pages = []
                else:
                    all = None
                pdfContent = open_and_read(filename)
                result = pageCatcher.storeFormsInMemory(pdfContent, pages, prefix=pfx, all=all,
                            verbose=0, BBoxes=0, extractText=0, fformname=None)
                pageCatcher.restoreFormsInMemory(result[1], canv, verbose=0, formnames=None, allowDuplicates=1)

class catchForms(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        try:
            filename = sdict["storageFile"]
            pfx = pages = None
        except KeyError:
            try:
                filename = sdict['pdfFile']
            except KeyError:
                raise ValueError("catchForms tag needs either storageFile or pdfFile attribute")
            pfx = sdict.get('pfx','PF')
            pages = sdict.get('pages','all').strip()
            if pages!='all': pages = map(int,pages.split())
        return FormsCatcher(filename, pfx, pages)
Controller["catchForms"] = catchForms()

class plugInFlowable(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        try:
            dirName = sdict["baseDir"]
        except KeyError:
            dirName = None
        moduleName = sdict["module"]
        functionName = sdict["function"]
        return LazyPlugInFlowable(dirName, moduleName, functionName, pcontent)

Controller["plugInFlowable"] = plugInFlowable()

class scale(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        sx = sy = 1.0
        if sdict.has_key("sx"):
            sx = readLength(sdict["sx"],context=context)
        if sdict.has_key("sy"):
            sy = readLength(sdict["sy"],context=context)
        def scaleOp(canv, doc, sx=sx,sy=sy):
            canv.scale(sx,sy)
        return scaleOp

Controller["scale"] = scale()

class translate(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        dx = dy = 0.0
        if sdict.has_key("dx"):
            dx = readLength(sdict["dx"],context=context)
        if sdict.has_key("dy"):
            dy = readLength(sdict["dy"],context=context)
        def scaleOp(canv, doc, dx=dx,dy=dy):
            canv.translate(dx,dy)
        return scaleOp

Controller["translate"] = translate()

class outlineAdd(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        level = 0
        closed = 0
        if sdict.has_key("closed"):
            closed = readBool(sdict["closed"])
        if sdict.has_key("level"):
            level = int(sdict["level"])
        return OutlineEntry(level, pcontent, closed)

Controller["outlineAdd"] = outlineAdd()

class namedString(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        default = None
        if sdict.has_key("default"):
            default = sdict["default"]
        return NameFlowable(id, default, context, pcontent)

Controller["namedString"] = namedString()

class evalString(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        imports = ""
        default = None
        if sdict.has_key("imports"):
            imports = sdict["imports"]
        if sdict.has_key("default"):
            default = sdict["default"]
        return EvaluateOp(default, pcontent, imports, context)

Controller["evalString"] = evalString()


class grid(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        xs = lengthSequence(sdict["xs"])
        ys = lengthSequence(sdict["ys"])
        def gridOp(canv, doc, xs=xs, ys=ys):
            canv.grid(xs, ys)
        return gridOp

Controller["grid"] = grid()

class rotate(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        degrees = float(sdict["degrees"])
        def rotateOp(canv, doc, degrees=degrees):
            canv.rotate(degrees)
        return rotateOp

Controller["rotate"] = rotate()

class skew(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        alpha = float(sdict["alpha"])
        beta = float(sdict["beta"])
        def skewOp(canv, doc, alpha=alpha, beta=beta):
            canv.skew(alpha, beta)
        return skewOp

Controller["skew"] = skew()

joinmap = {"round": 1, "mitered": 0, "bevelled": 2}
capmap = {"default": 0, "round": 1, "square": 2, "butt": 0}

class lineMode(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        join = cap = miterLimit = dash = width = None
        if sdict.has_key("join"):
            join = joinmap[sdict["join"].lower()] # could trap keyerror
        if sdict.has_key("cap"):
            cap = capmap[sdict["cap"].lower()] # could trap keyerror
        if sdict.has_key("miterLimit"):
            miterLimit = readLength(sdict["miterLimit"],context=context)
        if sdict.has_key("dash"):
            dash = lengthSequence(sdict["dash"])
        if sdict.has_key("width"):
            width = readLength(sdict["width"],context=context)
        def lineStyleop(canvas, doc, join=join, cap=cap, miterLimit=miterLimit, dash=dash, width=width):
            if join is not None:
                canvas.setLineJoin(join)
            if cap is not None:
                canvas.setLineCap(cap)
            if miterLimit is not None:
                canvas.setMiterLimit(miterLimit)
            if width is not None:
                canvas.setLineWidth(width)
            if dash is not None:
                canvas.setDash(dash, 0)
        return lineStyleop

Controller["lineMode"] = lineMode()

class lines(MapNode):
    """<lines>1 2 3 4 4 5 6 7</lines>
    means line from (1,2) to (3,4) and a line from (4,5) to (6,7)
    must be in groups of 4
    """
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        if len(pcontent)!=1:
            raise ValueError, "only nonempty point pairs allowed in lines %s" % repr(pcontent)
        seq = pcontent[0].split()
        if len(seq)%4 != 0:
            raise ValueError, "each line requires two pairs %s" % repr(seq)
        lines = []
        while seq:
            thisline = seq[:4]
            del seq[:4]
            thisline = map(readLength, thisline)
            lines.append(thisline)
        def lineop(canvas, doc, lines=lines):
            for l in lines:
                (x1,y1,x2,y2) = l
                canvas.line(x1,y1,x2,y2)
        return lineop

Controller["lines"] = lines()

class transform(MapNode):
    """<lines>1 2 3 4 4 5 6 7</lines>
    means line from (1,2) to (3,4) and a line from (4,5) to (6,7)
    must be in groups of 4
    """
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        if len(pcontent)!=1:
            raise ValueError, "only 6 numbers required for transform" % repr(pcontent)
        seq = pcontent[0].split()
        if len(seq) != 6:
            raise ValueError, "exactly 6 numbers required for transform %s" % repr(seq)
        m = map(float, seq)
        def lineop(canvas, doc, m=m):
            (a,b,c,d,e,f) = m
            canvas.transform(a,b,c,d,e,f)
        return lineop

Controller["transform"] = transform()

class curves(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        if len(pcontent)!=1:
            raise ValueError, "only nonempty point pairs allowed in curves %s" % repr(pcontent)
        seq = pcontent[0].split()
        if len(seq)%8 != 0:
            raise ValueError, "each curve requires 4 pairs %s" % repr(seq)
        curves = []
        while seq:
            this = seq[:8]
            del seq[:8]
            this = map(readLength, this)
            curves.append(this)
        def lineop(canvas, doc, curves=curves):
            for c in curves:
                (x1,y1,x2,y2,x3,y3,x4,y4) = c
                canvas.bezier(x1,y1,x2,y2,x3,y3,x4,y4)
        return lineop

Controller["curves"] = curves()

class path(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x0 = readLength(sdict["x"],context=context)
        y0 = readLength(sdict["y"],context=context)
        stroke = 1
        fill = 0
        close = 0
        if sdict.has_key("stroke"):
            stroke = readBool(sdict["stroke"])
        if sdict.has_key("fill"):
            fill = readBool(sdict["fill"])
        if sdict.has_key("close"):
            close = readBool(sdict["close"])
        # extract content
        ops = []
        content = list(pcontent)
        while content:
            strings = []
            while content and type(content[0]) is StringType:
                strings.append(content[0])
                del content[0]
            # if we have any strings they should be pairs of metrics
            if len(strings) > 1:
                strings = [" ".join(strings)]
            # collect the operations in the content, untagged point pairs are lineTos
            if strings:
                strings = strings[0].split()
                points = []
                metrics = map(readLength, strings)
                while metrics:
                    [x,y] = metrics[:2]
                    del metrics[:2]
                    points.append((x,y))
                def lineToOps(path, points=points):
                    for (x,y) in points:
                        #print "lineto", x,y
                        path.lineTo(x,y)
                ops.append(lineToOps)
            # if anything remains in content, head must be another op (curveto, moveto)
            if content:
                op = content[0]
                del content[0]
                ops.append(op)
        # return the canvas operation that will execute and draw the path
        # XXXX what about the clipping stuff? (deferred)
        def pathOps(canvas, doc, ops=ops, x0=x0, y0=y0, fill=fill, stroke=stroke, close=close):
            #print "pathOps"
            path = canvas.beginPath()
            path.moveTo(x0,y0)
            for op in ops:
                #print "op", op
                op(path)
            if close: path.close()
            canvas.drawPath(path, fill=fill, stroke=stroke)
        return pathOps

Controller["path"] = path()

class curvesto(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        if len(pcontent)!=1:
            raise ValueError, "only nonempty point pairs allowed in curvestos %s" % repr(pcontent)
        seq = pcontent[0].split()
        if len(seq)%6 != 0:
            raise ValueError, "each curve requires 3 pairs %s" % repr(seq)
        curves = []
        while seq:
            this = seq[:6]
            del seq[:6]
            this = map(readLength, this)
            curves.append(this)
        def curvesop(path, curves=curves):
            for c in curves:
                #print "curveto", c
                (x1,y1,x2,y2,x3,y3) = c
                path.curveTo(x1,y1,x2,y2,x3,y3)
        return curvesop

Controller["curvesto"] = curvesto()

class moveto(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        if len(pcontent)!=1:
            raise ValueError, "point required in moveto %s" % repr(pcontent)
        seq = pcontent[0].split()
        if len(seq) != 2:
            raise ValueError, "point required in moveto %s" % repr(seq)
        point = map(readLength, seq)
        def movetoop(path, point=point):
            #print "moveto", point
            (x,y) = point
            path.moveTo(x,y)
        return movetoop

Controller["moveto"] = moveto()

class illustration(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        height = readLength(sdict["height"],context=context)
        width = readLength(sdict["width"],context=context)
        I = IllustrationFlowable(width, height, pcontent)
        if sdict.has_key("align"): I.hAlign = sdict["align"]
        return I

Controller["illustration"] = illustration()

class IllustrationFlowable(Flowable):
    def __init__(self, width, height, drawOps):
        self.width = width
        self.height = height
        self.drawOps = drawOps
    def wrap(self, *args):
        return (self.width, self.height)
    def draw(self):
        canvas = self.canv
        doc = None # not used (I hope!)
        for op in self.drawOps:
            op(canvas, doc)

class place(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        width = readLength(sdict["width"],context=context)
        height = readLength(sdict["height"],context=context)
        return drawFlowables(x, y, width, height, pcontent)

Controller["place"] = place()

class storyPlace(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        width = readLength(sdict["width"],context=context)
        height = readLength(sdict["height"],context=context)
        origin = sdict.get("origin","local")
        return IllustrationFlowable(0, 0, [drawFlowables(x, y, width, height, pcontent,extra,context,origin)])
Controller["storyPlace"] = storyPlace()


class indent(MapNode):
    def MyProcessContent(self, content, controller, context, overrides):
        """erase whitespace, forbid non-whitespace CDATA"""
        newcontent = []
        for c in content:
            if type(c) is StringType:
                if c.strip(): raise ValueError("non white CDATA '%s' not allowed" % c.strip())
            else:
                newcontent.append(c)
        return controller.processContent(newcontent, context, overrides)

    def evaluate(self, tagname, sdict, pcontent, extra, context):
        from reportlab.platypus.doctemplate import Indenter
        left = readLength(sdict.get("left",'0'),context=context)
        right = readLength(sdict.get("right",'0'),context=context)
        return [Indenter(left,right)]+pcontent+[Indenter(-left,-right)]
Controller["indent"] = indent()

class FixedSize(Flowable):
    def __init__(self, tagname, sdict, pcontent, extra, context):
        self._info =  tagname, sdict, pcontent, extra, context
        self.flowables = None

    def _Initialize(self):
        if self.flowables is None:
            tagname, sdict, pcontent, extra, context = self._info
            self.width = readLength(sdict.get("width",'0'),context=context)
            self.height = readLength(sdict.get("height",'0'),context=context)
            self.flowables = get_local_story(tagname, sdict, pcontent, extra)

    def wrap(self, aW, aH):
        self._Initialize()
        for f in self.flowables:
            f.wrapOn(self.canv,aW,aH)
        return (self.width, self.height)

    def draw(self):
        self._Initialize()
        for f in self.flowables:
            f._drawOn(self.canv)

    def split(self,aW,aH):
        return []

class fixedSize(story):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return DeferredInitialization(FixedSize, tagname, sdict, pcontent, extra, context)
Controller["fixedSize"] = fixedSize()

class _pto_child(Flowable):
    def __init__(self, tagname, sdict, pcontent, extra, context):
        self.flowables = get_local_story(tagname, sdict, pcontent, extra)
        self.kind = tagname[4:]

class pto_child(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return DeferredInitialization(_pto_child, tagname, sdict, pcontent, extra, context)
Controller["pto_trailer"] = pto_child()
Controller["pto_header"] = pto_child()

class _PTO(PTOContainer):
    def __init__(self, tagname, sdict, pcontent, extra, context):
        F = get_local_story(tagname, sdict, pcontent, extra)
        H = T = None
        C = [_ for _ in F if isinstance(_,_pto_child)]
        F = [_ for _ in F if not isinstance(_,_pto_child)]
        for _ in C:
            if _.kind=='trailer': T = _.flowables
            elif _.kind=='header': H = _.flowables
        PTOContainer.__init__(self,F,T,H)

class pto(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return DeferredInitialization(_PTO, tagname, sdict, pcontent, extra, context)
Controller["pto"] = pto()

class drawFlowables:
    def __init__(self, x, y, width, height, pcontent, extra=None, context=None, origin='local'):
        self.x, self.y, self.width, self.height = x, y, width, height
        self.pcontent, self.extra, self.context, self.origin = pcontent, extra, context, origin
    def __call__(self, canvas, doc):
        origin = self.origin
        if origin != 'local':
            canvas.saveState()
            canvas.resetTransforms()
            if origin=='frame':
                frame = self.context.doc.frame
                canvas.translate(frame._x1,frame._y1)

        (aW, aH) = (self.width, self.height)
        for flowable in self.pcontent:
            if isinstance(flowable, DeferredInitialization):
                flowable = flowable.initialize() # create the flowable now and get the correct bindings
            (w, h) = flowable.wrapOn(canvas,aW, aH)
            if w<=aW and h<=aH:
                aH = aH-h # room left for next ones
                flowable.drawOn(canvas, self.x, self.y+aH)
            else:
                #content too big for place.  Strictness can be varies
                from reportlab.rl_config import allowTableBoundsErrors
                if allowTableBoundsErrors:
                    pass # we just don't draw the extra stuff
                else:  #fail, telling them which paragraph or whatever bombed and why
                    raise ValueError, "not enough room in 'place' for %s. \n (place size: x=%s y=%s width=%s height=%s;  \n offending flowable %s size (%s, %s)" % (
                        repr(self.pcontent),
                        self.x, self.y, self.width, self.height,
                        flowable.identity(), w, h)
        if origin != 'local': canvas.restoreState()

class drawString(MapNode):
    """Allow late string conversions in content (for, eg, page
       number or other varying values)"""
    op = "drawString"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        return delayedDrawString(self.op, x,y,pcontent)

class delayedDrawString:
    def __init__(self, op, x, y, stringseq):
        (self.op, self.x, self.y, self.stringseq) = (op, x, y, stringseq)
    def __call__(self, canvas, doc):
        s = "".join(map(str, self.stringseq))
        for (old, new) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
            s = s.replace(old, new)
        getattr(canvas, self.op)(self.x, self.y, s)

Controller["drawString"] = drawString()

class drawRightString(drawString):
    op = "drawRightString"

Controller["drawRightString"] = drawRightString()

class drawCentredString(drawString):
    op = "drawCentredString"

Controller["drawCentredString"] = drawCentredString()
Controller["drawCenteredString"] = drawCentredString()

class setFont(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        name = sdict["name"]
        size = readLength(sdict["size"],context=context)
        def setFontOp(canv, doc, name=name, size=size):
            canv.setFont(name, size)
        return setFontOp

Controller["setFont"] = setFont()

class stroke(MapNode):
    op = "setStrokeColor"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        width = sdict.get("width",None)
        if width is not None:
            width = readLength(width,context=context)
        def setStrokeOp(canv, doc, colorName=sdict["color"], width=width, context=context, op=self.op):
            color = getColor(colorName,context)
            getattr(canv, op)(color)
            if width is not None:
                getattr(canv,'setLineWidth')(width)
        return setStrokeOp

Controller["stroke"] = stroke()

class fill(stroke):
    op = "setFillColor"

Controller["fill"] = fill()

class rect(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        width = readLength(sdict["width"],context=context)
        height = readLength(sdict["height"],context=context)
        fill = 0
        if sdict.has_key("fill"):
            fill = readBool(sdict["fill"])
        round = 0
        if sdict.has_key("round"):
            round = readLength(sdict["round"],context=context)
        stroke = 1
        if sdict.has_key("stroke"):
            stroke = readBool(sdict["stroke"])
        def rectOp(canv, doc, x=x, y=y,
                   width=width, height=height, fill=fill, stroke=stroke, round=round):
            if round:
                canv.roundRect(x,y,width,height,round, fill=fill,stroke=stroke)
            else:
                canv.rect(x,y,width,height,fill=fill,stroke=stroke)
        return rectOp


Controller["rect"] = rect()

class circle(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        radius = readLength(sdict["radius"],context=context)
        fill = 0
        if sdict.has_key("fill"):
            fill = readBool(sdict["fill"])
        stroke = 1
        if sdict.has_key("stroke"):
            stroke = readBool(sdict["stroke"])
        def circleOp(canv, doc, x=x, y=y,
                   radius=radius, fill=fill, stroke=stroke):
            canv.circle(x,y,radius,fill=fill, stroke=stroke)
        return circleOp

Controller["circle"] = circle()

class ellipse(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        width = readLength(sdict["width"],context=context)
        height = readLength(sdict["height"],context=context)
        fill = 0
        if sdict.has_key("fill"):
            fill = readBool(sdict["fill"])
        stroke = 1
        if sdict.has_key("stroke"):
            stroke = readBool(sdict["stroke"])
        def ellipseOp(canv, doc, x=x, y=y,
                   width=width, height=height, fill=fill, stroke=stroke):
            #print "ellipse", x,y,width,height,fill,stroke
            canv.ellipse(x,y,width,height,fill=fill, stroke=stroke)
        return ellipseOp
Controller["ellipse"] = ellipse()

class frame(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        x1 = sdict["x1"]
        y1 = sdict["y1"]
        width = sdict["width"]
        height = sdict["height"]
        from reportlab.rl_config import overlapAttachedSpace
        overlapAttachedSpace = sdict.get('overlapAttachedSpace',None)
        if overlapAttachedSpace is not None: overlapAttachedSpace = int(overlapAttachedSpace)
        # Wait until we have access to our container's size before creating Frame
        # object so that we can allow percent (%) units for frame dimensions
        return DeferredFrame(id, x1, y1, width, height, overlapAttachedSpace,context)

class DeferredFrame:
    def __init__(self, id, x1, y1, width, height, overlapAttachedSpace,context):
        self._id = id
        self._x1 = x1
        self._y1 = y1
        self._width = width
        self._height = height
        self._oASpace = overlapAttachedSpace
        self._context = context

    def create(self, parentSize):
        x1 = readLength(self._x1, parentSize[0],context=self._context)
        y1 = readLength(self._y1, parentSize[1],context=self._context)
        width = readLength(self._width, parentSize[0],context=self._context)
        height = readLength(self._height, parentSize[1],context=self._context)
        return Frame(x1, y1, width, height, id=self._id,
                     #AR 2000-12-10 - should do this in Platypus?? makes life
                     #much easier positioning frames without boundaries.
                     leftPadding=0, rightPadding=0, topPadding=0,bottomPadding=0, overlapAttachedSpace=self._oASpace)
Controller["frame"] = frame()

#these classes are implemented as do nothings ie they utilize the base MapNode class
for _ in ("stylesheet", "b", "i", "font", "greek",
        "super", "sub", "seq", "seqDefault", "seqReset",
        "seqChain", "seqFormat", "bullet",
        ):
    Controller[_] = Controller[""]
del _

class blockTableStyle(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        id = sdict["id"]
        parentName = sdict.get("parent", None)
        if parentName:
            parent = context[parentName]
        else:
            parent = None
        st = TableStyle(pcontent,parent=parent,keepWithNext=_checkBool(sdict,"keepWithNext",0))
        context[id] = st
        return st

Controller["blockTableStyle"] = blockTableStyle()


class lineStyle(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        localdict = sdict.copy()
        kind = localdict["kind"]
        thickness = 1
        if localdict.has_key("thickness"):
            thickness = readLength(localdict["thickness"],context=context)
        color = getColor(localdict["colorName"],context)
        (start, stop) = startstop(localdict)
        if localdict.has_key("cap"): cap = capmap[localdict["cap"].lower()]
        else: cap = 'round'

        count = int(localdict.get("count","1"))
        if localdict.has_key('space'):
            space = readLength(localdict["space"],context=context)
        else:
            space = thickness


        return (kind, start, stop, thickness, color, cap, None, None, count, space)

def _int_or_split(x):
    try:
        return int(x)
    except ValueError:
        if x in ('splitfirst','splitlast'): return x
        raise

def startstop(dict):
    if dict.has_key("start"):
        start = dict["start"]
        start = lengthSequence(start, converter=_int_or_split)
    else:
        start = (0,0)
    if dict.has_key("stop"):
        stop = dict["stop"]
        stop = lengthSequence(stop, converter=_int_or_split)
    else:
        stop = (-1,-1)
    return (start, stop)

Controller["lineStyle"] = lineStyle()



class blockFont(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        fontname = sdict["name"]
        (start, stop) = startstop(sdict)
        result = ["FONT", start, stop, fontname]
        if sdict.has_key("size"):
            result.append(readLength(sdict["size"],context=context))
            if sdict.has_key("leading"):
                result.append(readLength(sdict["leading"],context=context))
        return tuple(result)

Controller["blockFont"] = blockFont()

class blockTextColor(MapNode):
    id = "TEXTCOLOR"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        (start, stop) = startstop(sdict)
        color = getColor(sdict["colorName"],context)
        return (self.id, start, stop, color)

Controller["blockTextColor"] = blockTextColor()

class blockSpan(MapNode):
    id = "SPAN"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        (start, stop) = startstop(sdict)
        return (self.id, start, stop)

Controller["blockSpan"] = blockSpan()

class blockLeading(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        (start, stop) = startstop(sdict)
        length = sdict["length"]
        length = readLength(length,context=context)
        return ("LEADING", start, stop, length)

Controller["blockLeading"] = blockLeading()

class blockAlignment(MapNode):
    # note: if value checking is added need to modify valign too!
    id = "ALIGNMENT"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        (start, stop) = startstop(sdict)
        return (self.id, start, stop, sdict["value"].upper())

Controller["blockAlignment"] = blockAlignment()

class blockValign(blockAlignment):
    id = "VALIGN"

Controller["blockValign"] = blockValign()

class blockLeftPadding(MapNode):
    id = "LEFTPADDING"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        (start, stop) = startstop(sdict)
        length = sdict["length"]
        length = readLength(length,context=context)
        return (self.id, start, stop, length)

Controller["blockLeftPadding"] = blockLeftPadding()

class blockRightPadding(blockLeftPadding):
    id = "RIGHTPADDING"

Controller["blockRightPadding"] = blockRightPadding()

class blockBottomPadding(blockLeftPadding):
    id = "BOTTOMPADDING"

Controller["blockBottomPadding"] = blockBottomPadding()

class blockTopPadding(blockLeftPadding):
    id = "TOPPADDING"

Controller["blockTopPadding"] = blockTopPadding()

class blockBackground(blockTextColor):
    id = "BACKGROUND"

Controller["blockBackground"] = blockBackground()

def _checkBool(dict,a,d=None):
    if d is not None:
        v = dict.get(a,str(d))
    else:
        v = dict[a]
    vl = v.lower()
    if vl in ['1','yes','true','on']: v = 1
    elif vl in ['0','no','false','off']: v = 0
    else: raise ValueError, 'Attribute "%s" given invalid value "%s"' % (a,v)
    return v

class paraStyle(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        parentname = None
        localdict = sdict.copy()
        if localdict.has_key("parent"):
            parentname = localdict["parent"]
            style = context[parentname]
            localdict["parent"] = style
        if not localdict.has_key("name"):
            raise ValueError, "name required for styles"
        name = localdict["name"]
        # fix attribute types
        for numericatt in ("fontSize", "leading",
                           "leftIndent", "rightIndent",
                           "firstLineIndent", "spaceBefore",
                           "spaceAfter", "bulletFontSize",
                           "bulletIndent"):
            if localdict.has_key(numericatt):
                value = localdict[numericatt]
                value = readLength(value,context=context)
                localdict[numericatt] = value
        for a in ['pageBreakBefore','frameBreakBefore','keepWithNext']:
            if localdict.has_key(a):
                localdict[a] = _checkBool(localdict,a)
        # test
        #localdict["spaceBefore"] = "invalid"
        if localdict.has_key("textColor"):
            localdict["textColor"] = getColor(localdict["textColor"],context)
        if localdict.has_key("alignment"):
            localdict["alignment"] = readAlignment(localdict["alignment"])
        context[name] = ParagraphStyle(**localdict)
        return "" # vanish

Controller["paraStyle"] =  paraStyle()


def quoteText(v):
    # quote ampersands, but not if already in an entity
    # XXXX this is a hack but because not all strings go through the platypus para
    # mechanism I need to make sure that "bare" &'s are allowed for
    # other strings, so I also need to hack/fix them here.
    # (avoiding reencoding paras as strings again should fix this (?)
    vs = v.split("&")
    if len(vs)<=1:
        return v
    out = [vs[0]]
    for frag in vs[1:]:
        #print frag
        doquote = 1
        if ";" in frag:
            tosemi = frag.split(";")[0]
            #print tosemi
            if len(tosemi.split())<=1:
                # this is an entity, don't quote it again
                doquote = 0
        if doquote:
            out.append("&amp;")
        else:
            out.append("&")
        out.append(frag)
    v = "".join(out)
    #print v
    return v

class pre(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        stylename = "pre.defaultStyle"
        if sdict.has_key("style"):
            stylename = sdict["style"]
        text = LazyJoin(pcontent)
        style = context[stylename]
        P = LazyPreformatted(text, style)
        return P
Controller["pre"] = pre()

class xpre(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        stylename = "pre.defaultStyle"
        if sdict.has_key("style"):
            stylename = sdict["style"]
        text = LazyJoin(pcontent)
        style = context[stylename]
        P = LazyXPreformatted(text, style)
        return P
Controller["xpre"] = xpre()

class hr(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        return HRFlowable(
                        width=readLength(sdict.get('width','80%'),"%",context=context),
                        thickness=readLength(sdict.get('thickness','1'),context=context),
                        lineCap=sdict.get('lineCap','round'),
                        color=getColor(sdict.get('color','lightgrey'),context=context),
                        hAlign=readAlignment(sdict.get('align','centre')),
                        spaceBefore=readLength(sdict.get('spaceBefore','1'),context=context),
                        spaceAfter=readLength(sdict.get('spaceAfter','1'),context=context),
                        )
Controller["hr"] = hr()

class para(MapNode):
     def evaluate(self, tagname, sdict, pcontent, extra, context):
        localdict = sdict.copy()
        stylename = "para.defaultStyle"
        if localdict.has_key("style"):
            stylename = localdict["style"]
            del localdict["style"]
        bulletText = None
        if localdict.has_key("bulletText"):
            bulletText = localdict["bulletText"]
            del localdict["bulletText"]
        style = context[stylename]
        # strip the first element if it is a string (hack to fix extra whitespace prob: should be fixed in platypus)
        if pcontent:
            pc0 = pcontent[0]
            if type(pc0) is StringType:
                pc1 = pc0.strip()
                if pc0[-1] in string.whitespace:
                    pc1 = pc1+" "
                pc0 = pc1
                pcontent[0] = pc0
        # get the text for the paragraph
        text = MapNode.evaluate(self, "para", localdict, pcontent, extra, context)
        #text = quoteText(text)
        #P = LazyParagraph(text, style, bulletText=bulletText)
        P = DeferredInitialization(LazyParagraph, text, style, bulletText=bulletText)
        #context.story.append(P)
        return P
Controller["para"] = para()

class h1(para):
    def __init__(self, defaultStyle="h1.defaultStyle"):
        para.__init__(self)
        self.defaultStyle = defaultStyle
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # sdict should be empty: reset it
        sdict = {"style": self.defaultStyle }
        return para.evaluate(self, tagname, sdict, pcontent, extra, context)


Controller["h1"] = h1()
Controller["h2"] = h1("h2.defaultStyle")
Controller["h3"] = h1("h3.defaultStyle")
Controller["title"] = h1("title.defaultStyle")

class blockTable(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):


        #first content element MIGHT be a style; is so we
        #use it, if not we may look for a style in the context.
        if len(pcontent) > 0 and isinstance(pcontent[0], TableStyle):
            data = pcontent[1:]
            style = pcontent[0]
        elif sdict.has_key("style"):
            data = pcontent
            stylename = sdict["style"]
            style = context[stylename]
        else:
            style = None
            data = pcontent

        #if the new first element is bulkData, grab its output as the
        #data; otherwise pcontent is just the data - a collection of
        #rows and cells which have done parsing themselves already.
        if len(pcontent) > 0 and isinstance(pcontent[0], BulkDataHolder):
            data = pcontent[0].value



        colWidths = None
        rowHeights = None
        splitByRow = 1
        repeatRows = 0
        repeatCols = 0
        # ... now change the defaults
        if sdict.has_key("colWidths"):
            colWidths = sdict["colWidths"]
            colWidths = lengthSequence(colWidths,converter=lambda x,context=context: readLength(x,percentTotal='%',context=context))
        if sdict.has_key("rowHeights"):
            rowHeights = sdict["rowHeights"]
            rowHeights = lengthSequence(rowHeights)
        if sdict.has_key("splitByRow"):
            splitByRow = int(sdict["splitByRow"])
        if sdict.has_key("repeatRows"):
            repeatRows = int(sdict["repeatRows"])
        if sdict.has_key("repeatCols"):
            repeatCols = int(sdict["repeatCols"])
        # XXXXXXX
        # make a default widths, if none specified # THIS IS HACKED BIGTIME!!! MUST BE DONE LAZILY!!!
        if 0 and data and style is None:
            nrows = len(data)
            ncols = len(data[0])
            pagewidth = context.pageSize[0]
            # TOTAL HACK!
            usewidth = pagewidth/2.0
            eachwidth = usewidth/ncols
            colWidths = [eachwidth]*ncols
##        T = xTable(data=data, colWidths=colWidths, rowHeights=rowHeights,
##                  style=style, splitByRow=splitByRow,
##                  repeatRows=repeatRows, repeatCols=repeatCols)
        T = DeferredTable(data=data, colWidths=colWidths, rowHeights=rowHeights,
                  style=style, splitByRow=splitByRow,
                  repeatRows=repeatRows, repeatCols=repeatCols)
        #context.story.append(T)
        return T

class StyledTD:
    def __init__(self,value,style_cmds):
        self.value = value
        self.style_cmds = style_cmds

class DeferredTable(DeferredInitialization):
    klass = [xTable] # for debugging
    def __init__(self, data, *args, **kw):
        self.data = data
        self.args = args
        self.kw = kw
    def initialize(self):
        data = self.data
        dataout = []
        dataout_append = dataout.append
        i = 0
        cmds = []
        cmds_append = cmds.append
        for row in data:
            rowout = []
            rowout_append = rowout.append
            j = 0
            for datum in row:
                if isinstance(datum,StyledTD):
                    IJ = (j,i), (j,i)
                    for cmd in datum.style_cmds:
                        cmds_append((cmd[0],)+IJ+(cmd[1],))
                    datum = datum.value
                if type(datum) is ListType:
                    #list of flowables
                    outdatum = []
                    for thing in datum:
                        if isinstance(thing, DeferredInitialization):
                            thing = thing.initialize()
                        outdatum.append(thing)
                    datum = outdatum
                elif isinstance(datum, DeferredInitialization):
                    #single deferred flowable
                    datum = datum.initialize()
                #else simple string
                rowout_append(datum)
                j += 1
            dataout_append(rowout)
            i += 1
        kw = self.kw
        if cmds:
            S = kw.get('style',None)
            kw['style'] = TableStyle(cmds,parent=S)
        T = xTable(*((dataout,)+tuple(self.args)), **kw)
        return T
Controller["blockTable"] = blockTable()

class tr(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # just return the row itself
        #print "row", pcontent
        #stop
        return pcontent

Controller["tr"] = tr()

_tdAMap = {
    'fontName': ('FACE',readString),
    'fontSize': ('SIZE',readLength),
    'fontColor': ('TEXTCOLOR',readColor),
    'leading': ('LEADING',readLength),
    'leftPadding': ('LEFTPADDING',readLength),
    'rightPadding': ('RIGHTPADDING',readLength),
    'topPadding': ('TOPPADDING',readLength),
    'bottomPadding': ('BOTTOMPADDING',readLength),
    'background': ('BACKGROUND',readColor),
    'align': ('ALIGN',readAlignU),
    'vAlign': ('VALIGN',readvAlignU),
    }

class td(NoWhiteContentMixin, MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # clean out the pcontent
        # STRIP ANY STRINGS, and do entity substitutions

        L = []
        #for now only support either a sequence of (lazy) strings or a single flowable
        hasFlowable = hasLazyString = None
        for x in pcontent:
            if isinstance(x, Flowable) or isinstance(x, DeferredInitialization): hasFlowable = 1
            elif isinstance(x, LazyStringForm): hasLazyString = 1
            elif type(x)==StringType:
                hasLazyString = 1
                x = x.strip()
                for (a,b) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
                    x = x.replace(a,b)
            L.append(x)
        lenL = len(L)
        if lenL==1: L = L[0]
        elif lenL==0: L = ""
        if hasLazyString:
            if hasFlowable: raise ValueError, "<td>cell tag cannot include both strings and flowables: "+repr(L)
            #strings only
            if lenL>1: L = LazyJoin(L)
        if sdict:
            #we have local attributes
            cmds = []
            a = cmds.append
            _readColor = lambda x,c=context: getColor(x,c)
            _readLength = lambda x,c=context: readLength(x,context=c)
            for k, v in sdict.items():
                K, f = _tdAMap[k]
                if f is readColor: f=_readColor
                elif f is readLength: f=_readLength
                a((K,f(v)))
            L = StyledTD(L,cmds)
        return L
Controller["td"] = td()

class BulkDataHolder:
    def __init__(self, value):
        self.value = value

class bulkDataParser(MapNode):
    "This holds field and row separated data to save tr/td tags"
    def evaluate(self, tagname, sdict, pcontent, extra, context):

        stripBlock = (sdict.get("stripBlock","yes") <> "no")
        stripRows = (sdict.get("stripRows","yes") <> "no")
        stripFields = (sdict.get("stripFields","no") <> "no")

        fieldDelim = sdict.get("fieldDelim", ",")
        #allow escaped tabs
        if fieldDelim == '\\t':
            fieldDelim = '\t'

        rowDelim = sdict.get("rowDelim", "\n")


        if stripBlock:
            raw = pcontent[0].strip()
        else:
            raw = pcontent[0]
        lines = raw.split(rowDelim)
        output = []
        for line in lines:
            if stripRows:
                line = line.strip()
            row = line.split(fieldDelim)
            if stripFields:
                row = map(lambda x:x.strip(), row)
            output.append(row)
        #we return the data wrapper, so that the enclosing table can
        #look for a bulkData object and not a list, which is a bit
        #ambiguous.
        return BulkDataHolder(output)
Controller["bulkData"] = bulkDataParser()

class spacer(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        length = readLength(sdict["length"],context=context)
        width = 1
        if sdict.has_key("width"):
            width = readLength(sdict["width"],context=context)
        return Spacer(width, length)
Controller["spacer"] = spacer()

class condPageBreak(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        length = readLength(sdict["height"],context=context)
        return CondPageBreak(length)
Controller["condPageBreak"] = condPageBreak()

###########################################################################
# Form and bar code widgets an support
###########################################################################



class BoxStyle(PropertySet):
    "Attribute collection used in form elements like checkboxes etc."

    defaults = {
        'fontName':'Courier-Bold',
        'fontSize':10,
        'alignment':TA_LEFT,
        'textColor':black,
        'labelFontName':'Times-Roman',
        'labelFontSize':8,
        'labelAlignment':TA_LEFT,
        'labelTextColor':black,
        'labelOffsetX':0,
        'labelOffsetY':0.2*cm,
        'fillColor':white,
        'strokeColor':black,
        'cellWidth':0.5*cm,
        'cellHeight':0.5*cm,
        }

    def fromStringDict(self, strDict, context):
        "converts string data to numbers, colors etc."
        # a big help when parsing...
        # not that if propertySet had some of th behavior
        # of the validators in graphics, we would
        # not even need this and propertySet could infer
        # it from a type catalog.

        # if it is not a string, assume parsed already and pass
        # through; RML handles color names separately
        localdict = strDict.copy()

        for (key, strValue) in strDict.items():
            if key in ("name", "parent"):
                # we already have those two form construction
                continue
            elif type(key) <> StringType:
                value = strValue
            elif key in ("fontSize",
                        "labelFontSize",
                        "cellWidth",
                        "cellHeight",
                        "labelOffsetX",
                        "labelOffsetY",
                        ):
                value = readLength(strValue,context=context)
            elif key in ("textColor","labelTextColor"):
                value = readColor(strValue)
            elif key in ("alignment",):
                value = readAlignment(strValue)
            else:
                value = strValue
            setattr(self, key, value)


class boxStyle(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        # need the parent if there is one
        #print '\n <boxStyle...>'
        if sdict.has_key("parent"):
            #print 'parent name %s' % sdict["parent"]
            parentStyle = context[sdict["parent"]]
        else:
            #print 'no parent found'
            parentStyle = None

        if not sdict.has_key("name"):
            raise ValueError, "name required for styles"
        styleName = sdict["name"]

        newstyle = BoxStyle(name=styleName, parent=parentStyle)
        newstyle.fromStringDict(sdict, context)
        #print 'created BoxStyle:'
        #newstyle.listAttrs()

        context[styleName] = newstyle
        return "" # vanish

Controller["boxStyle"] =  boxStyle()


class barCode(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        x = readLength(sdict["x"],context=context)
        y = readLength(sdict["y"],context=context)
        code = sdict["code"]
        value = pcontent[0]
        def barCodeOp(canv, doc,
                      x=x,
                      y=y,
                      code=code,
                      value=value):
            # Load all possible barcode modules.
            try:
                from rlextra.graphics.barcode import common
                from rlextra.graphics.barcode import code39
                from rlextra.graphics.barcode import code93
                from rlextra.graphics.barcode import code128
                from rlextra.graphics.barcode import usps
            except ImportError:
                try:
                    # try the canonical location
                    from reportlab.extensions.barcode import common
                    from reportlab.extensions.barcode import code39
                    from reportlab.extensions.barcode import code93
                    from reportlab.extensions.barcode import code128
                    from reportlab.extensions.barcode import usps
                except ImportError:
                    try:
                        # get it from other places, like site-packages
                        from barcode import common
                        from barcode import code39
                        from barcode import code93
                        from barcode import code128
                        from barcode import usps
                    except ImportError:
                        return

            # Find the one we want and draw it if found.
            # (should draw a dummy blank one if nothing found...?)
            for module in [common, code39, code93, code128, usps]:
                if code in module.__dict__.keys():
                    Barcode = module.__dict__[code]
                    b = Barcode(value)
                    b.drawOn(canv, x, y)

        return barCodeOp
Controller['barCode'] = barCode()

def extractStyledAttributes(sdict, attrMap, context):
    """Helper for the three box objects.

    The attribute map defines type conversions
    and defaults.  It first sees if the object
    has a 'style' attribute, and if so it looks
    up the style object.  All properties are
    then gathered in a new dictionary which
    gives priority to (1) the tag, (2) the
    associated style if any, and (3) attrMap
    defaults."""
    from reportlab.lib import colors
    from reportlab.lib.units import cm

    #is there a style name? if so, get the style object.
    if sdict.has_key("style"):
        styleName = sdict["style"]
        if styleName == '':
            style = None
        else:
            try:
                style = context[styleName]
                assert isinstance(style, BoxStyle), "LetterBox object has a style '%s' which is not a BoxStyle" % style.name
            except KeyError:
                style = None
    else:
        style = None

    # gather up the data into a dictionary.
    # try object, then style, then default.  This
    # could be in a subroutine.
    d = {}
    _readColor = lambda x,c=context: getColor(x,c)
    _readLength = lambda x,c=context: readLength(x,context=c)
    for (attrName, readerFunc, default) in attrMap:
        try:
            if readerFunc is readLength: readerFunc = _readLength
            elif readerFunc is readColor: readerFunc = _readColor
            value = readerFunc(sdict[attrName])
        except KeyError:
            if style is None:
                value = default
            else:
                try:
                    value = getattr(style, attrName)
                except AttributeError:
                    value = default

        d[attrName] = value

    return d

class textBox(MapNode):
    """A textbox with multi-line text, possibly shrinked to fit."""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        from reportlab.lib import colors
        from reportlab.lib.units import cm
##  style CDATA ""
##  x CDATA #REQUIRED
##  y CDATA #REQUIRED
##  boxWidth CDATA #REQUIRED
##  boxHeight CDATA #REQUIRED
##  boxStrokeColor CDATA #IMPLIED
##  boxFillColor CDATA #IMPLIED
##  lineWidth CDATA #IMPLIED
##  fontName CDATA #IMPLIED
##  fontSize CDATA #IMPLIED
##  align CDATA #IMPLIED
##  shrinkToFit CDATA #IMPLIED
##  label CDATA #IMPLIED
##>

        attrMap = [('x', readLength, 0.0),
                   ('y', readLength, 0.0),
                   ('count', int, 10),
                   ('boxWidth', readLength, 0.5*cm),
                   ('boxHeight', readLength, 0.6*cm),
                   ('boxStrokeColor', readColor, colors.black),
                   ('boxFillColor', readColor,colors.white),
                   ('textColor', readColor,colors.black),
                   ('lineWidth', readLength, 1) ,
                   ('fontName', str, 'Courier-Bold'),
                   ('fontSize', readLength, 10),
                   ('align', readAlign, 'left'),
                   ('vAlign', readvAlign, 'top'),
                   ('shrinkToFit', readBool, 11),
                   ('label', str, ''),
                   ('labelFontName', str, 'Helvetica'),
                   ('labelFontSize', readLength, 8),
                   ('labelTextColor', readColor,colors.black),
                   ('labelOffsetX', readLength,0),
                   ('labelOffsetY', readLength,0.2*cm),
                    ]

        d = extractStyledAttributes(sdict, attrMap, context)
        d['data'] = LazyJoin(pcontent)

        def textBoxOp(canv, doc,
                      x=d['x'],
                      y=d['y'],
                      boxWidth=d['boxWidth'],
                      boxHeight=d['boxHeight'],
                      boxStrokeColor=d['boxStrokeColor'],
                      boxFillColor=d['boxFillColor'],
                      textColor=d['textColor'],
                      lineWidth=d['lineWidth'],
                      fontName=d['fontName'],
                      fontSize=d['fontSize'],
                      align=d['align'],
                      vAlign=d['vAlign'],
                      shrinkToFit=d['shrinkToFit'],
                      label=d['label'],
                      labelFontName=d['labelFontName'],
                      labelFontSize=d['labelFontSize'],
                      labelTextColor=d['labelTextColor'],
                      labelOffsetX=d['labelOffsetX'],
                      labelOffsetY=d['labelOffsetY'],
                      text=d['data']):
            #raise "death", text
            from reportlab.lib import colors
            from reportlab.lib.units import cm


            text = unEscapeDrawString(str(text))

            # box
            canv.saveState()
            fill = boxFillColor is not None
            if fill: canv.setFillColor(boxFillColor)
            stroke = boxStrokeColor is not None
            if stroke:
                canv.setStrokeColor(boxStrokeColor)
                canv.setLineWidth(lineWidth)
            if stroke or fill:
                canv.rect(x, y, boxWidth, boxHeight, stroke=stroke, fill=fill)

            # label
            if label:
                canv.setFillColor(labelTextColor)
                canv.setStrokeColor(colors.black)
                canv.setFont(labelFontName, labelFontSize)
                for (a,b) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
                    label = label.replace(a,b)
                canv.drawString(x+labelOffsetX, y + boxHeight + labelOffsetY, label)

            # Split into lines and find longest one.
            from reportlab.pdfbase.pdfmetrics import stringWidth
            lines = str(text).split('\n')
            numLines = len(lines)
            maxStringWidth = 0
            for line in lines:
                m = stringWidth(line, fontName, fontSize)
                if maxStringWidth < m:
                    maxStringWidth = m

            distanceToBorder = 1 + lineWidth/2.0
            if shrinkToFit:
                # Shrink fontsize such that text fits in box.
                dtb2 = 2*distanceToBorder
                sw = maxStringWidth
                if sw > boxWidth - dtb2:
                    fontSize = fontSize * (boxWidth-dtb2)/sw
                if numLines > 1 and 1.2*fontSize*numLines > boxHeight - dtb2:
                    fontSize = fontSize * (boxHeight-dtb2)/(1.2*fontSize*numLines)

            # Draw text.
            canv.setStrokeColor(colors.black)
            canv.setFillColor(textColor)
            canv.setFont(fontName, fontSize)
            from reportlab.pdfbase import pdfmetrics
            n = len(lines)
            if vAlign=='bottom':
                font = pdfmetrics._fonts[fontName]
                descent = font.face.descent*fontSize/1000.
                yt0 = y - descent + (0.2+(n-1)*1.2)*fontSize
            elif vAlign=='middle':
                font = pdfmetrics._fonts[fontName]
                k = n/2+1
                descent = font.face.descent*fontSize/1000.
                ascent = font.face.ascent*fontSize/1000.
                yt0 = y+boxHeight/2.+int(n/2)*fontSize*1.2 -((ascent-descent)/2.+descent)
            else:   #top
                yt0 = y + boxHeight - fontSize
            for i in xrange(n):
                line = lines[i]
                yt = yt0 - 1.2*fontSize*i
                if align == 'left':
                    xt = x + distanceToBorder
                    canv.drawString(xt, yt, line)
                elif align == 'right':
                    xt = x + boxWidth - distanceToBorder
                    canv.drawRightString(xt, yt, line)
                elif align == 'center':
                    xt = x + boxWidth/2.0
                    canv.drawCentredString(xt, yt, line)

            canv.restoreState()

        return textBoxOp

Controller['textBox'] = textBox()


class checkBox(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        from reportlab.lib import colors
        from reportlab.lib.units import cm

        # (name, reader, default)
        attrMap = [('x', readLength, 0.0),
                   ('y', readLength, 0.0),
                   ('labelFontName', str, 'Helvetica'),
                   ('labelFontSize', readLength, 8),
                   ('labelTextColor', readColor,colors.black),
                   ('boxWidth', readLength, 0.5*cm),
                   ('boxHeight', readLength, 0.6*cm),
                   ('checkStrokeColor', readColor, colors.black),
                   ('boxStrokeColor', readColor, colors.black),
                   ('boxFillColor', readColor,colors.white),
                   ('lineWidth', readLength, 1) ,
                   ('line1', str, ''),
                   ('line2', str, ''),
                   ('line3', str, ''),
                   ('checked', readBool, 0),
                   ('bold', readBool, 0),
                   ('graphicOn', str, None),
                   ('graphicOff', str, None),
                   ]
        # this gives us the attributes to actually use,
        # after any style lookups and default substitutions
        d = extractStyledAttributes(sdict, attrMap, context)


        def checkBoxOp(canv, doc,
                       x=d['x'],
                       y=d['y'],
                       labelFontName=d['labelFontName'],
                       labelFontSize=d['labelFontSize'],
                       labelTextColor=d['labelTextColor'],
                       boxWidth=d['boxWidth'],
                       boxHeight=d['boxHeight'],
                       checkStrokeColor=d['checkStrokeColor'],
                       boxStrokeColor=d['boxStrokeColor'],
                       boxFillColor=d['boxFillColor'],
                       lineWidth=d['lineWidth'],
                       line1=d['line1'],
                       line2=d['line2'],
                       line3=d['line3'],
                       checked=d['checked'],
                       bold=d['bold'],
                       graphicOn=d['graphicOn'],
                       graphicOff=d['graphicOff'],
                       ):

            from reportlab.lib import colors
            from reportlab.lib.units import cm

            canv.saveState()

            # either we draw the box, or use the provided
            # pair of bitmaps
            if (graphicOn is not None or graphicOff is not None):
                assert (graphicOn and graphicOff), "a pair of graphics/bitmaps must be provided even if you only want to use one of them!"
                doGraphic = 1
            else:
                doGraphic = 0

            if doGraphic:
                if checked:
                    canv.drawImage(graphicOn, x, y, boxWidth, boxHeight)
                else:
                    canv.drawImage(graphicOff, x, y, boxWidth, boxHeight)
            else:  #draw a simple box ourselves
                fill = boxFillColor is not None
                stroke = boxStrokeColor is not None
                if stroke: canv.setStrokeColor(boxStrokeColor)
                if fill: canv.setFillColor(boxFillColor)
                canv.setLineWidth(lineWidth)
                if stroke or fill: canv.rect(x, y, boxWidth, boxHeight, stroke=stroke, fill=fill)
                if checked:
                    if checkStrokeColor:
                        canv.setStrokeColor(checkStrokeColor)
                    else:
                        canv.setStrokeColor(boxStrokeColor)
                    canv.line(x+1, y+1, x+boxWidth-1, y+boxHeight-1)
                    canv.line(x+1, y+boxHeight-1, x+boxWidth-1, y+1)

            # text - up to three lines right to check box
            linesUsed = 0
            if line1: linesUsed = linesUsed + 1
            if line2: linesUsed = linesUsed + 1
            if line3: linesUsed = linesUsed + 1
            #manually specify line positions
            canv.setFillColor(labelTextColor)
            fs = 8
            if bold:
                canv.setFont(labelFontName, labelFontSize)
            else:
                canv.setFont("Helvetica", fs)
            x1 = x + boxWidth + 0.1*cm
            ds = canv.drawString
            if linesUsed:
                lines = filter(lambda x:x != '', [line1, line2, line3])
                lines.reverse()
                for i in range(len(lines)):
                    line = lines[i]
                    yt = y + boxWidth/2.0 - (fs*linesUsed + fs*0.2*(linesUsed-1))/2.0 + i*fs*1.2
                    for (a,b) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
                        line = line.replace(a,b)
                    ds(x1, yt, line)

            canv.restoreState()

        return checkBoxOp

Controller['checkBox'] = checkBox()


class letterBoxes(MapNode):
    "Box grid with one letter per box"
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        from reportlab.lib import colors
        from reportlab.lib.units import cm

        # (name, reader, default)
        # should correspond to the DTD.
        attrMap = [('x', readLength, 0.0),
                   ('y', readLength, 0.0),
                   ('count', int, 10),
                   ('label', str, ''),
                   ('labelFontName', str, 'Helvetica'),
                   ('labelFontSize', readLength, 8),
                   ('labelTextColor', readColor,colors.black),
                   ('labelOffsetX', readLength,0),
                   ('labelOffsetY', readLength,0.2*cm),
                   ('boxWidth', readLength, 0.5*cm),
                   ('boxGap',readLength, 0),
                   ('boxHeight', readLength, 0.6*cm),
                   ('combHeight',float, 1),
                   ('boxStrokeColor', readColor, colors.black),
                   ('boxFillColor', readColor,colors.white),
                   ('lineWidth', readLength, 1) ,
                   ('fontName', str, 'Courier-Bold'),
                   ('fontSize', readLength, 10),
                   ]

        d = extractStyledAttributes(sdict, attrMap, context)
        try:
            d['data'] = pcontent[0]
        except:
            d['data'] = ''

        # embedded drawing function specific to this
        # text box.
        def letterBoxesOp(canv, doc,
                          x=d['x'],
                          y=d['y'],
                          count=d['count'],
                          label=d['label'],
                          labelFontName = d['labelFontName'],
                          labelFontSize = d['labelFontSize'],
                          labelTextColor = d['labelTextColor'],
                          labelOffsetX=d['labelOffsetX'],
                          labelOffsetY=d['labelOffsetY'],
                          boxWidth=d['boxWidth'],
                          boxHeight=d['boxHeight'],
                          boxGap=d['boxGap'],
                          combHeight=d['combHeight'],
                          boxStrokeColor=d['boxStrokeColor'],
                          boxFillColor=d['boxFillColor'],
                          lineWidth=d['lineWidth'],
                          fontName=d['fontName'],
                          fontSize=d['fontSize'],
                          data=d['data']
                          ):

            from reportlab.lib import colors
            from reportlab.lib.units import cm

            canv.saveState()

            fill = boxFillColor is not None
            stroke = boxStrokeColor is not None
            if stroke:
                canv.setStrokeColor(boxStrokeColor)
                canv.setLineWidth(lineWidth)
            if fill: canv.setFillColor(boxFillColor)

            xpoints = [x]
            y1 = y+combHeight*boxHeight
            if boxGap>0:
                y2 = y+boxHeight
                for i in xrange(count):
                    thisX = x+(i+1)*(boxWidth+boxGap)
                    x0 = xpoints[-1]
                    x1 = thisX - boxGap
                    xpoints.append(thisX)
                    if combHeight==1:
                        if fill or stroke:
                            canv.rect(x0, y, boxWidth, boxHeight, fill=fill, stroke=stroke)
                    else:
                        if fill: canv.rect(x0, y, boxWidth, boxHeight, fill=fill, stroke=0)
                        if stroke:
                            canv.line(x0, y, x0, y1)
                            canv.line(x1, y, x1, y1)
                            canv.line(x0, y, x1, y)
                            canv.line(x0, y2, x1, y2)
            else:
                # rectangle
                if stroke or fill: canv.rect(x, y, boxWidth*count, boxHeight, fill=fill, stroke=stroke)

                # blue grid
                for i in xrange(count):
                    thisX = x+(i+1)*(boxWidth+boxGap)
                    xpoints.append(thisX)
                    if stroke: canv.line(thisX, y, thisX, y1)
                    #xpoints.append(x+(i+1)*boxWidth)
            ypoints = [y, y+boxHeight]
            #canv.grid(xpoints, ypoints)

            # label
            if label:
                canv.setFillColor(labelTextColor)
                canv.setFont(labelFontName, labelFontSize)
                for (a,b) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
                    label = label.replace(a,b)
                canv.drawString(x+labelOffsetX, y + boxHeight + labelOffsetY, label)

            # data
            if data:
                #&amp; and friends to be displayed as expected
                for (old, new) in ENTITY_SUBSTITUTIONS_DRAWSTRING:
                    data = data.replace(old, new)

                # Shrink fontsize such that text fits in box.
                from reportlab.pdfbase.pdfmetrics import stringWidth
                maxStringWidth = 0
                for d in data:
                    m = stringWidth(d, fontName, fontSize)
                    if maxStringWidth < m:
                        maxStringWidth = m

                # correct fontSize
                distanceToBorder = 1 + lineWidth/2.0
                dtb2 = 2*distanceToBorder
                sw = maxStringWidth
                if sw > boxWidth - dtb2:
                    fontSize = fontSize * (boxWidth-dtb2)/sw

                # fill in text in boxes
                canv.setFillColor(colors.black)
                data = data[0:count]  # data might be too long
                canv.setFont(fontName, fontSize)
                yt = y + boxHeight - fontSize
                for i in range(len(data)):
                    xt = xpoints[i] + boxWidth*0.5
                    canv.drawCentredString(xt, yt, data[i])

            canv.restoreState()

        return letterBoxesOp

Controller['letterBoxes'] = letterBoxes()

###########################################################################
# End of... form stuff
###########################################################################


###########################################################################
# easier way to hook up to graphics....
###########################################################################

class Param(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        key = sdict['name']
        value = unEscapeDrawString(pcontent[0])
        try:
            value = eval(value)
        except:
            value = value.strip()
        return ("setParam", key, value)

Controller["param"] = Param()


class DiagraWidgetFactory(MapNode):
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        moduleName = sdict["module"]
        funcName = sdict["function"]
        alias = sdict["name"]
        # the path for the imports should include:
        # 1. document directory
        # 2. python path if baseDir not given, or
        # 3. baseDir if given
        try:
            dirName = sdict["baseDir"]
        except:
            dirName = None
        importPath = _rml2pdf_locations()
        if dirName is None:
            importPath.extend(sys.path)
        else:
            importPath.insert(0, dirName)

        modul = recursiveImport(moduleName, baseDir=importPath)
        initArgs = eval(sdict.get("initargs", '[]'))
        widget = getattr(modul, funcName)(*initArgs)
        #print 'created a widget called %s with args %s' % (widget.__class__.__name__, str(initArgs))
        return ("addWidget", widget, alias)

Controller["widget"] = DiagraWidgetFactory()

class DiagraFlowableWrapper(MapNode):
    """Call a standard drawing module"""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        moduleName = sdict["module"]
        funcName = sdict["function"]

        showBoundary = int(sdict.get("showBoundary", "0"))

        hAlign = sdict.get("hAlign", "CENTER")


        # the path for the imports should include:
        # 1. document directory
        # 2. python path if baseDir not given, or
        # 3. baseDir if given
        try:
            dirName = sdict["baseDir"]
        except:
            dirName = None
        importPath = _rml2pdf_locations()
        if dirName is None:
            importPath.extend(sys.path)
        else:
            importPath.insert(0, dirName)

        modul = recursiveImport(moduleName, baseDir=importPath)
        func = getattr(modul, funcName)
        drawing = func()

        drawing.hAlign = hAlign
        if showBoundary:
            drawing._showBoundary = 1

        #pcontent is the children, but they have been
        #traversed into already
        if pcontent is not None:
            for thing in pcontent:
                if type(thing) is TupleType:
                    if thing[0] == 'setParam':
                        (command, key, value) = thing
                        recursiveSetAttr(drawing, key, value)
                    elif thing[0] == 'addWidget':
                        (command, widget, alias) = thing
                        drawing.add(widget, alias)
                    else:
                        raise ValueError, "unknown Diagra command verb %s" % thing[0]
        # as a high-level convenience, it can also return one of
        # our figure objects, which wrap up drawings and let you shrink/scale
        # them and add captions


        return drawing

Controller["drawing"] = DiagraFlowableWrapper()

#######################################################################
#
#  PageCatcher flowable support. Need a flowable which does all
#  the storyplace stuff.
#
########################################################################


class IncludePdfPages(MapNode):
    """PDF Pages to be included inline within the story"""
    def expandPageNumbers(self, txt):
        "Convert expression like '1,3,7-10,12' to a list of numbers"
        chunks = txt.split(',')
        pages = []
        for chunk in chunks:
            bits = chunk.split('-')
            if len(bits) == 1:
                pages.append(int(chunk))
            elif len(bits) == 2:
                start = int(bits[0])
                end = int(bits[1])
                for pg in range(start, end+1):
                    pages.append(pg)
            else:
                pass
        return pages

    def evaluate(self, tagname, sdict, pcontent, extra, context):
        from rlextra.pageCatcher.pageCatcher import storeFormsInMemory
        from rlextra.pageCatcher.pageflow import LoadPdfFlowable, ShowPdfFlowable
        fileName = sdict["filename"]
        strPageList = sdict.get("pages", None)
        templateName = sdict.get("template", None)

        outlineText = sdict.get("outlineText", None)
        outlineLevel = int(sdict.get("outlineLevel", 0))
        outlineClosed= int(sdict.get("outlineClosed", 0))
        leadingFrame = readBool(sdict.get('leadingFrame','1'))


        if strPageList is None:
            pageNumbers = None
        else:
            pageNumbers = self.expandPageNumbers(strPageList)

        # this one is unusual in that it returns a list of objects to
        # go into the story.
        output = []

        if templateName:
            output.append(NextPageTemplate(templateName))


        #read in the PDF file right now and get the pickled object
        # and names
        pdfContent = open_and_read(fileName)
        prefix = os.path.splitext(fileName)[0].replace('/','_') + '_page'
        (formNames, pickledStuff) = storeFormsInMemory(
                pdfContent,
                prefix=prefix,
                all=1,
                verbose=0,
                BBoxes=0,
                extractText=0,
                fformname=None)

        #if explicit pages requested, slim it down.
        if pageNumbers:
            newNames = []
            for pgNo in pageNumbers:
                newNames.append(formNames[pgNo-1])
            formNames = newNames

        #make object 1 for story
        loader = LoadPdfFlowable(pickledStuff)
        output.append(loader)

        #now do first page.  This is special as it might
        #have an outline
        formName = formNames[0]
        if leadingFrame: output.append(PageBreak())
        if outlineText:
            output.append(OutlineEntry(outlineLevel, outlineText, outlineClosed))
        output.append(ShowPdfFlowable(formName))

        #now make a shower for each laterpage, and a page break
        for formName in formNames[1:]:
            output.append(PageBreak())
            output.append(ShowPdfFlowable(formName))

        return output

Controller["includePdfPages"] = IncludePdfPages()

###########################################################################
# dynamic behaviour
###########################################################################

def loseText(content):
    out = []
    for elem in content:
        if type(elem) <> type(''):
            out.append(elem)
    return out

class ConditionalFlowable(MapNode, Flowable):
    """This wraps up and selects conditional content."""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        """This returns a conditional flowable object - itself, actually -
        which when drawn decides what content it is."""
        print 'conditional tag arrived'
        self.cases = []
        self.elseCase = None
        elseCase = None
        for elem in pcontent:
            if type(elem) == type(''):
                continue
            elif isinstance(elem, Case):
                print 'found a case object'
                self.cases.append(elem)
            elif isinstance(elem, Else):
                print 'found an else object'
                self.elseCase = elem
        return self

##        print 'evaluating cases...'
##        nameSpace = {
##            'pageNumber':context.doc.page
##            }
##        nameSpace.update(context.names)
##
##        #skip out and return first true one
##        for case in cases:
##            if eval(case.expr):
##                print 'case %s evaluated true' % case.expr
##                return case.content
##            else:
##                print 'case %s evaluated false' % case.expr
##
##        if elseCase:
##            print 'else case present'
##            return elseCase.content
##
##        #still here
##        return None


Controller["condFlowable"] = ConditionalFlowable()

class Case(MapNode):
    """This wraps up and selects conditional content."""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        self.expr = sdict["expr"]
        self.content = loseText(pcontent)
        return self

Controller["case"] = Case()

class Else(MapNode):
    """Returns its content if called."""
    def evaluate(self, tagname, sdict, pcontent, extra, context):
        self.content = loseText(pcontent)
        return self

Controller["else"] = Else()


def getoutfile(parsedtext):
    if type(parsedtext) is StringType:
        return None
    (name, atts, content, misc) = parsedtext
    if name=="document":
        #print atts
        return atts["filename"]
    else:
        if content is None:
            return None
        for e in content:
            t = getoutfile(e)
            if t is not None:
                return t
        return None

def _vStr2DTDName(s):
    '''Map a version string to an rmldtd filename'''
    return 'rml_%s.dtd' % s.replace('.','_')

DTDName=_vStr2DTDName(version)
CompatibleDTDNames = ['dynamic_rml.dtd']+map(_vStr2DTDName,compatible_versions)

from rlextra.preppy import preppy
def go(xmlInputText, outputFileName=None, outDir=None, dtdDir=None,
       passLimit=2, permitEvaluations=1, ignoreDefaults=0,
       pageCallBack=None,
       progressCallBack=None,
       preppyDictionary=None, preppyIterations=1,
       dynamicRml=0, dynamicRmlNameSpace={},
       encryption=None,
       verbose=0):
    # passLimit of None means "keep trying until done"
    #   of 3 means, "try 3 times then quit"
    # permitEvaluations when false disallows the evalString tag for security (web apps, eg)
    # ignoreDefaults=1 means "do one pass and use the default values where values are not found"
    # pageCallBack is a callback to execute on final formatting of each page.
    # progressCallBack is a cleverer callback; see the progressCB function
    # in reportlab\platypus\doctemplate
    # preppyDictionary if set to a dictionary indicates that the xmlInputText should be
    #  preprocessed using preppy with the preppyDictionary as argument.
    # if preppyDictionary is not None and preppyIterations is >1 then the preppy preprocessing will
    #  be repeated preppyIterations times (max of 3) with the same dict,
    #  to generate, eg, table of contents...
    # If encryption is set it must be an encryption object.
    # For example
    # rlextra.utils.pdfencrypt.StandardEncryption("User", "Owner", canPrint=0, canModify=0, canCopy=0, canAnnotate=0)

    # DO PREPPY PREPROCESSING IF REQUESTED
    if progressCallBack:
        progressCallBack('STARTED', 0)

    if preppyDictionary is not None:
        if preppyIterations not in (1,2,3):
            raise ValueError, "preppy iterations (%s) must be 1,2, or 3" % preppyIterations
        if verbose:
            print "preprocessing the text"
        # get the preppy module derived from the xml text
        template = preppy.getModule("rml_template", sourcetext=xmlInputText)
        if verbose:
            print "prep: template made"
        # iterate the template the number of times requested
        pD = preppyDictionary.copy()
        for preppyCount in range(preppyIterations):
            preppyOutput = getStringIO()
            template.run(pD, outputfile = preppyOutput)
            if verbose:
                print "prep: done with prep", preppyCount
        # replace the input text with the preprocessor output
        xmlInputText = preppyOutput.getvalue()

    if progressCallBack:
        progressCallBack('BEFORE_PARSE_XML', 0)

    # PARSE THE XML
    if verbose:
        print "parsing XML"
    from reportlab.lib.rparsexml import parsexml

    if dtdDir is None: dtdDir = _rml_dtd_dirs()
    else:
        if not callable(dtdDir): dtdDir = _FS_Checker(dtdDir)
        if type(dtdDir) not in (type(()),type([])): dtdDir = [dtdDir]
    dtd_list = []
    def _eoDTD(s,possibles=CompatibleDTDNames,target=DTDName,dtdDir=dtdDir,dtd_list=dtd_list,path_basename=os.path.basename):
        if path_basename(s) in possibles:
            t = 'dynamic_rml.dtd'
            if not s.endswith(t): t = target
            if dtdDir:
                for d in dtdDir:
                    s = d(t)
                    if s:
                        dtd_list.append(t)
                        return s
            else:
                if s!=t: s = t
        return s

    # need to extract the filename by tricky methods
    # because we need the document defined *before* processing
    parsed = parsexml(xmlInputText,eoCB=_eoDTD,entityReplacer=None)
    if progressCallBack:
        progressCallBack('AFTER_PARSE_XML', 0)

    if dynamicRml or 'dynamic_rml.dtd' in dtd_list:
        #do a preprocessing pass on tuple tree.
        from rlextra.radxml import xpreppy
        parsed = xpreppy.preProcess(parsed, dynamicRmlNameSpace)

    # DETERMINE THE OUTPUT FILE (xxxx IS THIS HISTORICAL?)
    if not outputFileName:
        outputFileName = getoutfile(parsed)
        #print "output file is", outputFileName
    #doc = SimpleDocTemplate(outputFileName)
    #formatted = pprettyprint(parsed)
    #print formatted
    # outputFileName must be allowed to be a file-like object or a string!
    if outDir and (type(outputFileName) is StringType) and (not os.path.isabs(outputFileName)):
        outputFileName = os.path.join(outDir, outputFileName)

    if progressCallBack:
        progressCallBack('BEFORE_PRE_PROCESS', 0)
    # NOW PROCESS THE PARSED XML
    # create a names archive (to allow multipass crossreferencing
    namearchive = NamesArchive(passLimit)
    # repeat until done
    done = 0
    count = 1
    while not done:
        if verbose:
            print "Generating PDF...", count
        count = count+1
        if progressCallBack:
            progressCallBack('BEFORE_RML_CONTEXT', 0)
        context = RMLContext(namearchive, permitEvaluations)
        if progressCallBack:
            progressCallBack('AFTER_RML_CONTEXT', 0)
        #context.doc = doc
        context.filename = outputFileName
        if progressCallBack:
            progressCallBack('BEFORE_CONTROLLER_PROCESS', 0)

        formatted = Controller.process(parsed, context, parsed=1)
        if progressCallBack:
            progressCallBack('AFTER_CONTROLLER_PROCESS', 0)
        context.filename = outputFileName
        # hacker check
        if CiD != id(CanvasOps):
            raise SV, SV
        if AiD != id(alignmentAdjust):
            raise SV, SV
        # if it is a simple document, just render the pages
        pDs = context.pageDrawings
        if pDs:
            if progressCallBack:
                progressCallBack('BEGIN_FORMAT', 0)
                progressCallBack('SIZE_EST', len(pDs))
            if verbose:
                print "Gen: pageDrawing style document"
            cI = context.canvasInfo
            cI.filename = context.filename
            from reportlab.pdfgen import canvas
            PAGECOMPRESSION = 0
            canv = canvas.Canvas(cI.filename, cI.pageSize, pageCompression=PAGECOMPRESSION,
                                 invariant=context.invariant)
            if encryption:
                canv._doc.encrypt = encryption
            canv.setPageCallBack(pageCallBack)
            canv.imageCaching = cI.imageCaching
            # security
            alignmentAdjust(canv)
            doc = None # not used (?)
            for pD in pDs:
                pD(canv, doc)
                canv.showPage()
                if progressCallBack:
                    progressCallBack('PROGRESS', canv.getPageNumber())
                    progressCallBack('PAGE', canv.getPageNumber())
            done = namearchive.complete()
            if done: canv.save()
            # destroy the context object (break possible cycles)
            context.destroy()
            # return the canvas info object, with last canvas attached
            cI.canv = canv
            outputDevice = canv._filename
            result = cI
        else:
            if verbose:
                print "Gen: page template style document"
            doc = context.doc
            doc.invariant = context.invariant
            doc.setPageCallBack(pageCallBack)
            doc.setProgressCallBack(progressCallBack)
            Story = context.story
            # prepare the context
            context.prepare()
            doc._doSave = 0
            doc.build(Story)
            doc._doSave = done = namearchive.complete()
            if encryption:
                doc.canv._doc.encrypt = encryption
            if done: doc.canv.save()
            context.destroy()
            result = doc
            outputDevice = doc.canv._filename
        if not done:
            done = ignoreDefaults # if set this overrides, forces early termination
        if not done:
            namearchive.reset() # raises error on no progress or if too many passes are executed
            if verbose:
                print "Gen: namearchive not completely defined: trying again..."
    #... go to top and try again (unless limit reached)

    # the progress callback should be told the size of the output if possible
    if hasattr(outputDevice, 'tell'): # stream
        size = outputDevice.tell()
    elif os.path.isfile(outputDevice): #file on disk
        size = os.stat(outputDevice)[6]
    else:
        size = 0    # we won't guarantee all cases yet

    if progressCallBack:
        progressCallBack('PDF_SIZE', size)
        progressCallBack('FINISHED', 0)
    return result

TOPLEVELCLASSES = (LazyParagraph, Table, xTable, Spacer, CondPageBreak, LazyPreformatted,
                    NextPageTemplate, PageBreak, ActionFlowable, IllustrationFlowable,
                   LazyPlugInFlowable, OutlineEntry, FrameBreak.__class__)

class ProgressMonitor:
    "Callback for logging. Gives main events and every 10th page."
    import time
    def __init__(self):
        self.started = time.clock()

    def __call__(self, typ, value):
        if typ == 'PROGRESS':
            return
        elif typ == 'PAGE' and value % 10 <> 0:
            return
        else:
            elapsed = time.clock() - self.started
            print "%0.4f seconds:  %s     %s" % (elapsed, typ, value)

def main(quiet=0,exe=0,fn=[],outDir=None):
    '''use exe=1 if this is the exe'''
    cwd = os.getcwd()
    if outDir and not os.path.isabs(outDir):
        outDir = os.path.join(cwd,outDir)
    try:
        if len(fn):
            # extract options
            if '-v' in fn:
                verbose = 1
                fn.remove('-v')
            else:
                verbose = 0

            for inputPath in fn:
                if not os.path.isabs(inputPath):
                    inputPath = os.path.join(cwd,inputPath)
                targetDir=os.path.dirname(inputPath)
                InputFile = open(inputPath)
                os.chdir(targetDir)
                if verbose:
                    progressCB = ProgressMonitor()
                else:
                    progressCB = None

                test = go(InputFile.read(),outDir=outDir, progressCallBack=progressCB,
                          dynamicRml=1)
                os.chdir(cwd)
                if not quiet: print test.filename
        else:
            test = go(sys.stdin.read(),outDir=outDir)
            if not quiet: print test.filename
    finally:
        os.chdir(cwd)

if __name__=="__main__":
    main(exe=1,fn=sys.argv[1:])
